///! api.rs June 2022
///! This module provides control structs for defining, executing,
///! and reporting the result of tasks. Each task accepts a single
///! set of specifications for a given multiplication need. The task
///! then computes a series of multipliers that meet these
///! specifications, and can either act on all of them or act only
///! on the "best."

use crate::*;

/// `Signage` is a property of a factor, subfactor, product, or subproduct.
/// `Unsigned` and `Signed` have obvious meaning. `Either` means the
/// control decoder decides at the time of the multiplication, by sending
/// a 1 bit for Signed and 0 for Unsigned.
#[derive(Copy, Clone, Debug, PartialEq)]
pub enum Signage {
    Unsigned,
    Signed,
    Either,
}
pub use Signage::*;

/// `Specs` is a set of parameters sufficient for deterministically
/// creating a usable multiplier. This struct is intended to represent
/// a whole-word multiplier, so it's not used to describe parameters
/// for subword (subproduct) multiplication.
#[derive(Clone, Copy, Debug)]
pub struct Specs {
    pub width_l: usize,
    pub width_r: usize,
    pub sue_l: Signage,
    pub sue_r: Signage,
    pub l0_ram_dims: (usize, usize),    // TODO stratify by layer
}

/// `Which` controls which synthesized multipliers are eligible for processing.
#[derive(Clone, Copy, Debug, PartialEq)]
pub enum Which {
    Omit,
    Best,
    All,
}
pub use Which::*;

/// `Scope` determines what is to happen during a single Specs task.
#[derive(Clone, Copy, Debug)]
pub struct Scope {
    pub n_test_muls: usize,             // # of test products, per multiplier
    pub draw: Which,                    // what will be drawn
    pub synth: Which,                   // what will be synthesized
    pub test: Which,                    // what will be tested (requires synth)
    pub abandon_overcost: bool,         // halt designs once known to be losing
    pub d_flip_flops_avail: bool,       // enables cross-layer composition
    pub strategy: OptimizeStrategy,     // how exhaustively to search
    pub join_reservation: usize,        // only for multilayer carry-skip
}

/// `Report` is the essential output to the main logic of the program.
#[derive(Clone, Debug)]
pub struct Report {
    // copy of input parameters for job
    pub specs: Specs,                   // desired multiplier parameters
    pub scope: Scope,                   // desired operation of this run

    // job output
    pub product_width: usize,           // width of product in bits
    pub heading: ReportHeading,         // gets printed at top of output
    pub best_soln: Vec<Package>,        // cheapest solution achieving specs
    pub best_cost: SolutionCost,        // cost of best_soln
    pub something_failed: bool,         // true if any error is captured
    pub n_configs: u64,                 // # of solutions found to specs
    pub n_tests_right: u64,             // # of correct test products
    pub n_tests_wrong: u64,             // # of incorrect test products
}

/// `SolutionCost` means
/// (# layers, # packages, # active bits in total, # of composed RAMs).
pub type SolutionCost = (usize, usize, usize, usize);

/// `ReportHeading` exists because it needs to be printed at different
/// times, depending on whether we're single-threaded and printing all
/// the configs, or multi-threaded and printing just the best.
#[derive(Clone, Copy, Debug)]
pub struct ReportHeading {
    pub n_l_parts: usize,               // # of left subwords
    pub n_r_parts: usize,               // # of right subwords
    pub est_n_subprod_bits: usize,      // # approx subproduct total bits
}

/// `Tasker` accepts a Spec with a Scope, runs what is needed concurrently,
/// retrieves a Report, and prints the outcome. The constructor is
/// `with_threads` and takes the desired number of threads.
pub struct Tasker {
    n_threads: usize,
    jh_rx: Vec<(std::thread::JoinHandle<()>,
            std::sync::mpsc::Receiver<Report>)>,
    anything_failed: bool,              // true if any errors were captured
    total_reports: u64,                 // # of reports tabulated
    total_configs: u64,                 // # solutions found for all specs
    total_tests_right: u64,             // # correct products for all specs
    total_tests_wrong: u64,             // # incorrect products for all specs
}

/// `Display` for `Signage` writes 's', 'u', or 'e'.
/// `Debug` for `Signage` writes `Signed`, `Unsigned`, or `Either`.
impl std::fmt::Display for Signage {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", match self {
            Unsigned => 'u',
            Signed => 's',
            Either => 'e',
        })
    }
}

impl std::fmt::Display for Specs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}{} x {}{} w/ l0 = {:?}", self.width_l, self.sue_l,
                self.width_r, self.sue_r, self.l0_ram_dims)
    }
}

impl Tasker {
    // The main thread calls this constructor exactly once.
    // This struct will distribute the work of this program among
    // the number of threads requested here.
    pub fn with_threads(n_threads: usize) -> Tasker {
        assert!(n_threads != 0);
        Tasker {
            n_threads,
            jh_rx: Vec::new(),
            anything_failed: false,
            total_reports: 0,
            total_configs: 0,
            total_tests_right: 0,
            total_tests_wrong: 0,
        }
    }

    // The main thread does this (except for the closure for spawn).
    // Schedule another multiplier design, and start the next design
    // in the queue if a thread is available.
    pub fn go(&mut self, specs: &Specs, scope: &Scope) {
        // UNTHREADED MODE
        if self.n_threads == 1 {
            let report = plan_multiplier(specs, scope);
            self.process_report(report);
            return;
        }

        // THREADED MODE
        // Don't draw from multiple threads simultaneously.
        if scope.draw == All {
            fail("Scope of draw may not be All using multiple threads.");
        }

        // If we're out of threads, wait for the first-queued thread
        // to send a report (which we print) and exit. Using separate
        // channels in order of creation will cause this program's
        // printed output to be in the natural order of work, regardless
        // of the number of threads used. For efficient scheduling,
        // try to queue the tasks in ascending order of CPU time.
        if self.jh_rx.len() == self.n_threads {
            self.block_until_report();
        }

        // Start a new thread with its own channel for output.
        let (specs, scope) = (*specs, *scope);
        let (tx, rx) = std::sync::mpsc::sync_channel(0);
	let builder = std::thread::Builder::new().stack_size(512 * 1024 * 1024);
        let jh = builder.spawn(move || {
            // THIS BLOCK IS THE ENTIRE CHILD THREAD.
            let report = plan_multiplier(&specs, &scope);
            tx.send(report).unwrap();
        }).unwrap();

        // Add the handle and receiver to the queue.
        self.jh_rx.push((jh, rx));
    }

    // The main thread does this. Wait for a report from another thread,
    // process the report, then wait for the thread to have exited.
    fn block_until_report(&mut self) {
        // Wait for the first thread to report and exit.
        let (jh, rx) = self.jh_rx.remove(0);
        let report = rx.recv().unwrap();

        // Use the available core to process the report.
        self.process_report(report);

        // That should have given the thread time to quit.
        jh.join().unwrap();
    }

    // The main thread does this.
    // Digest information from a report. Print solution if requested.
    fn process_report(&mut self, r: Report) {
        self.anything_failed |= r.something_failed;
        self.total_reports += 1;
        self.total_configs += r.n_configs;
        self.total_tests_right += r.n_tests_right;
        self.total_tests_wrong += r.n_tests_wrong;

        if r.scope.draw == Omit {
            return;
        }

        // If we're drawing all the configs, we needed the heading printed
        // earlier. And so we did that.
        if r.scope.draw == Best {
            r.heading.print(&r.specs);
        }

        if r.best_cost == MAX_SOLN_COST {
            p!(R "No supported solutions found.\n");
        } else {
            let pruned = if r.scope.abandon_overcost
                    { "(pruned) " } else { "" };
            p!("Cheapest costs {:?} after {} {}trials:\n",
                r.best_cost, r.n_configs, pruned);
            draw_solution
                    (&r.best_soln, r.product_width, r.scope.join_reservation);
        }
    }
}

impl ReportHeading {
    pub fn print(&self, specs: &Specs) {
        p!(B "  configuration:  {}", specs);
        p!(B "       subwords:  {} x {}", self.n_l_parts, self.n_r_parts);
        p!(B "subproduct bits:  {} +/-\n", self.est_n_subprod_bits);
    }
}

/// This keeps the Tasker (and therefore the program) alive until all
/// threads have finished processing and all output has been printed.
///
/// This can cause confusion (even for single-thread runs) if the program
/// panics, because you'll get the "pristine" message after a crash!
/// To overcome this, put panic = "abort" in Cargo.toml.
impl Drop for Tasker {
    fn drop(&mut self) {
        while !self.jh_rx.is_empty() {
            self.block_until_report();
        }

        p!("total reports: {}", self.total_reports);
        p!("total configs: {}", self.total_configs);
        p!("mults correct: {}", self.total_tests_right);
        p!("mult failures: {}", self.total_tests_wrong);

        if self.anything_failed {
            p!(e "This run had problems.");
        } else {
            p!(G "This run is pristine.");
        }
    }
}
