///! bigrand.rs February 2022
///! This module generates multiplies large numbers programatically
///! (no RAMs) for the purpose of generating regression test cases.

use crate::*;

/// Arbitrary-length signed or unsigned integers.
/// We can generate, multiply, and print out in hexadecimal. That's about all.
#[derive(Clone)]
pub struct BigRand {

    /// true iff this BigRand is signed
    pub is_signed: bool,

    /// effective word size of this BigRand
    pub n_bits: usize,

    /// mask of bits used in the most significant u64
    pub top_mask: u64,

    /// mask indicating location of "sign bit", if any, in most significant u64
    pub neg_mask: u64,

    /// BigRand contents as vector of u64s, least significant first
    pub v: Vec<u64>,
}

impl BigRand {

    /// Create a BigRand with value 0.
    /// You specify word size and whether or not signed.
    pub fn zero(is_signed: bool, n_bits: usize) -> BigRand {
        if n_bits == 0 && is_signed {
            fail("signed integers cannot be 0 bits wide");
        }

        let q = n_bits / 64;
        let r = n_bits % 64;
        let n_u64s = (r != 0) as usize + q;

        let mut v = vec!(0; n_u64s);

        let top_mask;
        let neg_mask;
        if r != 0 {
            neg_mask = if is_signed { 1 << r - 1 } else { 0 };
            top_mask = (1 << r) - 1;
            v[q] &= top_mask;
        }
        else {
            neg_mask = if is_signed { 1 << 63 } else { 0 };
            top_mask = u64::MAX;
        }

        BigRand { is_signed, n_bits, top_mask, neg_mask, v }
    }

    /// Create and initialize a BigRand. You specify word size and
    /// whether or not signed. Value comes from /dev/urandom.
    pub fn new(is_signed: bool, n_bits: usize, rng: &mut Urand64) -> BigRand {
        let mut r = BigRand::zero(is_signed, n_bits);

        for e in r.v.iter_mut() {
            *e = rng.next();
        }

        r.canonicalize();
        r
    }

    /// Create a BigRand with value 2**n.
    pub fn power_of_two(is_signed: bool, n_bits: usize, n: usize) -> BigRand {
        assert!(n < n_bits);
        let q = n / 64;
        let r = n % 64;
        let mut x = BigRand::zero(is_signed, n_bits);
        x.v[q] = 1 << r;
        x
    }

    /// Create a BigRand from a Tally. Note that Tally is unsigned always.
    pub fn from(t: &Tally) -> BigRand {
        let mut r = BigRand::zero(false, t.bits.len());
        for (pv, bit) in t.bits.iter().enumerate() {
            r.set_bit(pv, *bit);
        }
        r
    }

    /// Read a specific bit from a BigRand.
    pub fn read_bit(&self, bit: usize) -> bool {
        assert!(bit < self.n_bits);
        let q = bit / 64;
        let r = bit % 64;
        self.v[q] & 1 << r != 0
    }

    /// Set a specific bit in a BigRand.
    pub fn set_bit(&mut self, bit: usize, val: bool) {
        assert!(bit < self.n_bits);
        let q = bit / 64;
        let r = bit % 64;
        self.v[q] = self.v[q] & !(1 << r) | (val as u64) << r;
    }

    /// Print a BigRand directly in base 2.
    pub fn print_bin(&self, prefix: &str, postfix: &str) {
        pp!("{}", prefix);
        for e in self.v.iter().rev() {
            pp!("{:064b}", e);
        }
        pp!(" {} {}", if self.is_negative() { '-' } else { '+' }, postfix);
    }

    /// Print a BigRand directly in base 16.
    /// We don't need printing in base 10, so we don't support that.
    pub fn print_hex(&self, prefix: &str, postfix: &str) {
        pp!("{}", prefix);
        for e in self.v.iter().rev() {
            pp!("{:016x}", e);
        }
        pp!(" {} {}", if self.is_negative() { '-' } else { '+' }, postfix);
    }

    /// Print a BigRand directly in base 16, but supply a "minus()" wrapper
    /// to indicate negative values. Used for Python regression testing.
    fn print_hex_python(&self) {
        let n = self.is_negative();
        pp!(n: "minus(");
        pp!("0x");
        for e in self.v.iter().rev() {
            pp!("{:016x}", e);
        }
        pp!(n: ")");
    }

    /// Canonicalize a BigRand so there are no unnecessary u64s,
    /// and no bits are set past the word size.
    fn canonicalize(&mut self) {
        let q = self.n_bits / 64;
        let r = self.n_bits % 64;
        let n_u64s = (r != 0) as usize + q;

        self.v.truncate(n_u64s);
        if n_u64s != 0 {
            self.v[n_u64s - 1] &= self.top_mask;
        }
    }

    /// Compute the twos complement of a BigRand, whether or not
    /// it is signed.
    fn complement(&self) -> Self {
        let mut c = self.clone();
        for chunk in c.v.iter_mut() {
            *chunk = !*chunk;
        }
        for chunk in c.v.iter_mut() {
            let one_up = chunk.wrapping_add(1);
            *chunk = one_up;
            if one_up != 0 {
                break;
            }
        }
        c.canonicalize();
        c
    }

    /// Determine whether or not a BigRand is negative.
    fn is_negative(&self) -> bool {
        if !self.is_signed {
            return false;
        }
        // is_signed guarantees self.v is non-empty.
        self.v.last().unwrap() & self.neg_mask != 0
    }

    /// Multiply two BigRand, possibly signed, returning the product.
    pub fn multiply(&self, other: &BigRand) -> BigRand {

        /// Helper function: add u64 `x` with carry into vector `sum` of u64
        /// starting at position `pos`. `sum` must already be wide enough.
        fn accumulate(sum: &mut Vec<u64>, mut x: u64, mut pos: usize) {
            while x != 0 {
                let (s, carry) = sum[pos].overflowing_add(x);
                sum[pos] = s;
                pos += 1;
                x = carry.into();
            }
        }

        // The product of two 1-bit signed numbers is unsigned!
        // Otherwise, the product is signed iff either factor is signed.
        let prod_signed = if self.is_signed && other.is_signed
                && self.n_bits == 1 && other.n_bits == 1 {
            false
        }
        else { self.is_signed | other.is_signed };

        // *a is a Vec representing abs(self).
        let mut neg_result = false;
        let a;
        let a_comp;
        if self.is_negative() {
            neg_result = !neg_result;
            a_comp = self.complement();
            a = &a_comp.v;
        }
        else {
            a = &self.v;
        }

        // *b is a Vec representing abs(other).
        let b;
        let b_comp;
        if other.is_negative() {
            neg_result = !neg_result;
            b_comp = other.complement();
            b = &b_comp.v;
        }
        else {
            b = &other.v;
        }

        // Compute *c, a Vec representing abs(self * other).
        let mut prod = BigRand::zero(prod_signed, self.n_bits + other.n_bits);
        let mut c = &mut prod.v;

        // Multiply the subproducts and accumulate their sum.
        // Rust has a widening_mul() method for u64 in nightly builds, but
        // it is not yet standardized as of this writing. We won't use it.
        for (ai, ae) in a.iter().enumerate() {
            for (bi, be) in b.iter().enumerate() {
                let hilo = *ae as u128 * *be as u128;
                let hi = (hilo >> 64) as u64;
                let lo = hilo as u64;
                accumulate(&mut c, lo, ai + bi);
                accumulate(&mut c, hi, ai + bi + 1);
            }
        }

        // Return the product with the correct sign.
        if neg_result { prod.complement() } else { prod }
    }

    // Add two unsigned, equal-sized BigRand with wrap.
    pub fn special_case_add(&self, other: &Self) -> Self {
        assert!(!self.is_signed);
        assert!(!other.is_signed);
        assert!(self.n_bits == other.n_bits);

        let mut sum = BigRand::zero(self.is_signed, self.n_bits);
        let mut c_old = false;
        for i in 0..sum.v.len() {
            let mut s;
            let mut c;

            // Sorry for the ugliness. I'm running in Rust 1.57,
            // and destructuring assignments are unstable still.
            {
                let (ss, cc) = self.v[i].overflowing_add(other.v[i]);
                s = ss;
                c = cc;
            }

            if c_old {
                let (ss, cc) = s.overflowing_add(1);
                s = ss;
                c |= cc;
            }

            sum.v[i] = s;
            c_old = c;
        }

        sum.canonicalize();
        sum
    }

}

/// Function to verify correct BigRand multiplication:
/// Manually use Python to ensure each produced a * b - c is actually zero.
pub fn test_bigrand() {
    p!(Y "test_bigrand output:\n");
    let mut rng = Urand64::new();
    for i in 1..132 {        // TODO was 132, should revert 25Apr2022
        p!("{} bits:", i);
        for _ in 0..10 {
            let a = BigRand::new(true, i, &mut rng);
            let b = BigRand::new(true, i, &mut rng);
            let c = a.multiply(&b);
            a.print_hex_python();
            pp!(" * ");
            b.print_hex_python();
            pp!(" - ");
            c.print_hex_python();
            p!();
        }
        p!();
    }
    p!(b "# Python 3:");
    p!(b "def minus(x):");
    p!(b "\treturn -(~x + 1 & (1 << x.bit_length()) - 1)\n");
}
