///! compo.rs June 2022
///! Find pairs of RAMs that can be composed into one.

use crate::*;

pub fn optimize_compositions(w: &mut Wires, do_vert: bool, do_horiz: bool) {
    remove_deadbeat_joiners(w);

    w.n_compos = 0;
    if do_vert {
        optimize_vertical_compositions(w);
    }
    if do_horiz {
        // Time complexity of this is not good for large layers.
        // Don't call it for 4096u x 4096u OptimizeMulti tasks!
        optimize_horizontal_compositions(w);
    }

    w.ram_layers.retain(|x| !x.is_empty());
}

fn optimize_vertical_compositions(w: &mut Wires) {
    loop {
        let mut any_swaps = false;
        let pairs = find_vertical_compositions(w);
        break_if!(pairs.is_empty());

        for (s_marker, d_marker) in pairs {

            // For both markers, I need the RAM, layer number, and index
            // within the layer. We know the d layer will be after the s layer.
            // These might not be found if something has changed.
            if let Some((s_ram, s_layer_number, s_index)) =
                    find_marker(&w.ram_layers, s_marker) {
                if let Some((d_ram, d_layer_number, d_index)) =
                        find_marker(&w.ram_layers[s_layer_number..], d_marker) {
                    let d_layer_number = d_layer_number + s_layer_number;

                    // Compose and replace these two RAMs.
                    let replacement = s_ram.compose_top_bot(d_ram);
                    w.ram_layers[s_layer_number].remove(s_index);
                    w.ram_layers[d_layer_number][d_index] = replacement;
                    any_swaps = true;
                    w.n_compos += 1;

                    // Note if we've contaminated the subproduct RAM layer.
                    if s_layer_number == 0 {
                        w.l0_dirty = true;
                    }
                }
            }
        }
        break_if!(!any_swaps);
    }

}

// Map a marker -> (RAM, layer_number, index_within_layer).
fn find_marker<'y>(layers: &'y[Vec<RAM>], marker: usize) ->
        Option<(&'y RAM, usize, usize)> {
    for (layer_number, layer) in layers.iter().enumerate() {
        for (ram_index, ram) in layer.iter().enumerate() {
            if ram.marker == marker {
                return Some((ram, layer_number, ram_index));
            }
        }
    }
    None
}

// Put presently-unique markers on all RAMs.
fn set_markers(w: &mut Wires) {
    let mut next_marker = 0;
    for layer in w.ram_layers.iter_mut() {
        for ram in layer.iter_mut() {
            ram.marker = next_marker;
            next_marker += 1;
        }
    }
}

// This function returns a list of vertically-composed pairs of RAMs.
// The list is by marker, not pointer, so that deletion and creation of
// new nodes will not cause undefined behavior. Returns a Vec of
// (source_marker, sink_marker).
fn find_vertical_compositions(w: &mut Wires) -> Vec<(usize, usize)> {
    let mut pairs = Vec::new();

    // Vertical composition is only applicable if there are at least two layers.
    if w.ram_layers.len() < 2 {
        return pairs;
    }

    set_markers(w);
    let w: &Wires = w;                  // drop mutability

    // Have pointers that mean "zero RAMs" or "more than one RAM" instead
    // of the usual &RAM that means "this particular RAM."
    let nobody = &RAM::bogus();
    let multiple = &RAM::bogus();

    // Make a directory of which RAM reads from each node.
    let mut who_reads = vec![nobody; w.bits.len()];
    for layer in w.ram_layers[1..].iter() {
        for ram in layer {
            for bit_pos in &ram.inputs {
                if ptr_eq!(who_reads[*bit_pos], nobody) {
                    who_reads[*bit_pos] = ram;
                }
                else if !ptr_eq!(who_reads[*bit_pos], ram) {
                    who_reads[*bit_pos] = multiple;
                }
            }
        }
    }

    // Identify 'source' RAMs where:
    // a) all outputs lead to the same 'sink' RAM,
    // b) # sink inputs + # source inputs - # source outputs
    //    <= # avail sink inputs
    for source_layer in w.ram_layers[0..w.ram_layers.len() - 1].iter() {
        'next_source: for source in source_layer {
            assert!(source.outputs.len() != 0);
            let sink = who_reads[source.outputs[0]];
            for j in 1..source.outputs.len() {
                if !ptr_eq!(sink, who_reads[source.outputs[j]]) {
                    continue 'next_source;
                }
            }
            if sink.is_bogus() {
                continue 'next_source;
            }

            // If we're here, all `source` outputs lead to `sink`.
            // The composing RAM's dims will be the same as `sink`.
            // Outputs of `sink` are the same as the composing RAM.
            // Make sure we can fit all the inputs.
            if sink.inputs.len() + source.inputs.len() -
                    source.outputs.len() > sink.dims.0 {
                continue 'next_source;
            }

            // If we get here, we've found a composable pair.
            // That means we can replace these two RAMs with a single RAM.
            pairs.push((source.marker, sink.marker));
        }
    }

    // 22Aug2022: I thought that taking this list in reverse order may benefit
    // composition for OptimizeMulti, to make sure a Joiner is "tucked into"
    // the last CarrySkipFinish instead of being pulled up and away.
    // So far I haven't been able to find a case where this actually helps.
    pairs.reverse();
    pairs
}

// Determine how many elements of unsorted `a` and `b` are common.
// Neither vector is empty, and the members are unique within each.
fn n_overlapping(a: &[usize], b: &[usize]) -> usize {

    // Assume bounding boxes are faster than sorting.
    let a_min = a.iter().min().unwrap();
    let a_max = a.iter().max().unwrap();
    let b_min = b.iter().min().unwrap();
    let b_max = b.iter().max().unwrap();
    if a_max < b_min || b_max < a_min {
        return 0;
    }

    // Assume sorting is faster than hashing.
    let mut c = Vec::from(a);
    c.extend(b);
    c.sort();

    let mut n = 0;
    for j in 1..c.len() {
        if c[j-1] == c[j] {
            n += 1;
        }
    }

    n
}

fn optimize_horizontal_compositions(w: &mut Wires) {
    loop {
        let mut any_swaps = false;
        let pairs = find_horizontal_compositions(w);
        break_if!(pairs.is_empty());

        for (l_marker, r_marker) in pairs {

            // For both markers, I need the RAM, layer number, and index
            // within the layer. We know the d layer will be after the s layer.
            // These might not be found if something has changed.
            if let Some((l_ram, l_layer_number, l_index)) =
                    find_marker(&w.ram_layers, l_marker) {
                if let Some((r_ram, r_layer_number, r_index)) =
                        find_marker(&w.ram_layers, r_marker) {

                    // Compose and replace these two RAMs.
                    let replacement = l_ram.compose_l_r(r_ram);
                    w.ram_layers[r_layer_number][r_index] = replacement;
                    w.ram_layers[l_layer_number].remove(l_index);
                    any_swaps = true;
                    w.n_compos += 1;

                    // Note if we've contaminated the subproduct RAM layer.
                    if l_layer_number == 0 || r_layer_number == 0 {
                        w.l0_dirty = true;
                    }
                }
            }
        }
        break_if!(!any_swaps);
    }
}

// This function returns a list of horizontally-composed pairs of RAMs.
// The list is by marker, not pointer, so that deletion and creation of
// new nodes will not cause undefined behavior. Returns a Vec of
// (left_marker, right_marker). Incidentally, left and right aren't named
// after the RAM circuit positions, but their output bit positions in the
// composed RAM.
fn find_horizontal_compositions(w: &mut Wires) -> Vec<(usize, usize)> {
    let mut pairs = Vec::new();

    set_markers(w);
    let w: &Wires = w;                  // drop mutability

    for layer in &w.ram_layers {
        for a in layer {
            for b in layer {
                if a.marker >= b.marker {
                    continue;           // same RAM or symmetric dupe
                }
                if a.outputs.len() + b.outputs.len() > b.dims.1 {
                    continue;           // too many outputs together
                }
                if a.inputs.len() + b.inputs.len() -
                            n_overlapping(&a.inputs, &b.inputs) > b.dims.0 {
                    continue;           // too many inputs together
                }
                pairs.push((a.marker, b.marker));
            }
        }
    }
    pairs
}

// Remove Propagator (they're actually Joiner) RAMs with unread outputs.
// The caller will remove any empty layers after this.
fn remove_deadbeat_joiners(w: &mut Wires) {
    // Make a directory of which nodes are ever read.
    let mut ever = vec![false; w.bits.len()];
    for layer in w.ram_layers[1..].iter() {
        for ram in layer {
            for bit_pos in &ram.inputs {
                ever[*bit_pos] = true;
            }
        }
    }

    // See which Propagator RAMs have no outputs read.
    // They will come out with .marker != 0.
    for layer in w.ram_layers[1..].iter_mut() {
'ram:   for ram in layer.iter_mut() {
            ram.marker = if let RAMParms::Propagator { .. } = ram.func_parms {
                1
            } else { 0 };
            for bit_pos in &ram.outputs {
                if ever[*bit_pos] {
                    ram.marker = 0;
                    continue 'ram;
                }
            }
        }

        layer.retain(|x| x.marker == 0);
    }
}
