///! draw.rs May 2022
///! Solution-drawing routines for solver.rs

use std::clone::Clone;      // trait
use std::fmt::Debug;        // trait

use crate::Package;
use crate::PackageKind::{self, *};

/// Draw a tagged dot diagram of subproduct bits for one layer.
/// The input is a slice of 3-tuples, (global_pv, attribute, attribute),
/// and a function that can draw a single bit with the attributes given.
/// The bits are drawn from MSB (left) to LSB (right), with as many
/// lines as needed to show all of the bits.
pub fn draw_dots<S: Clone + Debug, T: Clone + Debug>
        (d: &[(usize, S, T)], print_dot: fn(S, T, &mut &'static str),
        product_width: usize, old_color: &mut &'static str) {
    let mut d = Vec::from(d);

    // This is a stable, reverse-order sort that ignores attr1 & attr2.
    // By ignoring the attributes, we're able to show bit values without
    // moving the bits around.
    d.sort_by(|a, b| b.0.cmp(&a.0));

    if d.is_empty() {
        return;
    }

    // This loop makes a "waterfall" effect where if a place value
    // has more than one element (very typical), successive elements
    // will be drawn on successive lines.
    while !d.is_empty() {
        let mut d_next = Vec::new();

        // Here is the "horizontal raster" loop.
        let mut prev_pv = product_width;

        for (pv, attr1, attr2) in d {
            if prev_pv == pv {
                d_next.push((pv, attr1, attr2));
                continue;
            }
            prev_pv -= 1;
            while prev_pv > pv {        // inequality indicates an empty pv
                change_color("\x1b[m", old_color);
                pp!(" ");
                prev_pv -= 1;
            }
            print_dot(attr1, attr2, old_color);
        }
        change_color("\x1b[m", old_color);
        p!();                           // eol
        d = d_next;
    }
}

/// This is a simple bit-drawing function for draw_dots, where the
/// first attribute determines what character 0-9, a-z, A-Z is printed.
/// The second attribute is ignored.
pub fn digit_dot<T>(digit: usize, _: T) {
    let digit = (digit % 62) as u8;
    let symbol = match digit {
        0..=9 => b'0' + digit,
        10..=35 => b'a' + digit - 10,
        36..=61 => b'A' + digit - 36,
        _ => b'?',
    };
    pp!("{}", symbol as char);
}

/// This de-duplicates ANSI terminal control sequences.
/// It doesn't de-duplicate subsequences, so there is still a little bloat.
/// It also doesn't de-duplicate catenated control sequences, although
/// this module doesn't employ those.
fn change_color(want: &'static str, have: &mut &'static str) {
    if want != *have {
        pp!("{}", want);
        *have = want;
    }
}

/// This is a digit_dot wrapper that adds color.
/// *old_color should ordinarily be "\x1b[m" on the first call to this.
/// Restoring terminal default color when done is the caller's responsibility.
fn package_kind_dot(a: usize, b: &PackageKind, old_color: &mut &'static str) {
    let color = match b {
        ArbGeom { finishing: false, .. } => "\x1b[0;33m",
        ArbGeom { finishing: true, .. } => "\x1b[0;7;33m",
        CarrySkipFinish { .. } => "\x1b[0;7;34m",
        CarrySkipSetup { finishing: false, .. } => "\x1b[0;31m",
        CarrySkipSetup { finishing: true, .. } => "\x1b[0;7;31m",
        Propagator { .. } => "\x1b[0;36m",
        Joiner { .. } => "\x1b[0;35m",
        Wireflop => {
            change_color("\x1b[m", old_color);
            pp!("-");
            return;
        },
        EndOfLayer => "\x1b[1;30;47m",        // error => inverse video
    };
    change_color(color, old_color);
    digit_dot(a, 0);
}

// `solution` is a list of packages for all layers, with EndOfLayer
// packages at the end of each and every layer. Make a list of
// slices for each layer, leaving EndOfLayer out.
pub fn split_solution_by_layers<'x>(solution: &'x[Package])
        -> Vec<&'x[Package]> {
    let mut layer_solns = Vec::new();
    let mut start = 0;
    for i in 0..solution.len() {
        if solution[i].kind == EndOfLayer {
            let slice = &solution[start..i];
            layer_solns.push(slice);
            start = i + 1;
        }
    }
    assert_eq!(start, solution.len());  // last solution included EndOfLayer
    layer_solns
}

// FIXME
// Drawing correctly from layer to layer has been an issue for a
// few days now. Would it be more expedient to explicitly list
// the consumed pvs within each package in the solver, and draw
// from that? Later on when things appear to work, we could look
// at transferring that solver overhead into a more complex
// print routine here.
pub fn draw_solution(
    solution: &[Package],
    product_width: usize,
    _join_reservation: usize,
) {
pp!("\x1b[?7l");    // wrap off (22Aug2022)

    let mut old_color = "\x1b[m";

    if solution.is_empty() {
        // Draw single-package solution.
        p!("This is ! a single-package solution.\n\n");
        return;
    }

    // Loop through the layer solutions.
    let layer_solns = split_solution_by_layers(solution);
    for layer_soln in &layer_solns {

        if layer_soln.is_empty() {
            p!(hw "[empty layer]\n");
            continue;
        }

        if let [Package { kind: Propagator { .. }, .. }, ..] = layer_soln {
            p!(hw "[carry propagation layer]\n");
            continue;
        }

        if let [Package { kind: Joiner { .. }, .. }, ..] = layer_soln {
            p!(hw "[joiner layer]\n");
            continue;
        }

        // Set up a "job" that will print this layer.
        // It's a list of place values, ordinal package number,
        // and the kind of package.
        let mut job = Vec::new();
        let mut present = vec![false; product_width];

        let mut pkg_id = 0;
        for pkg in *layer_soln {
            if pkg.kind != Wireflop {
                pkg_id += 1;
            }
            for pv in &pkg.input_pvs {
                present[*pv] = true;
                job.push((*pv, pkg_id, &pkg.kind));
            }
        }

        for (i, here) in present.iter().enumerate() {
            if !*here {
                job.push((i, 0, &Wireflop));
            }
        }
//p!(g "layer {}:", n); n += 1;
        draw_dots(&job, package_kind_dot, product_width, &mut old_color);
        change_color("\x1b[m", &mut old_color);
        p!();
    }
    p!();
}

/// Diagnostic
pub fn list_solution(solution: &[Package]) {
    for p in solution {
        if p.kind == EndOfLayer {
            p!();
            continue;
        }
        p!("{:?}", p);
    }
    p!();
}

/// This function corrects the output of a large run by
/// de-duplicating specific escape sequences. stdin to stdout.
pub fn dedupe () {
    use std::io::Write;

    let from_to = [
        (&b"\x1b[33m"[..], &b"\x1b[0;33m"[..]),
        (&b"\x1b[7;33m"[..], &b"\x1b[0;7;33m"[..]),
        (&b"\x1b[7;34m"[..], &b"\x1b[0;7;34m"[..]),
        (&b"\x1b[31m"[..], &b"\x1b[0;31m"[..]),
        (&b"\x1b[7;31m"[..], &b"\x1b[0;7;31m"[..]),
    ];

    let mut o = std::io::stdout();
    let mut have = Vec::from(&b"\x1b[m"[..]);
    let mut want = Vec::from(&b"\x1b[m"[..]);

    loop {
        // Get a line of input as a slice.
        let mut s = String::new();
        std::io::stdin().read_line(&mut s).unwrap();
        if s.len() == 0 {
            break;
        }
        let mut s = s.as_bytes();

        while s.len() != 0 {
            if s[0] != 0x1b {

                if have != want {
                    o.write_all(&want).unwrap();
                    have = want.clone();
                }

                // normal character
                o.write_all(&s[0..1]).unwrap();
                s = &s[1..];
                continue;
            }

            // uh oh, ANSI sequence
            // I always end these with 'm', so it's simple....
            let m_index = s.iter().position(|&r| r == 109).unwrap();
            let seq = &s[..=m_index];
            s = &s[m_index + 1..];

            want = seq.to_owned();
            for (f, t) in from_to {
                if f == want {
                    want = t.to_owned();
                    break;
                }
            }
        }
    }
}
