#![allow(unused_macros)]    // needed when maybe_color is disabled

///! macros.rs April 2022
///! This module supplies Marc's declarative macros p!() and pp!()
///! for conditional printing and color printing. See macro_tests()
///! below for a gist of how they work. Tested with rustc 1.57.0.

// XXX Dear reader, the available literature about Rust declarative
// macros is rather opaque. Here's a crucial point: nested macro
// substitutions happen from outside to inside, not the other way
// around. You have to plan for that. One consequence is that the
// result of a macro substitution cannot be matched to produce another
// substitution. So if you're expanding outside!(inside!(something)),
// outside! is removed from the expression first, and then inside!.
// If you really need inside! removed first, you'll need a procedural
// macro crate (*not* macro_rules!) -- that's a whole different game.

/// XXX Uncomment exactly one line to enable or disable ANSI escape sequences.
macro_rules! maybe_color {
    ( $($yes_color: tt)* ) => { $($yes_color)* };
//  ( $($no_color: tt)* ) => { "" };
}

// Convert a 1- or 2-letter code into part of an ANSI escape sequence
// to select a foreground color. Submacro.
macro_rules! color_code {
    // XXX The first seven colors omit the "0;" for normal intensity,
    // and thereby depend on an all-attribute reset between colors.
    (r) => { "31" };        // red
    (g) => { "32" };        // green
    (y) => { "33" };        // yellow
    (b) => { "34" };        // blue
    (p) => { "35" };        // purple
    (t) => { "36" };        // teal
    (w) => { "37" };        // white
    (R) => { "1;31" };      // bold red
    (G) => { "1;32" };      // bold green
    (Y) => { "1;33" };      // bold yellow
    (B) => { "1;34" };      // bold blue
    (P) => { "1;35" };      // bold purple
    (T) => { "1;36" };      // bold teal
    (W) => { "1;37" };      // bold white
    (hr) => { "2;31" };     // half red
    (hg) => { "2;32" };     // half green
    (hy) => { "2;33" };     // half yellow
    (hb) => { "2;34" };     // half blue
    (hp) => { "2;35" };     // half purple
    (ht) => { "2;36" };     // half teal
    (hw) => { "2;37" };     // half white
    (e) => { "1;2;7;37" };// serious error (crazy attributes)
}

// Convert a 1- or 2-letter code into an ANSI escape sequence
// to select a foreground color. Submacro.
macro_rules! color_esc {
    ($c: ident) => { concat!("\x1b[", color_code!($c), "m") };
}

// Progressively expand a pp!() or p!() format string
// to incorporate ANSI escape sequences for setting foreground colors.
//
// Instead of writing:
//     p!("Format string: {} green {}", "red", "blue");   =>
//
// We can now write:
//     p!("Format string: " r "{} " g "green " b "{}", "red", "blue");
macro_rules! reflow_fmt {
    (   $print_macro: ident 
        concat! $cats: tt
        $color: ident
        $fmt: literal
        $($rest: tt)*
    ) => {
        reflow_fmt!($print_macro concat!(concat! $cats, maybe_color!
            (color_esc!($color)), $fmt, maybe_color!("\x1b[m")) $($rest)* )
    };

    (   $print_macro: ident 
        concat! $cats: tt
        $fmt: literal
        $($rest: tt)*
    ) => {
        reflow_fmt!($print_macro concat!(concat! $cats, $fmt) $($rest)* )
    };

    (   $print_macro: ident 
        concat! $cats: tt
        $($rest: tt)*
    ) => {
        $print_macro!(concat! $cats $($rest)*)
    };
}

// Allow printing to happen conditionally based on an identifier that
// evaluates to boolean. May use with color output macros. Example:
//
//     p!(debug: "x is presently {:?}", x);
//
// can be used instead of:
//
//     if (debug) {
//         p!("x is presently {:?}", x);
//     }
//
macro_rules! try_opt_cond {
    // Format is preceded with a condition.
    ($print_macro: ident , $cond: ident : $($rest: tt)*) => {
        if $cond {
            reflow_fmt!($print_macro concat!() $($rest)*)
        }
    };

    // Format does not include a condition.
    ($print_macro: ident , $($rest: tt)*) => {
        reflow_fmt!($print_macro concat!() $($rest)*)
    };
}

/// Generalized print (with newline) with optional 
/// boolean condition and optional multiple colorization
macro_rules! p {
    ($($rest: tt)*) => {
        try_opt_cond!(println, $($rest)*);
    }
}

/// Generalized "partial print" (no newline) with optional
/// boolean condition and optional multiple colorization
macro_rules! pp {
    ($($rest: tt)*) => {
        try_opt_cond!(print, $($rest)*);
    }
}

/// Expand a list of identifiers with an optional prefix identifier
/// before it, and optional suffix characters after it. Rust's limited
/// set of fragment specifiers constrains what this macro can achieve.
macro_rules! many {
    ($($prefix: ident)+ , $item: ident $($items: ident)* , $($suffix: tt)*) => {
        $($prefix)+ $item $($suffix)*
        many!($($prefix)+, $($items)*, $($suffix)*);
    };  // recursively expand next identifier

    ($($prefix: ident)+ , , $($suffix: tt)*) => {
    };  // terminate
}

/// plain conditional break
macro_rules! break_if {
    ($($cond: tt)+) => { if $($cond)+ { break } }
}

/// plain conditional continue
macro_rules! continue_if {
    ($($cond: tt)+) => { if $($cond)+ { continue } }
}

/// conditional return with value (usually ())
macro_rules! return_if {
    ($cond: expr, $value: expr) => { if $($cond)+ { return $value } }
}

/// diagnostic return without unreachable warning for next statement
macro_rules! poof {
    ($value: expr) => { if true { return $value; } }
}

/// "insertable" panic that doesn't complain about unreachability
/// for statements that follow
macro_rules! polite_panic {
    ($($arg:tt)*) => { if true { panic!($($arg)*) } }
}

/// safe direct comparison of two pointers for same memory location
macro_rules! ptr_eq {
    ($left: expr, $right: expr) =>
        { std::ptr::eq($left as *const _, $right as *const _) }
}

pub fn macro_tests() {
    let debug = true;
    let nope = false;

    p!("hello");
    p!("You will see two blank lines.");
    p!();
    p!(debug:);
    p!(nope:);
    p!("testing {}", 1);
    p!("testing {} {}", 1, "two");
    p!("testing {} {:?}", 1, ("two", 3.));
    p!(r "Red!");
    p!(r "Red!" t " Teal!" b " Blue!");
    p!(hr "half" r " medium" R " full");
    p!(hy "half" y " medium" Y " full");
    p!(debug: r "Red!" g " Green!" R " Bold!" " You should see this!");
    p!(nope: r "Red!" g " Green!" R " Bold!" " You should not see this!");
    p!("Testing");
    p!("Testing {}", 1);
    p!("Testing {}", 1 + 1);
    pp!(debug: "Partial ");
    p!("line.");
    p!(debug: "Yes!");
    p!(nope: "You should not see this!");
    p!("multiple " "strings!");
    p!(Y "bold yellow " r "normal red");
    p!("Format string: " r "{} " g "green " b "{}", "red", "blue");
    p!(debug: "Counting " r "{} " g "{} " b "{}", 1, "two", 2 + 1);
}


