// included from solver.rs

// Parses the remainder of this layer as a sequence of ArbGeom and Wireflop
// without backtracking. This conserves space on the stack somewhat.
// The only order constraint is that Wireflop can't come first, because
// such can appear within a carry-skip setup sequence.
fn parse_arb_geom<'m>(
    g: &Globals,
    m: &mut MutGlobals,
    mut ms: Mindset,
    mut layer: &'m[LayerVocab],
) -> Option<&'m[LayerVocab]> {
    let mut yes = false;

    loop {
        if let Some((new_ms, new_layer)) = parse_wireflop(m, ms, layer) {
            ms = new_ms;
            layer = new_layer;
            yes = true;
            continue;
        }
        if let Some((new_ms, new_layer)) = subparse_arb_geom(g, m, ms, layer) {
            ms = new_ms;
            layer = new_layer;
            yes = true;
            continue;
        }
        break;
    }

    if yes {
        Some(layer)
    } else {
        None
    }
}

fn subparse_arb_geom<'m>(
    g: &Globals,
    m: &mut MutGlobals,
    ms: Mindset,
    mut layer: &'m[LayerVocab],
) -> Option<(Mindset, &'m[LayerVocab])> {
    // May configure this here a/o 24Aug2022.
    const NO_SINGLE_SETUPS: bool = false;

    // Disallow singleton carry-skip setup RAMs by backtracking over them.
    // But certain strategies force setups to be taken, including singletons,
    // so we're cautious about that.
    if NO_SINGLE_SETUPS && ms.stage_two_setup() && g.strategy == OptimizeAll {
        // XXX TODO
        // This is a problem, because these singleton carry-skips will be
        // replaced by greedier ArbGeoms, which in some instances result
        // in more expensive solutions. For example, symmetric 24u, 34u, 46u,
        // 47u, 52u, 58u, 64u show this behavior. So singleton carry-skip
        // setup RAMs are actually a GOOD thing, except they're a special
        // case of a much larger family of missing search space. To complicate
        // this more, 41u x 41u has the reverse problem, where singleton
        // carry-skip somehow manages to /worsen/ the solution. 21Aug2022
        return None;
    }

    // If carry-skip finish information was delivered here from the
    // prior layer, it reduces the number of input bits, but also
    // contributes carry and propagate bits that can increase the sum.
    // Overall it's usually a net decrease.
    let (
        finishing,          // true / false
        mut in_bits_used,   // computed max sum this package can produce
        mut sum,            // keep package from running out of input wires
    ) = if let [ FinishInfo { prop_carry_baggage, max_eff_carry_val }, .. ]
                = layer {
        layer = &layer[1..];
        (true, *prop_carry_baggage, *max_eff_carry_val)
    } else {
        (false, 0, 0)
    };

    let mut have = ms.missing_bits();   // # of cols we "have" taken
    let mut overshoot_complement = 0;
    let mut input_pvs = Vec::with_capacity(g.ram_dims.0);

    // The first Column or OffsetComplement initializes these.
    let mut base_col = 0;
    let mut cols_to_end = 0;
    let mut max_sum = 0;                // 0 is never an initialized max_sum.

    'full: while !layer.is_empty() {

        // Watch for the new RAM's base column.
        if max_sum == 0 {
            base_col = match layer[0] {
                OffsetComplement { gpv } => gpv,
                Column { gpv, .. } => gpv,
                _ => return None,
            };

            // Set correct limit on how high this package may add.
            cols_to_end = g.product_width - base_col;
            max_sum = if cols_to_end < g.ram_dims.1 {
                usize::MAX              // package truncated => no maximum
            } else {
                (1 << g.ram_dims.1) - 1 // maximum package can output
            };
        }

        match layer[0] {

            // Include an overshoot complement in this RAM.
            OffsetComplement { gpv } => {

                // If the sum would be too high, leave it out and stop.
                let addend = 1 << gpv - base_col;
                let new_sum = sum + addend;
                if new_sum > max_sum {
                    break;
                }

                // This RAM can fit this overshoot complement.
                sum = new_sum;
                overshoot_complement += addend as u64;
            },

            // Include as many columns we can. There may be partial columns.
            Column { gpv, n: want } => {
                let addend = 1 << gpv - base_col;

                // Include all the bits we can from this column.
                while have < want {

                    // Can we fit another input bit?
                    assert!(in_bits_used <= g.ram_dims.0);
                    if in_bits_used == g.ram_dims.0 {
                        break 'full;
                    }

                    // Will the sum stay representable in the output bits?
                    let new_sum = sum + addend;
                    if new_sum > max_sum {
                        break 'full;
                    }

                    // We can fit this bit.
                    sum = new_sum;
                    in_bits_used += 1;
                    have += 1;
                    input_pvs.push(gpv);
                }

                // We've taken the entire column.
                have = 0;
            },

            _ => break,
        }

        layer = &layer[1..];
    }

    assert_eq!(sum == 0, max_sum == 0);
    if sum == 0 {
        return None;
    }

    // Determine the output width of this RAM.
    let out_width = sum.bit_length().min(cols_to_end);

    // Inform package to left if we've taken a partial column.
    let ms = if have == 0 { NeverMind } else { MissingBits(have) };

    // There is a situation where we might not take any addend bits.
    assert!(finishing || !input_pvs.is_empty() || overshoot_complement != 0);

    // Emit the package.
    let pkg = Package {
        kind: ArbGeom { overshoot_complement, finishing },
        mem_used: out_width << in_bits_used,
        output_pvs: (base_col, base_col + out_width),
        input_pvs,
    };
    m.trial_soln.push(pkg);

    // Put this ArbGeom's output bits into the next layer.
    // It's possible the first bit will overlap a prior ArbGeom.
    m.overlap_arbgeom_output(base_col, base_col + out_width);

    // Return a new mindset, and new position.
    Some((ms, layer))
}
