// included from solver.rs

// This module provides multilayer carry-skip setup.
// It's only valid today for the OptimizeMulti strategy, because it
// doesn't accept carry-skip finish information from the layer above it.
// Everything above should be ArbGeom.

// MWA MWA MWA MWA MWA MWA MWA MWA MWA MWA MWA
/* Here is some data for you that I trust as of 22Aug2022.

      configuration:  32u x 32u w/ l0 = (18, 18)
           subwords:  5 x 3
    subproduct bits:  256 +/-
    Cheapest costs (6, 40, 86839360, 0) after 1 (pruned) trials:

      configuration:  64u x 64u w/ l0 = (18, 18)
           subwords:  8 x 7
    subproduct bits:  960 +/-
    Cheapest costs (6, 143, 287047690, 1) after 1 (pruned) trials:

      configuration:  128u x 128u w/ l0 = (18, 18)
           subwords:  16 x 13
    subproduct bits:  3712 +/-
    Cheapest costs (7, 526, 1415782400, 6) after 1 (pruned) trials:

      configuration:  256u x 256u w/ l0 = (18, 18)
           subwords:  32 x 26
    subproduct bits:  14848 +/-
    Cheapest costs (8, 2042, 5378101864, 25) after 1 (pruned) trials:

      configuration:  512u x 512u w/ l0 = (18, 18)
           subwords:  57 x 57
    subproduct bits:  58368 +/-
    Cheapest costs (9, 7899, 21684594176, 52) after 1 (pruned) trials:
    real 3.949s

      configuration:  1024u x 1024u w/ l0 = (18, 18)
           subwords:  114 x 114
    subproduct bits:  233472 +/-
    Cheapest costs (10, 31295, 85062327270, 179) after 1 (pruned) trials:
    real 16.019s

      configuration:  2048u x 2048u w/ l0 = (18, 18)
           subwords:  228 x 228
    subproduct bits:  933888 +/-
    Cheapest costs (12, 124653, 338455229862, 304) after 1 (pruned) trials:
    real 126.824s

      configuration:  4096u x 4096u w/ l0 = (18, 18)
           subwords:  456 x 456
    subproduct bits:  3735552 +/-
    Cheapest costs (13, 497306, 1347697573632, 523) after 1 (pruned) trials:
    real 1014.939s

      configuration:  8192u x 8192u w/ l0 = (18, 18)
           subwords:  911 x 911
    subproduct bits:  14925824 +/-
    Cheapest costs (14, 1981799, 5404324177388, 1687) after 1 (pruned) trials:
    real 7080s
*/

// This parses an entire layer containing only Column, with 1 or 2 bits
// in every place value. If successful, the entire balance of the
// multiplier except the final EndOfLayer is appended to m.trial_soln,
// and this function returns true. Otherwise nothing is changed, and
// this function returns false.
pub fn parse_multilayer_setup<'m>(
    g: &Globals,
    m: &mut MutGlobals,
    mut layer: &'m[LayerVocab],
) -> bool {
    assert!(g.strategy == OptimizeMulti);

    // It's possible to reserve a specific number of input bits
    // on finish RAMs for joining multilayer carry-skip information.
    //
    // 0 means "reserve on a case-by-case basis the exact number
    //   needed to do the entire aggregation in the finish RAM."
    //   This setting will not work for all problem sizes and
    //   RAM dimensions.
    //
    // 1 means reserve just one bit. In many cases, this will force
    //   insertion of a joining RAM layer between the multilayer
    //   input tree and the finish RAMs, for a product latency
    //   penalty of one clock cycle.
    //
    // 2 and other even numbers are not advantageous over their
    //   next-smaller odd number.
    //
    // 3 reserves 3 bits, and is sufficient to always eliminate
    //   the one-cycle penalty. Any additional joining can happen
    //   in parallel with the input tree RAMs. In practice the
    //   output tree is wired as if only 1 bit was reserved, but
    //   two more bits on the finish RAM are left free. There will
    //   be an unnecessary layer inserted between the input tree
    //   and the finish RAMs to do the final join. But that gets
    //   corrected by post-synthesis vertical composition, which
    //   will see the optimization and apply it.

    // This feature has a limit of two input bits per place value.
    // Confirm that we don't exceed that. The "real" limit is no
    // more than one carry bit per stage, but a look at real-life
    // data shows that when there are *sometimes* more than two inputs
    // per place value, it happens *often*, essentially guaranteeing
    // that the one carry bit limit would be violated. To keep the
    // code simple, I neglect the possibility of an exception.
    for stuff in layer {
        if let Column { n, .. } = stuff {
            if *n > 2 {
                return false;
            }
        } else {
            return false;
        }
    }

    // There is nothing to do for trailing columns of one bit.
    // XXX The caller may already have addressed this.
    while let [Column { n: 1, .. }, ..] = layer {
        layer = &layer[1..];
    }
    let layer = layer;                  // Drop mutability.

    // We've proven this can succeed, but we don't know how many
    // propagator layers we'll need. We'll try 0 first, and keep
    // increasing layers until the parsing succeeds.
    for n_layers in 0.. {
        let mut layers = make_propagator_layers(n_layers, &g.ram_dims);
        let mut joiners = trace_propagator_joiners(&layers);

        // We need a list of how many bits should be reserved
        // by each of the finish RAMs.
        let mut reserve = Vec::with_capacity(joiners.len());
        for j in &joiners {
            reserve.push(if g.join_reservation == 0 {
                let jl = j.len();
                match jl {
                    0 | 1 => jl,
                    _ => 2 * jl - 1,
                }
            } else { g.join_reservation });
        }

        // For each n_layers, start in the same correct place.
        let mut layer = layer;
        let mut rightmost = true;

        let mut pkgs = Vec::new();

        // Take all of the columns until we're out of allowed packages
        // (we're allowed one per element of `reserve`) or out of input.
        for reserve_bits in reserve {
            if layer.is_empty() {
                break;                  // successfully completed
            }

            if let Some((pkg, remainder)) =
                    parse_one_stage(g, layer, reserve_bits, rightmost) {
                pkgs.push(pkg);
                layer = remainder;
                rightmost = false;
            } else {
                // XXX This fails for RAMs smaller than (4, 3).
                return false;
            }
        }

        if !layer.is_empty() {
            // We need more setup packages than we can accommodate
            // using this number of layers. Enlarge.
            continue;
        }

        // We succeeded parsing. Before we generate the packages and
        // synthesize the RAMs, trim the tree to fit the packages.
        // FIXME I'm not sure the left side is trimmed fully. 22Aug2022
        trim_tree(&mut layers, &mut joiners, pkgs.len());

        // See doc for multilayer_scheme() for what now.
        multilayer_scheme(m, pkgs, layers, joiners);
        return true;
    }

    unreachable!();
}

fn parse_one_stage<'m>(
    g: &Globals,
    mut layer: &'m[LayerVocab],
    reserve_bits: usize,
    rightmost: bool,
) -> Option<(Package, &'m[LayerVocab])> {
    const UNINIT: usize = usize::MAX;

    let mut in_bits_used = 0;

    let mut base_col = UNINIT;

    // The single propagate bit is suppressed for the rightmost stage.
    // Setup RAMs with all-singleton inputs cannot possibly carry,
    // but in order to benefit from eliminating the carry bit from the
    // design to use that bit elsewhere, a lot of complexity would have
    // to be added. So we'll always produce a carry output bit, even
    // when it will never be set. The leftmost RAM also doesn't need a
    // carry output bit, but it has one anyway.
    let n_prop_bits = if rightmost { 0 } else { 1 };
    let n_carry_bits = 1;

    let mut width = 0;                  // number of columns taken
    let mut input_pvs = Vec::with_capacity(g.ram_dims.0);

    while !layer.is_empty() {

        // See if there's a column available to consider.
        let (gpv, n) = match layer[0] {
            Column { gpv, n } => (gpv, n),
            _ => break,
        };

        if base_col == UNINIT {
            base_col = gpv;
        }

        // We'll need several checks relating to the size of this column
        // before we commit to taking it. See c.1 through c.4.
        let next_width = width + 1;
        let next_in_bits_used = in_bits_used + n;

        // c.1  not too many input bits on this RAM
        if next_in_bits_used > g.ram_dims.0 {
            break;
        }

        // c.2  not too many output bits on this RAM
        if next_width + n_prop_bits + n_carry_bits > g.ram_dims.1 {
            break;
        }

        // c.3  not too many input bits on finishing (south) RAM
        if next_width + reserve_bits > g.ram_dims.0 {
            break;
        }

        // c.4  not too many output bits on finishing (south) RAM
        //      (already estbalished by c.2)
        // if next_width > g.ram_dims.1 {
        //     break;
        // }

        // We commit to taking this column, and update some limits.
        layer = &layer[1..];
        width = next_width;
        in_bits_used = next_in_bits_used;
        for _ in 0..n {
            input_pvs.push(gpv);
        }

        // Non-consecutive Column inputs from the caller would weak havoc.
        assert_eq!(base_col + width, gpv + 1);
    }

    // If we didn't take a column, the RAMs are probably too small.
    // Let the caller deal with it.
    if width == 0 {
        return None;
    }

    let pkg = Package {
        kind: CarrySkipSetup {
            overshoot_complement: 0,
            finishing: false,
            n_prop_bits,
            n_carry_bits,
        },
        mem_used: width << in_bits_used,
        output_pvs: (base_col, base_col + width),
        input_pvs,
    };

    Some((pkg, layer))
}

// Post-parsing fn multilayer_scheme() will add all this to m.trial_soln:
//
// a) the layer of setup RAMs from `pkgs`
// b) `n_layers` layers of input propagation tree RAMs
// c) output propagation RAMs parallel to portions of (b)
// d) one layer final propagation RAMs
// e) one layer of CarrySkipFinish
//
// Composition optimization will typically merge layer (d) into (e)
// and merge a lot of (c) with itself after synthesis. But that's later.
fn multilayer_scheme(
    m: &mut MutGlobals,
    mut pkgs: Vec<Package>,
    layers: Vec<Vec<PropNode>>,
    joiners: Vec<Vec<PropNodeID>>,
) {
//show_layers(&layers);
//show_joiners(&joiners);

    // Before we consume pkgs, create the finish layer.
    // The first setup stage is final and doesn't have a finish stage.
    let mut finishers = Vec::with_capacity(pkgs.len());
    for setup_pkg in &pkgs[1..] {
        if let Package {
            kind: CarrySkipSetup { .. },
            output_pvs, ..
        } = setup_pkg {
            let width = output_pvs.1 - output_pvs.0;
            finishers.push(Package {
                kind: CarrySkipFinish {
                    stage_widths: vec![width as u64],
                },
                mem_used: width << 1 + width,   // +1 is carry decision input.
                output_pvs: *output_pvs,
                input_pvs: (output_pvs.0..output_pvs.1).collect(),
            });
        }
        else {
            unreachable!();
        }
    }

    // First layer: setup RAMs
    m.trial_soln.append(&mut pkgs);

    // Next layers: propagator tree, input side
    let mut skip_first_layer = true;
    for layer in layers {
        if skip_first_layer {
            skip_first_layer = false;
            continue;
        }
        m.add_layer_sep();
        for PropNode { inputs, rightmost } in layer {
            let in_bits = 2 * inputs.len() - rightmost as usize;
            let out_bits = (2 - rightmost as usize) * (inputs.len() - 1);
            m.trial_soln.push(Package {
                kind: Propagator { inputs, rightmost },
                mem_used: out_bits << in_bits,
                output_pvs: (0, 0),
                input_pvs: Vec::new(),
            });
        }
    }

    // This separator precedes either joiners or the finish RAMs.
    m.add_layer_sep();

    let mut stage = 0;
    for inputs in joiners {
        if inputs.is_empty() {
            if stage == 0 {
//              p!(G "stage 0 is final, so no joiners");
            } else {
                p!(G "why empty? stage {}", stage);
            }
            continue;
        }

        // This is a coarse estimate of firmware memory needed,
        // because synth is going to break this into multiple
        // packages, and then composition is going to fuse them
        // into fewer packages.
        let in_bits = 2 * inputs.len() - 1;
        let out_bits = 1;

        stage += 1;
        m.trial_soln.push(Package {
            kind: Joiner { stage, inputs },
            mem_used: out_bits << in_bits,
            output_pvs: (0, 0),
            input_pvs: Vec::new(),
        });
    }

    // Separate the joiners (if any) from the finish RAMs.
    if stage != 0 {
        m.add_layer_sep();
    }

    // Final layer: finish RAMs
    // XXX This PackageKind::CarrySkipFinish layer is used by
    // draw_solution() for graphics, but it's disregarded by
    // synthesize_layer() for netlisting. The RAMParms::CarrySkipFinish
    // layer that is synthesized with the root nodes of the corresponding
    // Joiner trees. This can't be great software engineering.
    m.trial_soln.append(&mut finishers);

    // Caller supplies final EndOfLayer because of how it's written.

}

// `layers` likely points to packages that don't exist.
// Trim up accordingly.
fn trim_tree(
    layers: &mut Vec<Vec<PropNode>>,
    joiners: &mut Vec<Vec<PropNodeID>>,
    n_pkgs: usize,
) {
    // Layer 0 indicates leaves, one per CarrySkipSetup.
    // Only keep the ones which exist.
    layers[0].truncate(n_pkgs);

    // Subsequent layers are interior nodes, which represent Propagators.
    // We need to process each layer in order.
    // Each propnode only has inputs from the layer prior.
    for layer_ix in 1..layers.len() {

        // Here is a Rust challenge: I have this item `layers`
        // that I make changes to in certain places, based on what's
        // happening in certain other places. That means multiple
        // borrowing with at least one mutable borrow. Not legal in
        // Rust. One workaround is to use a lot of scalar array
        // indices, which solves the issue at the expense of much
        // harder-to-read code, likely longer runtimes, and circumvents
        // Rust's memory assurances. Here's my workaround instead:
        // the changes needed here never reference the layer the changes
        // happen in. Here, I temporarily remove the layer to be
        // modified (without disturbing the array order in `layers`),
        // allowing me whatever mutable reference I need with immutable
        // references outside that layer still permitted. This removal
        // and eventual replacement take O(1) time.
        let mut layer = std::mem::take(&mut layers[layer_ix]);

        // Loop over this layer's propnodes. Order does not matter.
        for p in layer.iter_mut() {

            // Retain only those inputs for which
            // (1) the source RAM still exists, and
            // (2) the identified source RAM output still exists.
            p.inputs.retain(
                |(layer_ix, stage_ix, inst_ix)|
                *stage_ix < layers[*layer_ix].len() &&
                (*layer_ix == 0 ||
                *inst_ix + 1 < layers[*layer_ix][*stage_ix].inputs.len())
            );
        }

        // Retain only propnodes that still have outputs.
        layer.retain(|p| p.inputs.len() > 1);

        // Put the edited layer back where it goes.
        layers[layer_ix] = layer;
    }

    // Remove empty layers that may exist toward the root.
    layers.retain(|l| !l.is_empty());    

    // Remove the joiners that correspond to non-existent CarrySkipSetups.
    joiners.truncate(n_pkgs);
}
