// included from solver.rs

#[derive(Clone, Copy, Debug, PartialEq)]
pub enum OptimizeStrategy {
    OptimizeNone,       // all ArbGeom
    OptimizeMulti,      // ArbGeom + multilayer carry-skip
    OptimizePackage,    // consider one package at a time
    OptimizeLayer,      // consider one layer at a time
    OptimizeAll,        // consider all summation layers at once
}
pub use OptimizeStrategy::*;

/// This looks at a left portion of a layer to parse, and
/// determines if it would generate anything other than Wireflop.
/// Otherwise, there is "no work left," and there are implications.
fn no_work_left(layer: &[LayerVocab]) -> bool {
    for pkg in layer {
        if let Column { n: 1, .. } = *pkg {
            continue;
        }
        return false;
    }
    true
}

// To reduce needless allocations, this returns None instead of [].
fn parse_carry_skip_setup<'m>(
    g: &Globals,
    m: &mut MutGlobals,
    ms: Mindset,
    mut layer: &'m[LayerVocab],
) -> Option<Vec<SetupFlavor<'m>>> {
    const UNINIT: usize = usize::MAX;

    // Never set up for carry skip after Wireflop or Arbgeom,
    // because complete incoming carry information does not exist.
    // Don't worry, we now allow all-singleton CarrySkipSetups
    // to be parsed in the presence of baggage!
    // XXX This test might be fragile if more PackageKind are added.
    match &m.trial_soln[..] {
        [.., Package { kind, .. }] => {
            if let ArbGeom { .. } = kind {
                return None;
            }
            if let Wireflop = kind {
                return None;
            }
        },
        _ => (),
    }

    // If carry-skip finish information was delivered here from the
    // prior layer, it can reduce the number of input columns, but also
    // contributes carry and propagate bits that can increase the sum.
    // Overall it's usually a net decrease. We account for these with
    // these new variables:
    //
    //   finishing       true / false
    //   max_setup_sum   computed max sum this package can produce
    //   in_bits_used    keep package from running out of input wires
    //
    let (finishing, mut in_bits_used, mut max_setup_sum) = if let [ FinishInfo
            { prop_carry_baggage, max_eff_carry_val }, .. ] = layer {
        layer = &layer[1..];
        (true, *prop_carry_baggage, *max_eff_carry_val)
    } else {
        (false, 0, 0)
    };

    let mut outs: Vec<SetupFlavor> = Vec::new();
    let b = ms.baggage();               // propagate-carry info from east

    let mut base_col = UNINIT;
    let n_prop_bits = b.max_eff_carry_val.bit_length();

    let mut width = 0;                  // number of columns taken
    let mut input_pvs = Vec::with_capacity(g.ram_dims.0);

    // overshoot_gpv indicates the most recent gpv seen for which
    // a complement must be added.
    let mut overshoot_gpv = usize::MAX;
    let mut overshoot_complement = 0;

    while !layer.is_empty() {

        // For carry-skip setup, OffsetComplement always precedes
        // a non-empty Column.
        if let OffsetComplement { gpv } = layer[0] {
            overshoot_gpv = gpv;
            layer = &layer[1..];
            continue;
        }

        // See if there's a column available to consider.
        let (gpv, n) = match layer[0] {
            Column { gpv, n } => (gpv, n),
            _ => break,
        };

        if base_col == UNINIT {
            base_col = gpv;
        }

        // If we have unremoved overshoot complement at this
        // column, force its removal in the proposed column.
        // Maybe not globally optimal, but typically reasonable.
        let remove = if overshoot_gpv == gpv {
            1 << gpv - base_col
        } else { 0 };

        // We'll need several checks relating to the size of this column
        // before we commit to taking it. See c.1 through c.5.
        let next_width = width + 1;
        let addend = n << gpv - base_col;
        let next_max_setup_sum = max_setup_sum + addend + remove;
        let next_in_bits_used = in_bits_used + n;

        // c.1  not too many input bits on this RAM
        if next_in_bits_used > g.ram_dims.0 {
            break;
        }

        // c.2  not too many output bits on this RAM
        let n_full_sum_bits = next_max_setup_sum.bit_length();
        let out_bits_used = n_prop_bits + n_full_sum_bits;
        if out_bits_used > g.ram_dims.1 {
            break;
        }

        // c.3  not too many input bits on finishing (southwest) RAM
        let n_carry_bits = n_full_sum_bits - next_width;
        if n_prop_bits + n_carry_bits + b.n_prop_carry_cumul > g.ram_dims.0 {
            break;
        }

        // c.4  not too many output bits on finishing (southwest) RAM
        //      (unlikely scenario)
        let max_finish_sum = next_max_setup_sum + b.max_eff_carry_val;
        let max_eff_carry_val = max_finish_sum >> next_width;
        if max_eff_carry_val.bit_length() > g.ram_dims.1 {
            polite_panic!("You gotta see this.");
            break;                      // if needed, this break is valid
        }

        // c.5  not too many output bits on finishing (south) RAM
        // XXX  (south and southwest finishing RAMs could use distinct terms)

        // c.5 is always satisfied, so long as the setup and finishing
        // ram_dims are the same, because the finishing RAM output width
        // is the same, but the finishing RAM doesn't also have prop &
        // carry outputs. So no test appears here.

        // c.6  not too many input bits on finishing (south) RAM
        //      (Learned the hard way this isn't automatically true.)
        if next_width + b.n_prop_carry_cumul > g.ram_dims.0 {
            break;
        }

        // We commit to taking this column, and update some limits.
        layer = &layer[1..];
        max_setup_sum = next_max_setup_sum;
        width = next_width;
        in_bits_used = next_in_bits_used;
        overshoot_complement += remove;
        for _ in 0..n {
            input_pvs.push(gpv);
        }

        // Although we've now taken this column, it's possible we
        // haven't taken enough columns to avoid trying to propagate more
        // than one carry through this package, which we don't support.
        // If so, we keep our updates, don't output a proposed package,
        // and try to get some more columns.
        if width < n_prop_bits {
            continue;
        }

        // Another problem that comes up is this might be a "leftmost"
        // CarrySkipSetup; that is, there won't be any more packages
        // left of this one. In that situation, what we want is ArbGeom,
        // not CarrySkipSetup, because the ArbGeom will solve us in
        // one layer fewer.
        if no_work_left(layer) {
            // If indeed this is a "leftmost" CarrySkipSetup, any
            // narrower configurations which may have been generated
            // to this point should also be eliminated, as we'll use
            // an ArbGeom instead. To kill these other configurations,
            // we return instead of continue.
            return None;
        }

        // The columns we've taken comprise a valid CarrySkipSetup,
        // but we only desire to try the widest ones we find for
        // any particular prop-carry baggage. Because usually we just
        // want a greedy outcome (as many columns as we can), but once
        // in a while we can get a better outcome if we lighten the
        // baggage we hand to the right. (This flexibility is at the
        // cost of a lot of CPU time.)
        //
        // So here, we determine where (if anywhere) among the output
        // cases we have if there is a prior solution with the same
        // baggage that we can replace. We don't have to compare widths
        // first, because `width` is increasing monotonically.
        let n_prop_carry_cumul =
                b.n_prop_carry_cumul + n_prop_bits + n_carry_bits;
        let mut index = outs.len();
        for (i, (bag, ..)) in outs.iter().enumerate() {
            if bag.n_prop_carry_cumul == n_prop_carry_cumul {
                index = i;
                break;
            }
        }

        let b_out = BaggageStruct {
            n_carry_in: n_carry_bits,
            n_prop_carry_cumul,
            max_eff_carry_val,
        };

        assert_eq!(base_col + width, gpv + 1);

        let p_out = Package {
            kind: CarrySkipSetup {
                overshoot_complement: overshoot_complement as u64,
                finishing, n_prop_bits, n_carry_bits
            },
            mem_used: out_bits_used << in_bits_used,
            output_pvs: (base_col, gpv + 1),
            input_pvs: input_pvs.clone(),
        };

        let mut next_layer = Vec::new();

        // This FinishWithWidth uses FinishInfo from a PREVIOUS
        // setup package (immediate east).
        if b.n_prop_carry_cumul != 0 {
            next_layer.push(FinishWithWidth {
                gpv: base_col,
                width,
            });
        }

        // This FinishInfo will be delivered with addend output
        // from something LATER (immediate west), whether
        // a CarrySkipSetup, ArbGeom, Wireflop, etc.
        next_layer.push(FinishInfo {
            prop_carry_baggage: n_prop_carry_cumul,
            max_eff_carry_val,
        });

        let out_tup = (b_out, layer, p_out, next_layer);

        if index == outs.len() {
            outs.push(out_tup);
        }
        else {
            outs[index] = out_tup;
        }
    }

    if outs.is_empty() { None } else { Some(outs) }
}
