///! propagator.rs June 2022
///! This module relates to multilayer carry-skip for large multipliers.

use crate::*;

#[derive(Debug)]
pub struct PropNode {
    pub inputs: Vec<PropNodeID>,
    pub rightmost: bool,
}

// layer
// stage within layer
// output position within stage
pub type PropNodeID = (usize, usize, usize);

// Compute fanout of "input half" of multilayer carry-skip internal nodes.
// Fanout is the number of inputs to a node, and is maximized for the size
// of the RAM. The number of outputs is one less than the fanout. The
// rightmost nodes lack the propagate bit on their rightmost inputs and all
// of their outputs, because they have no incoming carry that can be
// propagated. XXX Tested manually 20 June 2022.
fn calc_fanout(dims: &(usize, usize), rightmost: bool) -> usize {
    let rightmost = rightmost as usize;
    let tentative_max_in = (dims.0 + rightmost) / 2;
    let tentative_max_out = dims.1 / (2 - rightmost);
    tentative_max_in.min(1 + tentative_max_out)
}

// Layer 0 is leaves (carry setup RAMs) only.
// n_layers is the number of propagator RAMs, and does not count layer 0.
// XXX Test manually 21 June 2022.
pub fn make_propagator_layers(n_layers: usize, dims: &(usize, usize))
        -> Vec<Vec<PropNode>> {
    let mut layers: Vec<Vec<PropNode>> = Vec::new();

    // Determine normal and rightmost fanout.
    let n_fanout = calc_fanout(dims, false);
    let r_fanout = calc_fanout(dims, true);

    // Create empty layers.
    for _ in 0..=n_layers {
        layers.push(Vec::new());
    }

    // Root layer has just one item.
    layers[n_layers].push(PropNode { inputs: Vec::new(), rightmost: true });

    // TREE EXPANSION SECTION

    // `child_layer_ix` is an index into `layers` where children are added.
    for child_layer_ix in (0..n_layers).rev() {

        let parent_layer_ix = 1 + child_layer_ix;
        let n_parents = layers[parent_layer_ix].len();
        for parent_ix in 0..n_parents {

            // The number of new children varies with parent fanout.
            let p_fanout = if parent_ix == 0 { r_fanout } else { n_fanout };
            for _ in 0..p_fanout {

                // Will this be a rightmost child?
                let c_rightmost = layers[child_layer_ix].is_empty();

                // Location among the child's output will vary.
                let c_out_ix = if child_layer_ix == 0 {
                    0                   // Setup RAM has just 1 p/g output.
                } else if c_rightmost {
                    r_fanout - 2        // leftmost output ix on rightmost
                } else {
                    n_fanout - 2        // leftmost output ix on non-rightmost
                };

                // `child` is a new child.
                let child = PropNode {
                    inputs: Vec::new(),
                    rightmost: c_rightmost,
                };

                // `node_id` finds the child's generate & propagate signals.
                let node_id = (
                    child_layer_ix,
                    layers[child_layer_ix].len(),
                    c_out_ix,
                );

                // Connect the parent to the child.
                layers[parent_layer_ix][parent_ix].inputs.push(node_id);

                // Include the child with its generation.
                layers[child_layer_ix].push(child);
            }
        }
    }

    // The rightmost leaf has no propagate or carry input.
    // We can also add one final leaf on the left, that accepts the
    // carry decision for everything we've done above.
    layers[0].push(PropNode { inputs: Vec::new(), rightmost: false });

    layers
}

// Implement the output half algorithm to form the forest that combines
// the parital carry-skip information. This is done on all leaves.
pub fn trace_propagator_joiners(layers: &Vec<Vec<PropNode>>)
        -> Vec<Vec<PropNodeID>> {

    // The joiner for the rightmost leaf is the empty set,
    // because that setup RAM is final and has no prop or carry input.
    let mut joiners = vec![Vec::new()];

    // Layer 0 was extended by one RAM past what the input half tree
    // can reach. So we have one less iteration to do here. Because
    // we've inserted an empty set on the previous line, the carry
    // decisions are properly shifted to the columns where the finish
    // RAMs will appear.
    for leaf_ix in 0..layers[0].len() - 1 {
        joiners.push(trace_joiner(layers, (0, leaf_ix, 0)));
    }
    joiners
}

// Helper for trace_joiner() to locate where a PropNodeID is an input.
fn find_id(layers: &Vec<Vec<PropNode>>, id: &PropNodeID) -> Option<PropNodeID> {
    let layer_ix = id.0 + 1;
    if layer_ix < layers.len() {
        for (ram_ix, ram) in layers[layer_ix].iter().enumerate() {
            for (input_ix, input_id) in ram.inputs.iter().enumerate() {
                if *input_id == *id {
                    return Some((layer_ix, ram_ix, input_ix));
                }
            }
        }
    }
    None
}

// Implement the output half algorithm for a single leaf.
// The output is a list/tree of the nodes to compose to form the
// decision for finish RAM for the next leaf to the left.
//
// I'll rephrase. If layer 0 has N leaves, the tree accommodates up
// to N + 1 setup RAMs S0..SN and N finish RAMs F1..FN. There is no
// F0, because S0's output is final and nothing will carry or propagate
// into F0. The N joiner lists produced by this routine JL0..JLN-1
// produce carry decisions for F1..FN, so there is a left shift of
// one stage. This works correctly all the way down to the 0-layer case,
// where S0's carry can feed F1 directly.
//
// XXX Manually tested 3 layers, fanout 3, 27 setups, 21 June 2022.
fn trace_joiner(layers: &Vec<Vec<PropNode>>, mut id: PropNodeID)
        -> Vec<PropNodeID> {

    // Repeatedly locate the layer, RAM, and input where `id` is an input.
    while let Some(lri) = find_id(layers, &id) {
        // If this is a rightmost input, just keep `id`.
        if lri.2 == 0 {
            break;
        }
        // Otherwise, `id` to become the output corresponding to this input.
        id = (lri.0, lri.1, lri.2 - 1);
    }

    // If the RAM driving `id` has a right neighbor, recurse on that
    // neighbor's leftmost output. Otherwise provide an empty list.
    let mut list = if id.1 > 0 {
        let l_out_ix = if id.0 == 0 {
            0                           // leaf
        } else { layers[id.0][id.1 - 1].inputs.len() - 2 };
        let l_output = (id.0, id.1 - 1, l_out_ix);
        trace_joiner(layers, l_output)
    } else {
        Vec::new()
    };

    // Join `id` and `list` into a single list. We're done.
    list.push(id);          // new 13 July
//  list.insert(0, id);     // removed 13 July
    list
}

pub fn test_gator() {
    let layers = make_propagator_layers(1, &(18, 18));
    show_layers(&layers);
    p!();

    let joiners = trace_propagator_joiners(&layers);
    show_joiners(&joiners);
    p!();
}

// diagnostic
pub fn show_layers(layers: &Vec<Vec<PropNode>>) {
    for (i, layer) in layers.iter().enumerate() {
        p!(Y "Layer {}", i);
        for ram in layer {
            pp!("  RAM{}", if ram.rightmost { ">" } else { " " });
            for input in &ram.inputs {
                pp!(" {}.{}.{}", input.0, input.1, input.2);
            }
            p!();
        }
    }
}

// diagnostic
pub fn show_joiners(cols: &Vec<Vec<PropNodeID>>) {
    for (i, list) in cols.iter().enumerate() {
        pp!(Y "Joiner {}:", i);
        for lri in list {
            pp!(" {}.{}.{}", lri.0, lri.1, lri.2);
        }
        p!();
    }
}


