///! ram.rs February 2022
///! This module models RAMs and the functions they implement.

use crate::*;

type RAMFn = fn(u64, &RAMParms) -> u64;

/// `RAM` represents a RAM within a multiplier. We record its size,
/// where signals come from and go to, and a deterministic computation
/// to represent its contents. (Note that RAM contents never change
/// after the firmware is initialized. We are using RAM as fast ROM.)
#[derive(Clone, Debug)]
pub struct RAM {

    /// Here are the number of physical input & output bits for this RAM,
    /// whether or not they are actually wired.
    pub dims: (usize, usize),

    /// Here are indices into Wires.bits where bools come from during
    /// computation. The size of this Vec is the number of inputs
    /// actually wired. The bits are used in order, beginning with the
    /// least significant. Leftover bits would be connected to ground
    /// in an actual circuit.
    pub inputs: Vec<usize>,

    /// Here are indices into Wires.bits where bools go to during
    /// computation. The size of this Vec is the number of outputs
    /// actually wired.* The bits are used in order, beginning with the
    /// least significant. Leftover bits would be pulled to ground via
    /// resistors in an actual circuit.
    ///
    /// *For RAMParms::Subproduct, product_width will match outputs.len() here.
    pub outputs: Vec<usize>,

    /// Function `func` is used to generate this RAM's contents, or to
    /// simulate the operation of this RAM. Parameters needed to
    /// customize `func` within a circuit are available in `func_parms`.
    pub func: Funky<RAMFn>,
    pub func_parms: RAMParms,

    /// `marker` is scratch memory for use by any function.
    /// Actual usage is (i) to save the base column for CarrySkipSetup
    /// for use in multi-layer schemes, and later (ii) during composition
    /// to identify RAMs.
    pub marker: usize,
}

/// Rust makes it hard to implement or derive Debug on items that
/// contain functions. To fix this, here's a Funky struct to wrap
/// functions with a Debug implementation.
#[derive(Clone, Copy)]
//pub struct Funky(pub fn(u64, &RAMParms) -> u64);
pub struct Funky<F>(pub F);

/// `RAMParms` are parameters used to generate a RAM's contents.
/// There are a few kinds, depending on what a RAM has to do.
#[derive(Clone, Debug)]
pub enum RAMParms {

    Subproduct {
        r_val_mask: i64,        // mask of bits where right operand is
        r_msb_mask: i64,        // most significant bit of right operand
        r_sue_mask: i64,        // see DISCUSSION
        l_val_mask: i64,        // mask of bits where left operand is
        l_msb_mask: i64,        // most significant bit of left operand
        l_sue_mask: i64,        // see DISCUSSION
        l_val_shift: usize,     // right shift that aligns left operand
        width: usize,           // number of output bits actually used
        multiplier_width: usize,// number of output bits for whole mult'r
        was_widened: bool,      // RAM was widened 1 bit to complement offset
        offset_used: i64,       // amount added to product to force positive
        offset_unused: i64,     // reserve capacity to overcome complement
        global_pv: usize,       // product LSB's global place value

        // DISCUSSION - r_sue_mask depends on signage:
        //
        // Unsigned => all zeros
        // Signed => r_val_mask
        // Either => r_ctrl_decoder_signed_mask
        //
        // r_sue_mask & ram_input != 0 indicates that the right operand's
        // MSB has a negative weight. This test has a false negative if
        // the right operand is zero, but the MSB is clear anyhow in
        // that case.
        //
        // l_sue_mask works isomorphically to r_sue_mask.
    },

    ArbGeom {
        // Simply, know what is added for each true input bit.
        // XXX Positions in this Vec must exactly match positions in the
        //     RAM's input bits, because larger positions are reserved
        //     for carry and propagate inputs for finishing ArbGeom.
        input_to_addend: Vec<u64>,

        // This provides means after layer 0 for complementing overshoot.
        overshoot_complement: u64,

        // Extension for finishing ArbGeom:
        // See compute_skipped_carry() for meaning of this Vec.
        finish_srcs: Vec<(u64, u64)>,

        // Extension for non-rightmost (Finshing)CarrySkipSetup.
        // This causes propagate bits to be output. The extension is
        // disabled if prop_max is 0, which must be the case for
        // ArbGeom, and any carry-skip setups that
        // are rightmost and therefore don't output propagate bits.
        carry_lsb: u64,         // local pv of carry output's LSB
        prop_lsb: u64,          // local pv of prop output's LSB
        prop_max: u64,          // max prop value out == MECV from east
    },

    CarrySkipFinish {
        // See compute_skipped_carry() for meaning of this Vec.
        finish_srcs: Vec<(u64, u64)>,

        // This is the number of finish stages east of here,
        // because we need all of their sources too.
        n_finishes_so_far: usize,

        // Here are widths for each subpackage from right to left.
        stage_widths: Vec<u64>,
    },

    Bogus {                     // not an actual RAM; can be a pointer marker
    },

    VerticalCompose {           // composition: `top` RAM feeds into `bot` RAM
        top_parms: Box<RAMParms>,
        bot_parms: Box<RAMParms>,
        top_func: Funky<RAMFn>,
        bot_func: Funky<RAMFn>,
        top_mask: u64,
        top_to_bot_shuf: Vec<(u32, u32)>,
        pkg_to_bot_shuf: Vec<(u32, u32)>,
    },

    HorizontalCompose {         // composition: `l` and `r` RAMs simultaneous
        l_parms: Box<RAMParms>,
        r_parms: Box<RAMParms>,
        l_func: Funky<RAMFn>,
        r_func: Funky<RAMFn>,
        l_out_shift: u32,
        r_in_mask: u64,
        l_out_mask: u64,
        r_out_mask: u64,
        in_to_l_shuf: Vec<(u32, u32)>,
    },

    Propagator {
        n_stages: usize,
        rightmost: bool,
    },

}

impl RAMParms {

    /// Create parameters for a RAM that implements subproducts.
    /// These are intermediate values that allow subproduct_ram_fn()
    /// to operate correctly for a particular RAM.
    pub fn subproduct_from(
            width_l: usize, width_r: usize,
            sue_l: Signage, sue_r: Signage,
            global_pv: usize, multiplier_width: usize,
        ) -> RAMParms {

        let r_val_mask = (1i64 << width_r) - 1;
        let r_msb_mask = 1i64 << width_r - 1;

        let l_val_shift = width_r;
        let l_val_mask = ((1i64 << width_l) - 1) << l_val_shift;
        let l_msb_mask = (1i64 << width_l - 1) << l_val_shift;

        let mets = subproduct_metrics(width_l, width_r, sue_l, sue_r);

        // XXX Note that mets.width is not always width_l + width_r.
        //     It can be more or less, depending on such needs as
        //     additive offset (larger), 1-bit factors (smaller), etc.

        // XXX Sometime in March 2022, I was printing out Signed * Signed
        //     cases where mets.width != width_l + width_r, flagging them
        //     as "trouble". I don't know why, and I've removed this for
        //     the time being.

        let r_sue_mask = match sue_r {
            Unsigned => 0,
            Signed => r_val_mask,
            Either => 1 << width_l + width_r,
        };

        let l_sue_mask = match sue_l {
            Unsigned => 0,
            Signed => l_val_mask,
            Either => 1 << width_l + width_r + (sue_r == Either) as usize,
        };

        let offset_unused = if global_pv + mets.width >= multiplier_width {
            i64::MAX        // consider infinite; uses saturating_add()
        } else {
            (1 << mets.width) - mets.max_prod - 1
        };

        RAMParms::Subproduct {
            r_val_mask, r_msb_mask, r_sue_mask,
            l_val_mask, l_msb_mask, l_sue_mask, l_val_shift,
            width: mets.width, global_pv, was_widened: false,
            offset_used: mets.offset, offset_unused, multiplier_width,
        }
    }
}

impl RAM {

    /// Create a new RAM, not connected to any wiring yet.
    /// `func` determines the RAM's contents at the time of creation.
    /// This constructor consumes `parms`.
    /// XXX This function isn't always convenient to use.
    pub fn new(
        dims: (usize, usize),
        func: Funky<RAMFn>,
        parms: RAMParms,
    ) -> RAM {
        RAM {
            dims,
            inputs: Vec::new(),
            outputs: Vec::new(),
            func,
            func_parms: parms,
            marker: 0,
        }
    }

    /// Create a safe pointer that's like a RAM, but doesn't do anything.
    /// This can be used as a marker among RAMs, etc.
    pub fn bogus() -> RAM {
        RAM {
            dims: (0, 0),
            inputs: Vec::new(),
            outputs: Vec::new(),
            func: Funky(bogus_ram_fn),
            func_parms: RAMParms::Bogus {},
            marker: 0,
        }
    }

    /// This is a convenience test to see if a &RAM is a marker pointer instead.
    pub fn is_bogus(&self) -> bool {
        if let RAMParms::Bogus {} = self.func_parms { true } else { false }
    }

    /// Simulate the operation of a RAM.
    /// `bits` is the flat vector of all circuit nodes within the multiplier,
    /// of which this RAM uses only those to which it connects.
    pub fn simulate(&self, bits: &mut [bool]) {
        let mut ram_in = 0;

        // loop over input word place values
        for pv in 0..self.inputs.len() {
            let input = self.inputs[pv];
            if bits[input] {
                ram_in |= 1 << pv;
            }
        }

        // do the computation
        let ram_out = (self.func.0)(ram_in, &self.func_parms);

        // loop over output word place values
        for pv in 0..self.outputs.len() {
            bits[self.outputs[pv]] = ram_out & 1 << pv != 0;
        }
    }

    /// This special function is only for wired subproduct RAMs.
    /// Given global place value `pv` representing a desired overshoot
    /// complement bit, see if this RAM has leftover
    /// offset capacity to satisfy the request. Iff so, adjust its
    /// RAMParms accordingly and return true.
    ///
    /// If `widen_first` is true, we've previously called this function
    /// without success for this place value on all RAMs, and our now
    /// strategy is to see if a RAM can be widened by one output bit
    /// to achieve the offset.
    ///
    /// Returns two bools: (offset was removed, RAM just now widened).
    ///
    pub fn try_to_offset(&mut self, pv: usize, widen_first: bool)
            -> (bool, bool) {

        if let RAMParms::Subproduct {
                width, offset_used, offset_unused, global_pv, was_widened,
                multiplier_width, .. } = &mut self.func_parms {

            // Widening is contingent on having a free output bit
            // available for use.
            let widen_first = widen_first && !*was_widened
                    && *width < self.dims.1;

            // For counting place value eligiblity, we need widen_first
            // expressed as 1 or 0.
            let extra = widen_first as usize;

            // See if the requested place value is within the
            // place values this RAM can represent.
            if pv < *global_pv || pv >= *global_pv + *width + extra {
                return (false, false);      // No.
            }

            // Here is where we would widen this RAM if necessary.
            if widen_first {
                let more_offset = 1 << *width;
                *width += 1;
                *offset_unused = (*offset_unused).saturating_add(more_offset);
                if *global_pv + *width >= *multiplier_width {
                    *offset_unused = i64::MAX;
                }
                *was_widened = true;

                // The RAM now thinks it is widened, but the caller now
                // needs to connect the new output bit into the circuit.
            }

            // See if there is enough juice (sufficient unused offset).
            let juice = 1 << pv - *global_pv;
            if juice > *offset_unused {
                assert!(!widen_first);
                return (false, false);
            }

            // Adjust as requested and conclude.
            *offset_used += juice;
            *offset_unused -= juice;
            (true, widen_first)
        }
        else {
            fail("try_to_offset found strange parms");
        }
    }

    // Form a RAM that is a vertical composition of two small RAMs.
    pub fn compose_top_bot(&self, bot: &Self) -> Self {
        let top = self;

        // We keep the inputs to the top RAM as they are.
        // To these, we will extend the composed RAM inputs at 
        // higher indices to serve as bottom inputs.
        let mut inputs = top.inputs.clone();
        let top_mask = (1 << inputs.len()) - 1;

        // These "shuffle" array elements will have form (from_pv, to_pv).
        let mut top_to_bot_shuf = Vec::new();
        let mut pkg_to_bot_shuf = Vec::new();

        // For each input to the bottom RAM, find, attach, and sometimes
        // create its source.
        'next_bot_input:
        for (bot_pv, bot_bit_pos) in bot.inputs.iter().enumerate() {

            // Look for this bit_pos among the top RAM outputs.
            for (top_pv, top_bit_pos) in top.outputs.iter().enumerate() {
                if top_bit_pos == bot_bit_pos {
                    top_to_bot_shuf.push((top_pv as u32, bot_pv as u32));
                    continue 'next_bot_input;
                }
            }

            // This bit_pos did not appear among the top RAM outputs,
            // so we'll allocate a new composed RAM input for it.
            let pkg_pv = inputs.len();
            inputs.push(*bot_bit_pos);
            pkg_to_bot_shuf.push((pkg_pv as u32, bot_pv as u32));
        }

        // The composed RAM's outputs are exactly the outputs of
        // the bottom RAM. The composed functions and parameters
        // are simply copies from the two original RAMs.
        let func_parms = RAMParms::VerticalCompose {
            top_parms: Box::new(top.func_parms.clone()),
            bot_parms: Box::new(bot.func_parms.clone()),
            top_func: top.func,
            bot_func: bot.func,
            top_mask,
            top_to_bot_shuf,
            pkg_to_bot_shuf,
        };

        // The composed RAM's `marker` must not match any live markers.
        // I've chosen usize::MAX for this.
        RAM {
            dims: top.dims,
            inputs,
            outputs: bot.outputs.clone(),
            func: Funky(vertical_compose_ram_fn),
            func_parms,
            marker: usize::MAX,
        }
    }

    // Form a RAM that is a horizontal composition of two RAMs.
    pub fn compose_l_r(&self, r: &Self) -> Self {
        let l = self;

        // We keep the inputs to the right RAM as they are.
        let mut inputs = r.inputs.clone();
        let r_in_mask = (1 << inputs.len()) - 1;

        // These "shuffle" array elements will have form (from_pv, to_pv).
        let mut in_to_l_shuf = Vec::new();

        // For each input to the left RAM, find, attach, and sometimes
        // create its source.
        'next_l_input:
        for (l_pv, l_bit_pos) in l.inputs.iter().enumerate() {

            // Look for this bit_pos among the right RAM inputs.
            for (r_pv, r_bit_pos) in r.inputs.iter().enumerate() {
                if l_bit_pos == r_bit_pos {
                    in_to_l_shuf.push((r_pv as u32, l_pv as u32));
                    continue 'next_l_input;
                }
            }

            // This bit_pos did not appear among the right RAM inputs,
            // so we'll allocate a new composed RAM input for it.
            let pkg_pv = inputs.len();
            inputs.push(*l_bit_pos);
            in_to_l_shuf.push((pkg_pv as u32, l_pv as u32));
        }

        // The composed RAM's outputs are the catenated outputs of the
        // right and left RAMs. The composing functions and parameters
        // are simply copies from the two original RAMs.
        let func_parms = RAMParms::HorizontalCompose {
            l_parms: Box::new(l.func_parms.clone()),
            r_parms: Box::new(r.func_parms.clone()),
            l_func: l.func,
            r_func: r.func,
            l_out_shift: r.outputs.len() as u32,
            r_in_mask,
            l_out_mask: (1 << l.outputs.len()) - 1,
            r_out_mask: (1 << r.outputs.len()) - 1,
            in_to_l_shuf,
        };

        // The composed RAM's `marker` must not match any live markers.
        // I've chosen usize::MAX for this.
        let mut outputs = r.outputs.clone();
        outputs.append(&mut l.outputs.clone());
        RAM {
            dims: r.dims,
            inputs,
            outputs,
            func: Funky(horizontal_compose_ram_fn),
            func_parms,
            marker: usize::MAX,
        }
    }
}

/// Allow struct with Funky fields to (almost fully) derive Debug.
impl<F> std::fmt::Debug for Funky<F> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        #![allow(unused_must_use)]
        write!(f, "funky")
    }
}

/// Do a subproduct multiplication as a RAM. This can be used
/// for simulation and testing, or for firmware generation.
pub fn subproduct_ram_fn(i: u64, p: &RAMParms) -> u64 {
    if let RAMParms::Subproduct {
        r_val_mask,
        r_msb_mask,
        r_sue_mask,
        l_val_mask,
        l_msb_mask,
        l_sue_mask,
        l_val_shift,
        offset_used,
        ..
    } = *p {

        // Compute right factor, possibly signed, as i64.
        let i = i as i64;
        let mut r = i & r_val_mask;
        if i & r_sue_mask != 0 && i & r_msb_mask != 0 {
        // [--- is signed ---]    [-- is negative --]
            r -= 2 * r_msb_mask;
        }

        // Compute left factor, possibly signed, as i64.
        let mut l = i & l_val_mask;
        if i & l_sue_mask != 0 && i & l_msb_mask != 0 {
        // [--- is signed ---]    [-- is negative --]
            l -= 2 * l_msb_mask;
        }
        l >>= l_val_shift;

        // Compute the signed product.
        let lr = l * r;

        // Exaggerate the product so it's never negative.
        let lro = lr + offset_used;

        // XXX This should be way under 64 bits and convert losslessly,
        // but planetary-scale RAMs may violate this assumption. I have
        // not included a test to catch this.
        lro as u64
    }
    else {
        panic!("subproduct_ram_fn wrong parameters");
    }
}

/// Do an ArbGeom addition as a RAM. This can be used
/// for simulation and testing, or for firmware generation.
///
/// TODO Document input bit format, which now is
///      ccc ppp ccc ppp ccc xxx xxx xxx ...
/// The x's are addend bits, subject to configured place values for each.
/// The ccc and ppp fields vary in width and are propagate and carry
/// inputs. Least significant carry field is rightmost, by the leftmost xs.
/// So this implements ArbGeom.
pub fn arbgeom_ram_fn(i: u64, p: &RAMParms) -> u64 {
    if let RAMParms::ArbGeom {
            input_to_addend,
            overshoot_complement,
            finish_srcs,
            carry_lsb,
            prop_lsb,
            prop_max,
        } = p {

        // Do any carry-skip finish that is pending from northeast.
        // Also complement overshoot as directed.
        let mut sum = overshoot_complement +
                compute_skipped_carry(i, &finish_srcs);

        // Add the input bits in their correct place values.
        for (pv, addend) in input_to_addend.iter().enumerate() {
            if i & 1 << pv != 0 {
                sum += addend;
            }
        }

        // For (Finishing)CarrySkipSetup only, fill in propagate bit(s).
        if *prop_max != 0 {
            let carry_at = 1 << *carry_lsb;

            // This let forces prop_thresh > 0. If we don't or in any
            // ones, this RAM won't propagate carries, which is the
            // most common case.
            let prop_thresh = carry_at - (sum & (carry_at - 1));
            if prop_thresh <= *prop_max {
                sum |= prop_thresh << *prop_lsb;
            }
        }

        sum
    }
    else {
        panic!("arbgeom_ram_fn wrong parameters");
    }
}

/// Do a CarrySkipFinish addition as a RAM. This can be used
/// for simulation and testing, or for firmware generation.
/// The input bit format is the same as for ArbGeom, except
/// the non-incremented summand is a flat row. The propagate
/// and carry inputs are as with ArbGeom.
pub fn carry_skip_finish_ram_fn(i: u64, p: &RAMParms) -> u64 {
    if let RAMParms::CarrySkipFinish {
            finish_srcs,
            n_finishes_so_far,
            stage_widths,
        } = p {

        // This RAM finishes one or more carry-skip setups,
        // meaning that selective increments have to be applied
        // with wraparound to the bit fields from RAMs to the north.
        let mut sum = 0;
        let mut lsb = 0;                // lsb moves right w/ each stage

        // The first RAM has just carry inputs, but each successive
        // have both propagate and carry inputs. So this limit will
        // count 1, 3, 5, 7, ... as far as necessary. But not
        // necessarily start at 1, because a RAM to the right may
        // have already been counting.
        let mut n_carry_prop_fields = 1 + 2 * n_finishes_so_far;

        for width in stage_widths {
            // Select just the bits that apply to this stage.
            let mask = (1 << width) - 1;

            // Temporarily right-align these bits.
            let mut stage_sum = (i >> lsb) & mask;

            // Finish the carry-skip inputs up to this stage only.
            stage_sum += compute_skipped_carry
                    (i, &finish_srcs[0..n_carry_prop_fields]);

            // Wrap to width, restore original alignment, and include in sum.
            stage_sum &= mask;
            sum |= stage_sum << lsb;

            // Note new position, and add next propagate and carry fields.
            lsb += width;
            n_carry_prop_fields += 2;
        }

        sum
    }
    else {
        panic!("arbgeom_ram_fn wrong parameters");
    }
}

/// This function implements a RAM that is a vertical composition
/// of two small RAMs. The top (first) RAM has all of its outputs
/// fed to the inputs of the bottom (second) RAM. This can be used
/// for simulation and testing, or for firmware generation.
pub fn vertical_compose_ram_fn(i: u64, p: &RAMParms) -> u64 {
    if let RAMParms::VerticalCompose {
            top_parms, bot_parms,
            top_func, bot_func,
            top_mask,
            top_to_bot_shuf,
            pkg_to_bot_shuf,
        } = p {

        // The top RAM's original input bits are preserved as the
        // composed RAM's least-significant input bits. Copy those,
        // and do whatever the top RAM originally did.
        let top_in = i & top_mask;
        let top_out = top_func.0(top_in, top_parms);

        // The top RAM's output bits comprise some of the bottom RAM's
        // input bits. They are sensitive to order, of course.
        let mut bot_in = 0;
        for (top_pv, bot_pv) in top_to_bot_shuf {
            bot_in |= (1 & top_out >> top_pv) << bot_pv;
        }

        // The remaining package (composed RAM's) input bits lead to
        // the remaining bottom RAM input bits. Also sensitive to order.
        for (pkg_pv, bot_pv) in pkg_to_bot_shuf {
            bot_in |= (1 & i >> pkg_pv) << bot_pv;
        }

        // The bottom RAM's outputs are the same as the composed RAM's
        // outputs. Do the bottom RAM calculation, and return that.
        let bot_out = bot_func.0(bot_in, bot_parms);
        bot_out
    }
    else {
        panic!("vertical_compose_ram_fn wrong parameters");
    }
}

/// This function implements a RAM that is a horizontal composition
/// of two small RAMs. The RAM outputs are strictly parallel, but
/// some inputs may be shared. This can be used for simulation and
/// testing, or for firmware generation.
pub fn horizontal_compose_ram_fn(i: u64, p: &RAMParms) -> u64 {
    if let RAMParms::HorizontalCompose {
            l_parms, r_parms,
            l_func, r_func,
            l_out_shift,
            r_in_mask,
            l_out_mask,
            r_out_mask,
            in_to_l_shuf,
        } = p {

        // The right RAM's original input bits are preserved as the
        // composed RAM's least-significant input bits. Copy those,
        // and do whatever the right RAM originally did.
        let r_in = i & r_in_mask;
        let r_out = r_func.0(r_in, r_parms) & r_out_mask;

        // The left RAM's inputs are a subset of the package input bits,
        // with some overlap possible with the right RAM's input bits.
        let mut l_in = 0;
        for (in_pv, l_pv) in in_to_l_shuf {
            l_in |= (1 & i >> in_pv) << l_pv;
        }

        // The bottom RAM's outputs are the catenation of the right
        // and left outputs.
        let l_out = l_func.0(l_in, l_parms) & l_out_mask;
        let pkg_out = l_out << l_out_shift | r_out;
        pkg_out
    }
    else {
        panic!("horizontal_compose_ram_fn wrong parameters");
    }
}

/// Emulate a Wireflop as if it's a RAM. For simulation and testing.
pub fn wireflop_ram_fn(i: u64, _: &RAMParms) -> u64 {
    i
}

/// This is merely a placeholder and not to be executed.
fn bogus_ram_fn(_: u64, _: &RAMParms) -> u64 {
    panic!();
}

/// This is a helper for arbgeom_ram_fn().
///
/// For some RAM with input vector `i`, unpack the carry-skip information
/// and determine this RAM's effective carry value.
///
/// `src` is an alternating sequence c0, p1, c1, p2, c2, ....
/// The number of terms is either 0 or odd, so it never ends with a propagate.
/// Each member of `src` is a tuple (r_shift, mask), where r_shift is
/// applied to `i` before masking. So the numeric result of each term
/// is `i >> r_shift & mask` (note >> has higher precedence).
fn compute_skipped_carry(i: u64, mut srcs: &[(u64, u64)]) -> u64 {
    // XXX Tested manually 28 May 2022.
    if srcs.is_empty() {
        return 0;
    }

    let (cs, cm) = srcs[0];             // shift & mask for carry
    let mut ecv = i >> cs & cm;         // effective carry value
    srcs = &srcs[1..];

    while let [(ps, pm), (cs, cm), ..] = srcs {
        let p = i >> ps & pm;           // propagate (compare to ecv)
        let c = i >> cs & cm;           // next ecv w/o propagated increment

        let incr = (p != 0 && ecv >= p) as u64;
        ecv = c + incr;
        srcs = &srcs[2..];
    }

    ecv
}

/// This function aggregates carry propagation from right to left.
/// Given input bits: p_n c_n ... p_2 c_2 p_1 c_1 p_0 c_0,
/// determine output bits: p_n c_n ... p_2 c_2 p_1 c_1,
/// where each position includes all lower values. TODO elucidate.
/// One little hiccough: "rightmost" RAMs have no carries in from
/// their right to propagate, so they have no p_0 input and no
/// propagate outputs at all.
pub fn propagator_ram_fn(mut i: u64, parms: &RAMParms) -> u64 {
    const X: u8 = 0;                    // placeholder output for illegal input
    const TABLE: [u8; 16] = [0, 0, 0, X, 1, 1, 1, X, 0, 1, 2, X, X, X, X, X];
/*
    0000  -- becomes 00 -
    0001  -G becomes 00 -
    0010  -P becomes 00 -
    0011  illegal
    0100  G- becomes 01 G
    0101  GG becomes 01 G
    0110  GP becomes 01 G
    0111  illegal
    1000  P- becomes 00 -
    1001  PG becomes 01 G
    1010  PP becomes 10 P
    1011  illegal
    1100  illegal
    1101  illegal
    1110  illegal
    1111  illegal
*/

    if let RAMParms::Propagator {
        n_stages,                       // # of carries in; should always > 1
        rightmost,
    } = parms {

        // On rightmost RAMs, p_0 is implicitly false instead of a wired input.
        // Insert this 0 bit into the computation.
        if *rightmost {
            i = 2 * i - (i & 1);
        }

        // Work from right to left.

        for stage in 1..*n_stages {
            let question = i >> 2 * (stage - 1) & 15;
            let answer = TABLE[question as usize] as u64;
            let mask_out = !(3 << 2 * stage);
            i = i & mask_out | answer << 2 * stage;
        }

        // Stage 0 is an identity and is suppressed from output.
        i >>= 2;

        // Rightmost RAMs don't have any propagate outputs, so we
        // have to remove those bits from the sequence.
        if *rightmost {
            i = odd_bits(i);
        }

        i
    }
    else {
        panic!("aggregator_ram_fn wrong parameters");
    }
}

fn odd_bits(mut i: u64) -> u64 {
    let mut pv = 1;
    let mut o = 0;
    while i != 0 {
        if i & 1 != 0 {
            o |= pv;
        }
        pv <<= 1;
        i >>= 2;
    }
    o
}

