///! solver.rs May 2022
///! Attempt to find near-optimal subproduct adder structures.

/*
    LEXICON (fill in definitions later)

    final  gpv  immediately-finalizable  lpv  mecv  n  pv
*/

// TODO Some multipliers will need 3-layer carry-skip mechanisms, along
//      with a theory to support their use. Also 4-layer, .... Big issue.

// XXX "lmul ideas spreadsheet.ods" has some really helpful design doc.

use crate::*;

pub const MAX_SOLN_COST: SolutionCost =
        (usize::MAX, usize::MAX, usize::MAX, usize::MAX);

type SetupFlavor<'m> =
        (BaggageStruct, &'m[LayerVocab], Package, Vec<LayerVocab>);

/// These are the names I use for RAM modalities.
#[derive(Clone, Debug, PartialEq)]
pub enum PackageKind {
    // XXX Adding to this list might break a p_setup.rs "fragile" test.
    // Keep overshoot_complement at the same offset for compiler optimization.
    ArbGeom {
        overshoot_complement: u64,      // used by synth; sic. u64
        finishing: bool,
    },
    CarrySkipFinish {
        stage_widths: Vec<u64>          // used by synth; sic. u64
    },     // sic. u64
    CarrySkipSetup {
        overshoot_complement: u64,      // used by synth; sic. u64
        finishing: bool,
        n_prop_bits: usize,
        n_carry_bits: usize,
    },
    Propagator {                        // for multilayer carry-skip
        inputs: Vec<PropNodeID>,
        rightmost: bool,
    },
    Joiner {                            // for multilayer carry-skip
        stage: usize,
        inputs: Vec<PropNodeID>,
    },
    Wireflop,
    EndOfLayer,
}
pub use PackageKind::*;

/// Here is just enough info for each RAM to synthesize a multiplier once
/// a solution has been decided on.
#[derive(Clone, Debug)]
pub struct Package {
    pub kind: PackageKind,              // kind of device chosen
    pub mem_used: usize,                // # actively used RAM cells in package
    pub output_pvs: (usize, usize),     // (from_incl, to_excl)

    // Here are place values of the bits consumed by this package, excluding
    // propagate & carry input. The len() is the number of addends.
    pub input_pvs: Vec<usize>,
}

/// A layer to parse (solve) is a Vec of these.
#[derive(Clone, Copy, Debug, PartialEq)]
pub enum LayerVocab {

    // This is a place value that needs another 1 added somewhere
    // in order to undo overshoot.
    // It will appear in advance of the same place value's Column.
    OffsetComplement {
        gpv: usize,
    },

    // Column is for "plain old subproduct bits" that we look to add.
    // Its result will be Wireflop, CarrySkipSetup,
    // or ArbGeom. To produce the finishing package
    // variants, the involved column group is preceded by a FinishInfo.
    Column {
        gpv: usize,                     // column place value = 2**gpv
        n: usize,                       // # of bits here
    },

    // This carry-skip finish information appears immediately prior to
    // bits we need to make final. The next enum to the left can be
    // Column (producing finishing ArbGeom or finishing CarrySkipSetup) or
    // FinishWithWidth (producing CarrySkipFinish). Because either enum
    // leftward includes the applicable gpv, FinishInfo doesn't include it.
    FinishInfo {
        prop_carry_baggage: usize,      // cumulative # of p & c inputs needed
        max_eff_carry_val: usize,       // contributes to output width
    },

    // When this appears, it is always preceded by a FinishInfo and will
    // result in a CarrySkipFinish.
    FinishWithWidth {
        gpv: usize,                     // carry lsb place value = 2**gpv
        width: usize,                   // # of columns made final here
    },
}
use LayerVocab::*;

/// These global values aren't constant, but don't change during a search.
#[derive(Debug)]
pub struct Globals {                    // reference is `g`
    pub orig_problem: Vec<LayerVocab>,  // original, unsolved problem
    pub product_width: usize,           // output width of full multiplier
    pub ram_dims: (usize, usize),       // in & out width of RAMs
    pub o_mask: usize,                  // fast detection of out width overflow

    // Output control fields
    pub n_test_muls: usize,             // number of test multiplications to do
    pub draw: Which,                    // what will be drawn
    pub synth: Which,                   // what will be synthesized
    pub test: Which,                    // what will be tested (requires synth)
    pub abandon_overcost: bool,
    pub d_flip_flops_avail: bool,       // enables cross-layer composition
    pub strategy: OptimizeStrategy,     // how exhaustively to search
    pub join_reservation: usize,        // only for multilayer carry-skip
}

/// These global values change during a search.
// [derive(Debug)] not possible: Wires doesn't implement Debug.
pub struct MutGlobals {                 // reference is `m`
    // These will change within a single layer.
    pub trial_soln: Vec<Package>,       // solution under construction
    pub next_problem: Vec<LayerVocab>,  // problem for next layer
    recording: bool,                    // used by OptimizeLayer
    farthest_setup: usize,              // used by OptimizeLayer

    // These only change between layers.
    pub best_soln: Vec<Package>,        // best solution seen thus far
    pub best_cost: SolutionCost,        // cost of best solution
    pub layers_in: usize,               // for pruning overcost search trees

    // These are written at synthesis and test time and
    // will contribute to the Report.
    pub something_failed: bool,         // true if any error is captured
    pub n_configs: u64,                 // # of solutions found to specs
    pub n_tests_right: u64,             // # of correct test products
    pub n_tests_wrong: u64,             // # of incorrect test products

    // This is updated at solution time.
    pub wires: Wires,                   // owned multiplier
}

/// This is parsing memory from the last-output package.
#[derive(Clone, Copy, Debug, PartialEq)]
enum Mindset {
    NeverMind,                          // nothing is happening
    MissingBits(usize),                 // ArbGeom partly consumed column
    Baggage(BaggageStruct),             // prop-carry info from right
}
use Mindset::*;

#[derive(Clone, Copy, Debug, PartialEq)]
struct BaggageStruct {
    n_carry_in: usize,              // # carries from immed right
    n_prop_carry_cumul: usize,      // cumulative p & c from right
    max_eff_carry_val: usize,       // affects # prop bits, max sum
}

impl MutGlobals {

    /// ArbGeom output columns can overlap, which somewhat complicates
    /// building the next problem. The good news is, they are produced
    /// contiguously, and the first column produced is the rightmost.
    /// That makes this merge function relatively simple.
    fn overlap_arbgeom_output(&mut self, from_col: usize, to_col: usize) {
        // MWA Manually tested 20 May 2022 and looks good.
        let np = &mut self.next_problem;

        // Locate where `from_col` would go in relation to the
        // indices of the next layer problem.
        let np_len = np.len();
        let mut align = np_len;
        for i_rev in (0..np_len).rev() {
            if let Column { gpv, .. } = np[i_rev] {
                if gpv >= from_col {    // will reach right alignment IF
                    align = i_rev;      // assumptions are valid
                } else {
                    break;
                }

            } else {
                break;
            }
        }

        // Split into fast increment and append portions. Do the work.
        let n_overlap = (np_len - align).min(to_col - from_col);
        for i in 0..n_overlap {
            if let Column { n, .. } = &mut np[align+i] {
                *n += 1;
            } else {
                panic!();
            }
        }
        for i in (from_col + n_overlap)..to_col {
            np.push(Column { gpv: i, n: 1 });
        }
    }

    // Work-in-progress appends to some of the Vecs in this struct.
    // Lengths are saved here so we can un-append and backtrack.
    fn take_snapshot(&self) -> (usize, usize) {
        (self.trial_soln.len(), self.next_problem.len())
    }

    // Restore the shorter versions so we can backtrack.
    fn restore_snapshot(&mut self, u: (usize, usize)) {
        self.trial_soln.truncate(u.0);
        self.next_problem.truncate(u.1);
    }

    // Determine the cost of this solution, prior to composition.
    fn estimated_cost(&self, g: &Globals) -> SolutionCost {
        let mut cost = self.wires.l0_cost;
        let mut joiner_layer = false;
        for pkg in &self.trial_soln {
            if pkg.kind == EndOfLayer {
                cost.0 += 1;
            }
            cost.1 += match pkg.kind {
                Wireflop | EndOfLayer => 0,
                Joiner { .. } => { joiner_layer = true; 0 },
                _ => 1,
            };
            cost.2 += pkg.mem_used;
        }
        if joiner_layer && g.join_reservation >= 3 {
            cost.0 -= 1;
        }
        cost
    }

    // Add a layer separator package to the trial solution.
    fn add_layer_sep(&mut self) {
        self.trial_soln.push(Package {
            kind: EndOfLayer,
            mem_used: 0,
            output_pvs: (0, 0),
            input_pvs: Vec::new(),
        });
    }
}

impl Mindset {
    fn missing_bits(&self) -> usize {
        if let MissingBits(n) = self { *n } else { 0 }
    }

    fn baggage(&self) -> BaggageStruct {
        match self {
            Baggage(b) => *b,
            _ => BaggageStruct {
                n_carry_in: 0,
                n_prop_carry_cumul: 0,
                max_eff_carry_val: 0,
            },
        }
    }

    /// If we do a carry-skip setup, there need to be at least two RAMs
    /// involved. Each RAM of the setup is called a "stage," although
    /// they are simultaneous. If only the rightmost RAM does a setup,
    /// it has the same effect as an ArbGeom, because its carry will
    /// not propagate through a second stage. We seek to disallow this
    /// condition, because (i) the drawings will show a CarrySkipSetup
    /// even though it's more akin to an ArbGeom, and (ii) the solver
    /// will also attempt the ArbGeom path, increasing the number of
    /// trials by a possibly large amount without appreciable benefit.
    ///
    /// This function intends to determine if we are immediately west
    /// of a first-stage setup RAM. If so and we're parsing an ArbGeom,
    /// we can fail that parse and backtrack over the first stage,
    /// perhaps saving a lot of CPU time.
    ///
    /// XXX Issue: This appears to work, but the more-aggressive
    /// ArbGeoms that take partial columns in rare cases can cause
    /// solution cost to increase. Usually this is due to composition.
    /// With composition off, we see this in 59e x 59e, but nowhere else
    /// from 1e through 64e. With composition checking ALL configurations,
    /// 59e x 59e turns out to be a wash whether or not stage_two_setup()
    /// is used. 14 June 2022
    fn stage_two_setup(&self) -> bool {
        if let Baggage(b) = self {
            b.n_carry_in != 0 && b.n_prop_carry_cumul == b.n_carry_in
        } else { false }
    }
}

/// Starting from `Specs`, produce an optimized multiplier plan.
pub fn plan_multiplier(specs: &Specs, scope: &Scope) -> Report {

    if scope.test == All && scope.synth != All {
        fail("In scope, you may not test All unless you synth All.");
    }
    if scope.test == Best && scope.synth == Omit {
        fail("In scope, you may not test Best unless you synth All or Best.");
    }
    if scope.strategy == OptimizeMulti && scope.synth != All {
        // FIXME The reason for this message (22Aug2022) is a nasty cost-
        // determination bug that prints estimated cost (which is too small)
        // if synth is Omit or Best. We want the actual cost, which we can
        // get if we honor the silly rule here. The problem is that the
        // solver is keeping the minimum cost, and once the estimate is too
        // low, it will be retained even if we have an actual cost for the
        // same solution. The permanent fix may be a little tricky. Note
        // that synth All for OptimizeMulti will only synth once, so at least
        // it won't slow down much.
        fail("In scope, you may not OptimizeMulti unless you synth All.");
    }

    // Design and netlist the subproduct layer. Get subproduct column counts.
    let mut w = Wires::new(specs);
    let mut overshoot_pvs = &w.lay_subproduct_rams()[..];
    let col_counts = w.solver_counts_hook();

    // Merge `col_counts` and `overshoot_pvs` into a Vec<LayerVocab>.
    // Overshoot is to precede addends for each place value.
    // No need for a special case here for implicit trailing wireflops;
    // that's done by search_new_layer().
    let mut orig_problem = Vec::new();
    for (gpv, n) in col_counts.into_iter().enumerate() {
        if !overshoot_pvs.is_empty() && overshoot_pvs[0] == gpv {
            orig_problem.push(OffsetComplement { gpv });
            overshoot_pvs = &overshoot_pvs[1..];
        }
        orig_problem.push(Column { gpv, n });
    }
    for leftover in overshoot_pvs {
        orig_problem.push(OffsetComplement { gpv: *leftover });
    }

    // Set up global data for the subproduct adder design search.
    let g = Globals {
        orig_problem,
        product_width: w.product_width,
        ram_dims: specs.l0_ram_dims,
        o_mask: !((1 << specs.l0_ram_dims.1) - 1),
        n_test_muls: scope.n_test_muls,
        draw: scope.draw,
        synth: scope.synth,
        test: scope.test,
        d_flip_flops_avail: scope.d_flip_flops_avail,
        abandon_overcost: scope.abandon_overcost,
        strategy: scope.strategy,
        join_reservation: scope.join_reservation,
    };

    let mut m = MutGlobals {
        next_problem: Vec::new(),
        trial_soln: Vec::new(),
        best_soln: Vec::new(),
        best_cost: MAX_SOLN_COST,
        layers_in: 0,
        n_configs: 0,
        something_failed: false,
        n_tests_right: 0,
        n_tests_wrong: 0,
        wires: w,
        recording: false,
        farthest_setup: 0,
    };

    // Recursively try to design the layers.
    search_new_layer(&g, &mut m, &g.orig_problem);

    if m.best_cost == MAX_SOLN_COST {
        m.something_failed = true;      // No solution was found.
    }

    // synth_and_test() can't accept a pointer to a solution, because
    // it also uses `&mut m`, and the solution is attached thereto.
    // So what I've done is have synth_and_test() always process
    // `m.trial_solution`. But now we're done, and what we want run
    // is `m.best_soln`. So we force a quick move from `m.best_soln` to
    // `m.trial_soln`.
    m.trial_soln.clear();
    std::mem::swap(&mut m.best_soln, &mut m.trial_soln);
    // XXX Do not use m.best_soln hence.

    // synth_and_test() conditionally updates some run counters on `m`.
    if let Some(cost) = synth_and_test(&g, &mut m, true) {

        // If we haven't been synthesizing non-best circuits, this
        // was our first opportunity to reduce cost by composition.
        // So we capture that result here.
        if cost < m.best_cost {
            m.best_soln = m.trial_soln.clone();
//p!(B "best cost s&t {:?}", cost);
            m.best_cost = cost;
        }
    }

    Report {
        specs: *specs,
        scope: *scope,
        product_width: g.product_width,
        heading: m.wires.heading,
        best_soln: m.trial_soln,        // sic.
        best_cost: m.best_cost,
        something_failed: m.something_failed,
        n_configs: m.n_configs,
        n_tests_right: m.n_tests_right,
        n_tests_wrong: m.n_tests_wrong,
    }
}

/// This is called at the beginning of a problem statement for a layer. Its
/// solutions are subsequent problem statements for the next layer, on which 
/// this function is called recursively.
///
/// This function doesn't do much actual searching, but mostly detects empty
/// problem statements, which indicates that the solution up to then is
/// complete and should be reported. search_within_layer() continues.
fn search_new_layer(
    g: &Globals,                        // fixed parameters for run
    m: &mut MutGlobals,                 // vars and progressive solution
    mut layer: &[LayerVocab],           // problem statement
) {
    // This speeds up certain searches a little bit.
    if g.abandon_overcost && m.layers_in > m.best_cost.0 {
        return;
    }
    m.layers_in += 1;                   // # of times search_new_layer() entered

    // Trailing singletons are final. Don't include them in the problem.
    // This also means trailing wireflops are always implicit, and
    // will never appear in a partial or complete solution.
    while let [Column { n: 1, .. }, ..] = layer {
        layer = &layer[1..];
    }

    // If there is nothing else in the layer, we have a complete multiplier
    // design. Retain the cheapest version. "Multiplier complete."
    if layer.is_empty() {

        // Determine the cost of this multiplier.
        // Synthesis is preferred, because it accounts for composition.
        // If we didn't synthesize, use the plans to estimate the cost.
        let cost = if let Some(c) = synth_and_test(g, m, false) {
            c
        } else {
//p!(R "*estimating* cost");
            m.estimated_cost(g)
        };
//p!(R "cost at the moment is {:?}", cost);

        // If this cost is the least yet, hang on to this solution and cost.
        if cost < m.best_cost {
            m.best_soln = m.trial_soln.clone();
            m.best_cost = cost;
        }

        // If drawing is requested for all solutions (not just the best),
        // show the cost and drawing. If it's the first multiplier out
        // from this specification, also show a specs heading.
        m.n_configs += 1;
        if g.draw == All {
            if m.n_configs == 1 {
                m.wires.heading.print(&m.wires.specs);
            }
            p!("Config #{} costs {:?}:\n", m.n_configs, cost);
            draw_solution(&m.trial_soln, g.product_width, g.join_reservation);
        }
    }

    // Otherwise, find solutions for what remains.
    else {
        search_within_layer(g, m, NeverMind, &layer);
    }

    m.layers_in -= 1;
}

/// This strangely-named function parses the least significant columns of
/// a problem statement, for the purpose of removing anything that definitely
/// does not backtrack.
///
/// XXX It would be reasonable to merge this into search_new_layer().
fn search_within_layer(
    g: &Globals,
    m: &mut MutGlobals,
    mut ms: Mindset,
    mut layer: &[LayerVocab],
) {
    // Zero or more wireflops do not require backtracking.
    if let Some((ms_, layer_)) = parse_wireflop(m, ms, layer) {
        ms = ms_;
        layer = layer_;
    }

    // Zero or more plain carry-skip finishes do not require backtracking.
    if let Some((ms_, layer_)) = parse_carry_skip_finish(g, m, layer) {
        ms = ms_;
        layer = layer_;
    }

    // FIXME 21Aug2022 Explain how this works.
    if g.strategy == OptimizeLayer {
        m.recording = true;
        m.farthest_setup = usize::MAX;
        let snapshot = m.take_snapshot();   // added 21Aug2022
        fork_search_here(g, m, ms, layer);
        m.restore_snapshot(snapshot);       // added 21Aug2022
        m.recording = false;
    }

    fork_search_here(g, m, ms, layer);
}

/// This is the recursive portion of an in-layer search for solutions.
/// We've already done the non-backtracking part of the layer (plain
/// CarrySkipFinish & Wireflop), so up next is CarrySkipSetups and then
/// ArbGeoms, subject to run parameters.
fn fork_search_here(
    g: &Globals,
    m: &mut MutGlobals,
    ms: Mindset,
    mut layer: &[LayerVocab],
) {
    match g.strategy {
        OptimizePackage | OptimizeLayer | OptimizeAll => {
            let snapshot = m.take_snapshot();
            if let Some(mut results) =
                    parse_carry_skip_setup(g, m, ms, layer) {

                // OptimizeAll tries the widest setup for every 
                // prop_carry_baggage, backtracking the entire depth of the
                // trial solution tree. It also allows carry-skip opportunities
                // to be ignored (by falling through to ArbGeom).
                if g.strategy == OptimizeAll {
                    for (baggage, layer, pkg, mut next_layer) in results {
                        let ms = Baggage(baggage);
                        m.trial_soln.push(pkg);
                        m.next_problem.append(&mut next_layer);
                        fork_search_here(g, m, ms, layer);
                        m.restore_snapshot(snapshot);
                    }
                }

                // OptimizePackage picks the immediate package that gives us
                // the widest carry-skip setup.
                else if g.strategy == OptimizePackage {
                    let (mut i_best, mut len_best) = (0, usize::MAX);
                    for (i, (_, l, _, _)) in results.iter().enumerate() {
                        if l.len() < len_best {
                            i_best = i;
                            len_best = l.len();
                        }
                    }
                    let (baggage, layer, pkg, mut next_layer) =
                            results.remove(i_best);
                    let ms = Baggage(baggage);
                    m.trial_soln.push(pkg);
                    m.next_problem.append(&mut next_layer);
                    fork_search_here(g, m, ms, layer);
                    m.restore_snapshot(snapshot);
                    if true {
                        // This prohibits ArbGeom if setup possible.
                        // FIXME 21Aug2022 Compatible w/ stage_two_setup()?
                        return;
                    }
                }

                // This tries to get as far as possible into a layer
                // with carry-skip setups, and then proceed with that alone.
                // XXX It does not consider there may be more than one way
                // to get as far as possible, to try to choose within those
                // alternatives. It uses "luck of the draw."
                else if g.strategy == OptimizeLayer {
                    for (baggage, layer, pkg, mut next_layer) in results {

                        if m.recording {
                            m.farthest_setup =
                                    m.farthest_setup.min(layer.len());
                        }

                        let ms = Baggage(baggage);
                        m.trial_soln.push(pkg);
                        m.next_problem.append(&mut next_layer);
                        fork_search_here(g, m, ms, layer);
                        m.restore_snapshot(snapshot);
                    }

                    // Abandon the search if we're recording, or if we're
                    // playing back but not at the farthest point yet.
                    if m.recording || layer.len() != m.farthest_setup {
                        return;
                    }
                }
            }
        },

        // This seeks to do a multi-layer carry-skip with no more than
        // two bits per place value across the remaining problem. The two-bit
        // check is within parse_multilayer_setup(). This does not backtrack.
        OptimizeMulti => {
            if parse_multilayer_setup(g, m, layer) {
                layer = &layer[0..0];       // No more input - it's ready to synth!
            }
        }

        // OptimizeNone disables all carry skips.
        OptimizeNone => (),
    }

    // Zero or more arbgeoms "do not require backtracking."
    // FIXME 21Aug2022 I have a hypothesis that corner cases exist where
    // non-greedy ArbGeom filling may lead to cheaper solutions. It's scary,
    // because the search space could become way larger, or more complex
    // code, or heuristics need invented, or ....
    if let Some(layer_) = parse_arb_geom(g, m, ms, layer) {
        layer = layer_;
    }

    // XXX 16 June 2022 parse_arbgeom() may be trying to backtrack over a
    // one-stage CarrySkipSetup. Otherwise, I think this layer would be empty.
    if !layer.is_empty() {
        return;
    }

    // Mark the end of this layer.
    m.add_layer_sep();

    // We need to recurse with m.next_problem as the new
    // layer to solve, but m has an outstanding mutable
    // reference. The solution is to detach next_problem
    // from m temporarily.
    let mut detached = Vec::new();
    std::mem::swap(&mut detached, &mut m.next_problem);

    // Otherwise, solve this new layer.
    search_new_layer(g, m, &detached);

    // Reattach next_problem back into m so we can backtrack.
    std::mem::swap(&mut detached, &mut m.next_problem);
}

include!("p_arbgeom.rs");
include!("p_finish.rs");
include!("p_multi.rs");
include!("p_setup.rs");
include!("p_wireflop.rs");
