///! split.rs February 2022
///! This module selects geometry for the first layer of RAMs in a
///! multiplier, the layer where subproducts are generated.

use crate::*;

#[derive(Debug)]
pub struct SubproductMetrics {
    pub offset: i64,    // offset to force all any product non-negative
    pub max_prod: i64,  // maximum possible product including `offset`
    pub width: usize,   // width of `max_prod` in bits
}

/// Find a reasonable way to break the left and right factors
/// for subproduct multiplication within the available RAM size.
/// Start with considering all possible numbers to break into,
/// and tally their cost according to the number of output bits.
///
/// The return value is
/// (n_left_subwords, n_right_subwords, approx_n_subproduct_bits).
/// This is not the final count of subproduct bits. The final count
/// has been seen lower, because at least one of the subproduct RAM
/// outputs is beyond the final multiplier output place value. The
/// final count has been seen higher, because at least one of the
/// subproduct RAM outputs had to be widened in order to cancel the
/// additive overshoot. But it was a good metric for splitting into
/// subwords.
///
/// XXX 4096e x 4096e took 677 seconds here on 1 June 2022.
///
/// XXX Changed approach 5 June to fewest RAMs, then fewest bits.
///     Let's see if that fixes a new problem I'm having; that
///     new "expensive double loop" is causing preposterous
///     choices that are costly to the adder layers also.
pub fn _old_select_subproduct_split(s: &Specs) -> (usize, usize, usize) {

    let mut fewest_bits_and_rams = (usize::MAX, usize::MAX);
    let mut fewest_nl_nr = (0, 0);

    // Because all left subwords will be multiplied against all
    // right subwords, any RAM's input will need to accommodate
    // both the left and right maximum subword size simultaneously.
    // The minimum possible subword size depends on whether or not
    // we need signage input for the most significant subword. Thus:
    let min_l_sw_bits = 1 + (s.sue_l == Either) as usize;
    let min_r_sw_bits = 1 + (s.sue_r == Either) as usize;

    // This is the first of two input width checks.
    // I think the two checks are redundant, but I haven't confirmed. XXX
    if s.l0_ram_dims.0 < min_l_sw_bits + min_r_sw_bits {
        fail("RAM input width is too narrow for these signages.");
    }

    // Here are the widths of the words with any needed signage input.
    let width_l_signage = s.width_l + (s.sue_l == Either) as usize;
    let width_r_signage = s.width_r + (s.sue_r == Either) as usize;

    // Now we loop over the allocations of these bits, which determines
    // how many partitions of the left and right full words are needed.
    for n_l_sw_bits in min_l_sw_bits ..= s.l0_ram_dims.0 - min_r_sw_bits {
        let n_r_sw_bits = s.l0_ram_dims.0 - n_l_sw_bits;

        // These two lines are simply ceiling division.
        let n_l_parts = (width_l_signage + n_l_sw_bits - 1) / n_l_sw_bits;
        let n_r_parts = (width_r_signage + n_r_sw_bits - 1) / n_r_sw_bits;

        // This expensive double loop was added 1 June 2022, because neither
        // n_l_parts nor n_r_parts has considered the RAM may be too narrow
        // for the split they indicate. So these figures are a minimum number
        // of subwords to divide into. The maximum number is obvious.
        for n_l_parts in n_l_parts..=s.width_l {
            for n_r_parts in n_r_parts..=s.width_r {

                // There's a really heinous corner case, such as
                // 1s x 2s using l0_ram_dims (4, 2), where we can
                // represent the singleton subproduct fine in two bits
                // because it's offset, but the full product does not
                // itself fit in two bits. That's fine in solutions
                // with more than one RAM, but we can't get rid of
                // the overshoot complement when there is only one RAM
                // and that RAM is too small.
                if n_l_parts == 1 && n_r_parts == 1 && predict_product_width
                        (s.width_l, s.width_r, s.sue_l, s.sue_r)
                        > s.l0_ram_dims.1 {
                    continue;
                }

                // Now create the actual partitioning of these integers.
                // XXX This can be written more insightfully and run faster.
                let parts_l = partition_word_size(s.width_l, n_l_parts);
                let parts_r = partition_word_size(s.width_r, n_r_parts);

                // Now see if the multiplication can be done for these
                // integers, and hang onto the partitioning plan that
                // produces the fewest subproduct bits that need adding.
                // XXX This can be written more insightfully and run faster.
                match mult_possible_for(
                        &parts_l, &parts_r, s.sue_l, s.sue_r, s.l0_ram_dims) {
                    None => continue,
                    Some(n_prod_bits) => {

                        // We look for the solution with fewest product bits,
                        // and in the event of a tie, the fewest RAMs. One such
                        // tie occurrs for 128 Either * 128 Either, where
                        // 15 x 15 RAMs or 13 x 17 RAMs or 17 x 13 RAMs all
                        // produce 3869 subproduct bits.
                        //
                        // TODO Consider whether there can ever be a case where
                        // fewer RAMs might result in an acceptable increase in
                        // the number of subproduct bits.
                        if (n_prod_bits, n_l_parts * n_r_parts) <
                                fewest_bits_and_rams {
                            fewest_bits_and_rams =
                                    (n_prod_bits, n_l_parts * n_r_parts);
                            fewest_nl_nr = (n_l_parts, n_r_parts);
                        }

                        // If this succeeded at n_r_parts, no need to
                        // divide the right subword any smaller. This
                        // is a huge performance savings: 256e x 256e
                        // goes from 128 sec to 0.2 sec to split.
                        break;
                    },
                }
            }
        }
    }

    if fewest_bits_and_rams.0 == usize::MAX {
        fail("could not break problem into subproducts - RAM width issue?");
    }

    (fewest_nl_nr.0, fewest_nl_nr.1, fewest_bits_and_rams.0)
}

/// Find a reasonable way to break the left and right factors
/// for subproduct multiplication within the available RAM size.
/// Start with considering all possible numbers to break into,
/// and tally their cost according to the number of output bits.
///
/// The return value is
/// (n_left_subwords, n_right_subwords, approx_n_subproduct_bits).
/// This is not the final count of subproduct bits. The final count
/// has been seen lower, because at least one of the subproduct RAM
/// outputs is beyond the final multiplier output place value. The
/// final count has been seen higher, because at least one of the
/// subproduct RAM outputs had to be widened in order to cancel the
/// additive overshoot. But it was a good metric for splitting into
/// subwords.
///
/// XXX 4096e x 4096e took 677 seconds here on 1 June 2022.
pub fn select_subproduct_split(s: &Specs) -> (usize, usize, usize) {

    let mut fewest_rams_and_bits = (usize::MAX, usize::MAX);
    let mut fewest_nl_nr = (0, 0);

    // Because all left subwords will be multiplied against all
    // right subwords, any RAM's input will need to accommodate
    // both the left and right maximum subword size simultaneously.
    // The minimum possible subword size depends on whether or not
    // we need signage input for the most significant subword. Thus:
    let min_l_sw_bits = 1 + (s.sue_l == Either) as usize;
    let min_r_sw_bits = 1 + (s.sue_r == Either) as usize;

    // This is the first of two input width checks.
    // I think the two checks are redundant, but I haven't confirmed. XXX
    if s.l0_ram_dims.0 < min_l_sw_bits + min_r_sw_bits {
        fail("RAM input width is too narrow for these signages.");
    }

    // Here are the widths of the words with any needed signage input.
    let width_l_signage = s.width_l + (s.sue_l == Either) as usize;
    let width_r_signage = s.width_r + (s.sue_r == Either) as usize;

    // Now we loop over the allocations of these bits, which determines
    // how many partitions of the left and right full words are needed.
    for n_l_sw_bits in min_l_sw_bits ..= s.l0_ram_dims.0 - min_r_sw_bits {
        let n_r_sw_bits = s.l0_ram_dims.0 - n_l_sw_bits;

        // These two lines are simply ceiling division.
        let n_l_parts = (width_l_signage + n_l_sw_bits - 1) / n_l_sw_bits;
        let n_r_parts = (width_r_signage + n_r_sw_bits - 1) / n_r_sw_bits;

        // This expensive double loop was added 1 June 2022, because neither
        // n_l_parts nor n_r_parts has considered the RAM may be too narrow
        // for the split they indicate. So these figures are a minimum number
        // of subwords to divide into. The maximum number is obvious.
        for n_l_parts in n_l_parts..=s.width_l {
            for n_r_parts in n_r_parts..=s.width_r {

                // There's a really heinous corner case, such as
                // 1s x 2s using l0_ram_dims (4, 2), where we can
                // represent the singleton subproduct fine in two bits
                // because it's offset, but the full product does not
                // itself fit in two bits. That's fine in solutions
                // with more than one RAM, but we can't get rid of
                // the overshoot complement when there is only one RAM
                // and that RAM is too small.
                if n_l_parts == 1 && n_r_parts == 1 && predict_product_width
                        (s.width_l, s.width_r, s.sue_l, s.sue_r)
                        > s.l0_ram_dims.1 {
                    continue;
                }

                // Now create the actual partitioning of these integers.
                // XXX This can be written more insightfully and run faster.
                let parts_l = partition_word_size(s.width_l, n_l_parts);
                let parts_r = partition_word_size(s.width_r, n_r_parts);

                // Now see if the multiplication can be done for these
                // integers, and hang onto the partitioning plan that
                // produces the RAMs.
                // XXX This can be written more insightfully and run faster.
                // FIXME This function is not necessarily correct.
                match mult_possible_for(
                        &parts_l, &parts_r, s.sue_l, s.sue_r, s.l0_ram_dims) {
                    None => continue,
                    Some(n_prod_bits) => {
                        // We look for the solution with fewest RAMs
                        // and in the event of a tie, the fewest bits.
                        if (n_l_parts * n_r_parts, n_prod_bits) <
                                fewest_rams_and_bits {
                            fewest_rams_and_bits =
                                    (n_l_parts * n_r_parts, n_prod_bits);
                            fewest_nl_nr = (n_l_parts, n_r_parts);
                        }

                        // If this succeeded at n_r_parts, no need to
                        // divide the right subword any smaller. This
                        // is a huge performance savings: 256e x 256e
                        // goes from 128 sec to 0.2 sec to split.
                        break;
                    },
                }
            }
        }
    }

    if fewest_rams_and_bits.0 == usize::MAX {
        fail("could not break problem into subproducts - RAM width issue?");
    }

    (fewest_nl_nr.0, fewest_nl_nr.1, fewest_rams_and_bits.1)
}

/// About evenly split `width` into `n` parts.
/// The larger parts appear first.
/// partition_word_size(15, 4) -> [4, 4, 4, 3]
pub fn partition_word_size(width: usize, n: usize) -> Vec<usize> {
    let q = width / n;
    let r = width % n;
    let mut parts = vec!(q; n);
    for i in 0..r {
        parts[i] += 1;
    }
    parts
}

/// Given this partitioning of the left and right factors along
/// with their signage, determine whether this RAM size is wide
/// enough (both in and out) to multiply all subproducts.
///
/// Neither `parts_l` and `parts_r` will ever be empty.
///
/// The return value is `Some(n_bits)` or `None`, depending on
/// whether all subproducts can be multiplied by this RAM size.
/// `n_bits` is the total number of output subproduct bits,
/// and usually much larger than the product width.
fn mult_possible_for(
    parts_l: &[usize],
    parts_r: &[usize],
    sue_l: Signage,
    sue_r: Signage,
    ram_dims: (usize, usize)
) -> Option<usize> {

    // This is the second of two input width checks.
    // I think the two checks are redundant, but I haven't confirmed.
    // The RAM needs enough inputs to take the longest left subword
    // with the longest right subword. Get the maximum length from each:
    let mut longest_l = *parts_l.iter().max().unwrap();
    let mut longest_r = *parts_r.iter().max().unwrap();

    // None of these lengths consider the control decoder's signage
    // bits that may be present for the most significant subword.
    if sue_l == Either {
        longest_l = longest_l.max(1 + parts_l.last().unwrap());
    }
    if sue_r == Either {
        longest_r = longest_r.max(1 + parts_r.last().unwrap());
    }

    if longest_l + longest_r > ram_dims.0 {
        return None;
    }

    // Now consider the output width. We have to look at the width of
    // all subword products in light of the subword widths and their
    // signages. Note that all but the most significant subwords are
    // Unsigned, but the most significant subwords can be Unsigned,
    // Signed, or Either. When they are Signed or Either, the output
    // width must be sufficient to allow offset-binary representation
    // of the (possibly) signed outputs.
    let mut n_bits_out = 0;
    for (index_l, width_l) in parts_l.iter().enumerate() {
        let signage_l = if index_l + 1 == parts_l.len()
                { sue_l } else { Unsigned };
        for (index_r, width_r) in parts_r.iter().enumerate() {
            let signage_r = if index_r + 1 == parts_r.len()
                    { sue_r } else { Unsigned };
            let mets = subproduct_metrics
                    (*width_l, *width_r, signage_l, signage_r);
            n_bits_out += mets.width;
            if mets.width > ram_dims.1 {
                return None;
            }
        }
    }

    Some(n_bits_out)
}

/// Given two "small" words (which each may have a sign bit, but never a
/// signage bit), and their declared signages (Unsigned/Signed/Either),
/// determine how many bits are needed to hold their product. This is
/// merely for my understanding.
fn product_size_experiment(
    width_l: usize,
    width_r: usize,
    sue_l: Signage,
    sue_r: Signage
) -> usize {
    assert!(width_l < 31 && width_r < 31);

    // minimum and maximum possible inputs
    let (l_min, l_max) = min_max(width_l, sue_l);
    let (r_min, r_max) = min_max(width_r, sue_r);

    // largest and smallest products
    let p = (l_min * r_min, l_min * r_max, l_max * r_min, l_max * r_max);
    let p_min = p.0.min(p.1.min(p.2.min(p.3)));
    let p_max = p.0.max(p.1.max(p.2.max(p.3)));

    // necessary width of these extrema in twos complement
    let pos_width = (usize::BITS - p_max.leading_zeros()) as usize;
    let neg_width = if p_min < 0 {
        (usize::BITS -
            p_min.unsigned_abs().next_power_of_two().leading_zeros()) as usize
    } else { 0 };

    // assurance the negative place won't overlap a positive place value
    let neg_width = if neg_width != 0 && neg_width <= pos_width {
        1 + pos_width
    } else { neg_width };

    // larger of the two extrema widths
    neg_width.max(pos_width)
}

/// Given two word sizes and their declared signages
/// (Unsigned / Signed / Either), predict how many bits are needed to hold
/// their product. This is to verify my understanding of what an actual
/// computation of the product size would produce. I don't have a proof.
pub fn predict_product_width(
    width_l: usize,
    width_r: usize,
    sue_l: Signage,
    sue_r: Signage
) -> usize {
    if width_l == 1 && sue_l == Unsigned {
        width_r
    }
    else if width_r == 1 && sue_r == Unsigned {
        width_l
    }
    else if width_l == 1 && width_r == 1 {
        1
    }
    else {
        width_l + width_r
    }
}

/// Make a table using product_size_experiment().
pub fn _product_size_table() {
    for sue_l in &[Unsigned, Signed] {
        for sue_r in &[Unsigned, Signed] {
            pp!("{:?} x {:?} =>", sue_l, sue_r);
            for width_l in 1..30 {
                for width_r in 1..30 {
                    let prod_width = product_size_experiment
                            (width_l, width_r, *sue_l, *sue_r);
                    let guessed_width = predict_product_width
                            (width_l, width_r, *sue_l, *sue_r);
                    assert_eq!(prod_width, guessed_width);
                    pp!(" {}x{}->{},", width_l, width_r, prod_width);
                }
            }
            p!();
        }
    }
}

/// Given two subwords (which each may have a sign bit, but never a
/// signage bit), and their declared signages (Unsigned/Signed/Either),
/// determine metrics for representing their product.
pub fn subproduct_metrics(
    width_l: usize,
    width_r: usize,
    sue_l: Signage,
    sue_r: Signage
) -> SubproductMetrics {

    // minimum and maximum possible inputs
    let (l_min, l_max) = min_max(width_l, sue_l);
    let (r_min, r_max) = min_max(width_r, sue_r);

    // the four extrema products
    let p = (l_min * r_min, l_min * r_max, l_max * r_min, l_max * r_max);

    // This is a tedious way to write a minimum,
    // but not as tedious as using an array and iterator
    let p_min = p.0.min(p.1.min(p.2.min(p.3)));

    // If the smallest product is less than zero, we need an offset
    // that can overcome it. Otherwise, that offset can be 0.
    //
    // XXX I used to use p_min.unsigned_abs().next_power_of_two()
    //     to "simplify" the offset, but it turns out the total
    //     number of complement bits for the whole multiplier is
    //     much smaller on average if I don't round up the offset.
    //     Also, the complement bits are further leftward if we
    //     don't round up here, suggesting they will be less of
    //     a carry propagation load.
    //
    // XXX I wonder if it is possible to incorporate these offset
    //     bits into the subproduct RAMs, causing the need to add
    //     them in later layers to disappear. If so, is it ALWAYS
    //     possible, so we don't need code to do this in the later
    //     layers? Because code to perform this at layer 0 may
    //     present a bit of a project.
//  let offset = if p_min < 0 {
//      p_min.unsigned_abs().next_power_of_two() as i64
//  } else { 0 };
    let offset = if p_min < 0 { p_min.abs() } else { 0 };

    // Largest product + offset, and the width of that.
    // Here again, method syntax has tedious consequences sometimes.
    let max_prod = p.0.max(p.1.max(p.2.max(p.3))) + offset;
    let width = (usize::BITS - max_prod.leading_zeros()) as usize;

/*
    // XXX 9 Apr 2022 This is a TEST. 
    //
    // Test result: For RAM sizes of (8-in, 8-out) and larger, and
    // multipliers where both factors are 2 bits or more, this appears
    // to fully satisfy the additive offset need within layer 1.
    // When either factor is 1 bit wide, offset help is needed after layer 1.
    // When either factor is 0 bits wide, the multiplier requires no RAMs.
    // For (7-in, 8-out) and (8-in, 7-out) RAMs, offset help is needed
    // after layer 1.
    let width =
        if sue_l == Signed && sue_r == Signed { width + 1 } else { width };
*/

    // Final metrics are these.
    SubproductMetrics { offset, max_prod, width }
}

/// Find the largest and smallest integers that can be
/// represented in this width with the given signage.
fn min_max(width: usize, signage: Signage) -> (i64, i64) {
    match signage {
        Signed => (-(1 << width - 1), (1 << width - 1) - 1),
        Unsigned => (0, (1 << width) - 1),
        Either => (-(1 << width - 1), (1 << width) - 1),
    }
}

/// subproduct split testing
pub fn _test_split() {
    p!(Y "test_split output:\n");

    let small = 1..=36;
    let big = [32, 36, 64, 72, 128, 144, 256, 512,
            1024, 2048, 4096, 8192, 16384];

    for i in small.chain(big) {

        p!("{} x {}", i, i);

        let specs = Specs {
            width_l: i,         // XXX only symmetric cases tested here
            width_r: i,
            sue_l: Either,      // XXX only Either SUE here
            sue_r: Either,
            l0_ram_dims: (18, 18),
        };

        select_subproduct_split(&specs);
        p!();
    }
}
