///! synth.rs May 2022
///! Wire up a solution from Solver and run some trial multiplications.

use std::collections::HashMap;
use crate::*;

// This is for memoizing Joiner RAMs so dupes are not synthesized.
type Memo = HashMap<((usize, usize), (usize, usize)), (usize, usize)>;

// This function fuses nodes together in lieu of wireflops.
// So instead of using D flip-flops, there are simply wires feeding
// bits forward that will need to be held for future clock pulses.
//
// The benefit is, logic that composes two RAMs vertically is easier
// to write. Also, multiplier simulations will run a little faster.
//
// The drawback is, engineers who are trying to build real circuits
// from the output of this program may become confused.
fn dont_make_wireflops(
    w: &mut Wires,
    out_layer: usize,
    from_bits: &mut Vec<Vec<usize>>,
) {
    // `from_bits` exceeds the product width where propagates and carries
    // are attached. We don't fuse those. But `from_bits` could be
    // less than the product width, typically for Unsigned * Unsigned
    // multiplications' original subproduct layer. We have to clamp our
    // iteration to the smaller of the actual width and product width.
    let clamp = from_bits.len().min(w.product_width);
    for (pv, some) in from_bits[0..clamp].iter().enumerate() {

        // `some` is the list of bits we may not have processed at this
        // place value. Allowed are zero (nothing to do) or one (wireflop).
        // Any place values with more than one bit should have generated
        // other packages by now.
        let len = some.len();
        continue_if!(len == 0);
        assert!(len == 1);

        let bit_index: usize = some[0];
        w.fuse_node(out_layer, pv, bit_index);
    }
}

/// Synthesize a multiplier and see if its products are correct.
pub fn synth_and_test(g: &Globals, m: &mut MutGlobals, the_best: bool)
        -> Option<SolutionCost> {

    // SYNTHESIS PORTION

    if g.synth == Omit || !the_best && g.synth == Best {    // not requested
        return None;
    }

    // A helper function synthesizes each layer in order to keep
    // functions small. A meticulous Rust dance gives us immutable
    // input information and mutable output information to use here.
    let layer_solns = split_solution_by_layers(&m.trial_soln);
    let mut cookie = usize::MAX;        // wiring layer where setup RAMs output
    let mut nofinish = false;
    let mut joiner_map = HashMap::new();

    for layer_soln in &layer_solns {
        break_if!(layer_soln.len() == 0);       // FIXME Why this break?
        // XXX We commit here to truncate before this function returns.
//p!(b "layer: {:#?}", layer_soln);
        synthesize_layer(layer_soln, &mut m.wires, &mut cookie,
                &mut nofinish, &mut joiner_map);
    }

    // Some small RAMs can be composed into larger RAMs.
    optimize_compositions(
        &mut m.wires,
        g.d_flip_flops_avail,
        g.strategy != OptimizeMulti,    // horiz is slow, unneeded for 'Multi
    );

    m.wires.time_sanity_check();

//p!(B "ram_layers {:#?}", m.wires.ram_layers);

    // Retain the cost.
    let cost = Some(m.wires.actual_cost());

    // TEST PORTION

    // Only test if it's requested and hasn't happened yet.
    if the_best && g.test == Best || !the_best && g.test == All {

        // Spot check that this multiplier does arithmetic correctly.
        // There is no output printed if everything works.
        let (oops, right, wrong) = test_mul(&mut m.wires, g.n_test_muls);
        m.something_failed |= oops;
        m.n_tests_right += right;
        m.n_tests_wrong += wrong;
    }

    // Tear this circuit back to just subproducts so we can try another.
    m.wires.truncate_to_one_layer();

    cost
}

// Synthesize just one layer.
fn synthesize_layer(
    pkgs: &[Package],
    w: &mut Wires,
    cookie: &mut usize,
    nofinish: &mut bool,
    joiner_map: &mut Memo,
) {
    // Here is the set of "unclaimed" incoming physical bits.
    // We can pop instances from these by place value.
    let mut from_bits = w.from_lpi.last().unwrap().clone();

    // MWA 2-June-2022 8:53 PM added as a test FIXME
    while from_bits.len() < w.product_width {
        from_bits.push(Vec::new());
    }

    // index of the not-yet-existing layer of output nodes
    // (not valid for Joiner schemes)
    let out_layer = w.from_lpi.len();

    // counter to locate pv where carries and props alternate
    let mut next_carry_prop_pv = w.product_width;

    let mut n_finishes_so_far = 0;

    // RAMs to append
    let mut rams = Vec::new();

    // Look at each package specified.
    for p in pkgs {
        match &p.kind {
            // Not all wireflops are explicitly here, so we defer until
            // all else is done, and then harvest what is left.
            Wireflop => {
                assert_eq!(             // sanity test
                    p.input_pvs,
                    (p.output_pvs.0..p.output_pvs.1).collect::<Vec<usize>>(),
                )
            },

            ArbGeom { overshoot_complement, finishing } => {
                let base_col = p.output_pvs.0;
                let mut inputs = Vec::new();
                let mut outputs = Vec::new();
                let mut finish_srcs = Vec::new();

                // addend inputs by position within RAM
                let mut i2a = Vec::<u64>::new();

                for pv in &p.input_pvs {
                    // Grab an untaken `w.bits` index for this pv,
                    // and hook a RAM input to it.
                    let untaken = from_bits[*pv].pop().unwrap();
                    inputs.push(untaken);

                    // Let the RAM know the local additive weight of this input.
                    i2a.push(1 << *pv - base_col);
                }

                // The inputs continue with propagate and finish info.
                if *finishing {
                    let n_cols = 2 * n_finishes_so_far + 2;

                    // Cols here are props, carries, props, carries, ...
                    // We can't just pop instances from from_bits for
                    // props & carries, because multiple RAMs need them.
                    // Some columns may be empty, and the first props
                    // column is in fact skipped (nothing carries to it).
                    for l_col in 1..n_cols {

                        // Get the bit locations for this column.
                        let g_col = l_col + w.product_width;
                        let bit_indices = &from_bits[g_col];

                        // Show the RAM how to intepret these bits.
                        let shift = inputs.len() as u64;
                        let mask = (1 << bit_indices.len()) - 1;
                        finish_srcs.push((shift, mask));

                        // Connect these bits to RAM inputs.
                        for bit_index in bit_indices {
                            inputs.push(*bit_index);
                        }
                    }

                    n_finishes_so_far += 1;
                }

                for pv in p.output_pvs.0..p.output_pvs.1 {
                    // Create a `w.bits` index for a new output bit at this pv,
                    // and hook a RAM output to it.
                    let new_index = w.add_node(out_layer, pv);
                    outputs.push(new_index);
                }

                rams.push(RAM {
                    dims: w.specs.l0_ram_dims,
                    inputs,
                    outputs,
                    func: Funky(arbgeom_ram_fn),
                    func_parms: RAMParms::ArbGeom {
                        input_to_addend: i2a,
                        overshoot_complement: *overshoot_complement,
                        finish_srcs,
                        carry_lsb: 0,   // disabled by prop_max
                        prop_lsb: 0,    // disabled by prop_max
                        prop_max: 0,    // disable carry-skip setup
                    },
                    marker: 0,
                });
            },

            CarrySkipFinish { stage_widths } => {

                // Multilayer carry-skip schemes have a
                // PackageKind::CarrySkipFinish layer, but it
                // does not drive RAMParms::CarrySkipFinish synthesis.
                // PackageKind::Joiner causes RAMParms::CarrySkipFinish instead.
                if *nofinish {
                    return;
                }

                let mut inputs = Vec::new();
                let mut outputs = Vec::new();
                let mut finish_srcs = Vec::new();
                // XXX .clone necessary?
                let stage_widths = stage_widths.clone();
                let n_stages = stage_widths.len();

                for pv in &p.input_pvs {
                    // Grab an untaken `w.bits` index for this pv,
                    // and hook a RAM input to it.
                    let untaken = from_bits[*pv].pop().unwrap();
                    inputs.push(untaken);
                }

                // The inputs continue with propagate and finish info.
                let n_cols = 2 * (n_stages + n_finishes_so_far);

                // Cols here are props, carries, props, carries, ...
                // We can't just pop instances from from_bits for
                // props & carries, because multiple RAMs need them.
                // Some columns may be empty, and the first props
                // column is in fact skipped (nothing carries to it).
                for l_col in 1..n_cols {

                    // Get the bit locations for this column.
                    let g_col = l_col + w.product_width;
                    let bit_indices = &from_bits[g_col];

                    // Show the RAM how to intepret these bits.
                    let shift = inputs.len() as u64;
                    let mask = (1 << bit_indices.len()) - 1;
                    finish_srcs.push((shift, mask));

                    // Connect these bits to RAM inputs.
                    for bit_index in bit_indices {
                        inputs.push(*bit_index);
                    }
                }

                for pv in p.output_pvs.0..p.output_pvs.1 {
                    // Create a `w.bits` index for a new output bit at this pv,
                    // and hook a RAM output to it.
                    let new_index = w.add_node(out_layer, pv);
                    outputs.push(new_index);
                }

                rams.push(RAM {
                    dims: w.specs.l0_ram_dims,
                    inputs,
                    outputs,
                    func: Funky(carry_skip_finish_ram_fn),
                    func_parms: RAMParms::CarrySkipFinish {
                        finish_srcs,
                        n_finishes_so_far,
                        // TODO is .clone() actually needed?
                        stage_widths,
                    },
                    marker: 0,
                });

                n_finishes_so_far += n_stages;
            },

            CarrySkipSetup {
                overshoot_complement,
                finishing, n_prop_bits,
                n_carry_bits
            } => {
                let base_col = p.output_pvs.0;
                let mut inputs = Vec::new();
                let mut outputs = Vec::new();
                let mut finish_srcs = Vec::new();

                // addend inputs by position within RAM
                let mut i2a = Vec::<u64>::new();

                for pv in &p.input_pvs {
                    // Grab an untaken `w.bits` index for this pv,
                    // and hook a RAM input to it.
                    let untaken = from_bits[*pv].pop().unwrap();
                    inputs.push(untaken);

                    // Let the RAM know the local additive weight of this input.
                    i2a.push(1 << *pv - base_col);
                }

                // The inputs continue with propagate and finish info.
                if *finishing {
                    let n_cols = 2 * n_finishes_so_far + 2;

                    // Cols here are carries, props, carries, props, ...
                    // We can't just pop instances from from_bits for
                    // props & carries, because multiple RAMs need them.
                    // Some columns may be empty, and the first props
                    // column is in fact skipped (nothing carries to it).
                    for l_col in 1..n_cols {

                        // Get the bit locations for this column.
                        let g_col = l_col + w.product_width;
                        let bit_indices = &from_bits[g_col];

                        // Show the RAM how to intepret these bits.
                        let shift = inputs.len() as u64;
                        let mask = (1 << bit_indices.len()) - 1;
                        finish_srcs.push((shift, mask));

                        // Connect these bits to RAM inputs.
                        for bit_index in bit_indices {
                            inputs.push(*bit_index);
                        }
                    }

                    n_finishes_so_far += 1;
                }

                // The output pvs looped over here do not include carry bits.
                for pv in p.output_pvs.0..p.output_pvs.1 {
                    // Create a `w.bits` index for a new output bit at this pv,
                    // and hook a RAM output to it.
                    let new_index = w.add_node(out_layer, pv);
                    outputs.push(new_index);
                }

                // Get output node pseudo-pvs where carry & prop bits will
                // be sent to. Either pseudo-pv may be empty.
                let prop_pv = next_carry_prop_pv;
                let carry_pv = prop_pv + 1;
                next_carry_prop_pv += 2;

                // Connect the carry output bits.
                let carry_lsb = outputs.len() as u64;


                // Rightmost setup RAMs have no prop bits. And all-
                // singleton, non-leftmost CarrySkipSetup RAMs have
                // no carry bits. In either case, make sure an
                // empty list of instances are available in these
                // place values.
                w.reserve_for_node(out_layer, prop_pv);
                w.reserve_for_node(out_layer, carry_pv);

                for _ in 0..*n_carry_bits {
                    // Create a `w.bits` index for a new output bit at
                    // carry_pv, and hook the next RAM output to it.
                    let new_index = w.add_node_fast(out_layer, carry_pv);
                    outputs.push(new_index);
                }

                // Connect the propagate output bits.
                let prop_lsb = outputs.len() as u64;
                let prop_max = n_prop_bits.smo() as u64;
                for _ in 0..*n_prop_bits {
                    // Create a `w.bits` index for a new output bit at
                    // prop_pv, and hook the next RAM output to it.
                    let new_index = w.add_node_fast(out_layer, prop_pv);
                    outputs.push(new_index);
                }

                // Create the RAM and place it in the layer.
                rams.push(RAM {
                    dims: w.specs.l0_ram_dims,
                    inputs,
                    outputs,
                    func: Funky(arbgeom_ram_fn),
                    func_parms: RAMParms::ArbGeom {
                        input_to_addend: i2a,
                        overshoot_complement: *overshoot_complement,
                        finish_srcs,
                        carry_lsb,
                        prop_lsb,
                        prop_max,
                    },
                    marker: base_col,   // only for multilayer CarrySkipFinish
                });

                // Save the wiring layer number where the (last) setup RAM
                // outputs to in the "cookie." It's called back when we do
                // multilayer carry-skip, because it corresponds to
                // layer 0 of the input tree.
                *cookie = out_layer;
            },

            Propagator { inputs: cluster_inputs, rightmost } => {
                assert!(*cookie != usize::MAX);

                // A "cluster" here is a single wire (carry) or
                // two wires (propagate & carry) representing a
                // CarrySkipSetup stage or an aggregation thereof
                // via a Propagator. This Propagator being created
                // has one fewer clusters output than input. A
                // rightmost Propagator has one bit per output
                // cluster, otherwise two bits per. Therefore:
                let n_out_bits_per = 2 - *rightmost as usize;
                let n_out_clusters = cluster_inputs.len() - 1;
                let _n_bits_out = n_out_clusters * n_out_bits_per;

                // Here are the pseudo-pvs for where to stack the
                // carry and propagate aggregates from this Propagator.
                let prop_pv = next_carry_prop_pv;
                let carry_pv = prop_pv + 1;
                next_carry_prop_pv += 2;

                // Reserve space.                
                w.reserve_for_node(out_layer, prop_pv);
                w.reserve_for_node(out_layer, carry_pv);

                // Compute the nodes (indices in w.bits), identifying
                // them as RAM output bits.
                let mut outputs = Vec::new();
                for _ in 0..n_out_clusters {
                    let new_index = w.add_node_fast(out_layer, carry_pv);
                    outputs.push(new_index);
                    if *rightmost {
                        continue;
                    }
                    let new_index = w.add_node_fast(out_layer, prop_pv);
                    outputs.push(new_index);
                }

                // Go find the input bits from other RAM outputs.
                let mut bit_inputs = Vec::new();
                for (layer_ix, stage_ix, inst_ix) in cluster_inputs {

                    // Determine the actual *wiring* (not RAM) layer.
                    let layer_ix = layer_ix + *cookie;

                    // Find the pseudo-pvs for locating the bits.
                    let prop_pv = w.product_width + 2 * stage_ix;
                    let carry_pv = prop_pv + 1;

                    // Find and append the requested carry bit to the inputs.
                    let carry_ix = w.from_lpi[layer_ix][carry_pv][*inst_ix];
                    bit_inputs.push(carry_ix);

                    // If this stage the input bits are coming from
                    // is rightmost, it doesn't and can't have any
                    // propagate bits.
                    if *stage_ix == 0 {
                        continue;
                    }

                    // Find and append the requested prop bit to the inputs.
                    let prop_ix = w.from_lpi[layer_ix][prop_pv][*inst_ix];
                    bit_inputs.push(prop_ix);
                }

                rams.push(RAM {
                    dims: w.specs.l0_ram_dims,
                    inputs: bit_inputs,
                    outputs,
                    func: Funky(propagator_ram_fn),
                    func_parms: RAMParms::Propagator {
                        n_stages: cluster_inputs.len(),
                        rightmost: *rightmost,
                    },
                    marker: 0,
                });
            },

            Joiner { stage, inputs } => {
//p!(y "J stage {} inputs {:?}", stage, inputs);

                // `stage` indicates which setup RAM is participating.
                // It's indexed from 1, because setup RAM 0 is final and
                // and doesn't need to roll the odometer for carries.
                //
                // `inputs` indicates prop/carry pairs that require
                // combining. inputs[0] has no prop. It's a right-associative
                // tree that looks like this: if inputs = A B C D E, that
                // means ((((A join B) join C) join D) join E). I know this
                // looks left associative, but I mean right associative with
                // respect to stage place value. A is inputs[0], which is the
                // least significant stage, which would be rightmost if I
                // draw a picture of that.
                //
                // Note in the `join` operations, HIGHER
                // (layer_ix, stage_ix, inst_ix) corresponds to LESS
                // significance. So inputs[0] is at the highest layer.
                //
                // FIXME 22Aug2022 Because inputs[0] is available last, we
                // need to go left-associative so we use it last. This
                // increases complexity, because we get more cases and some
                // inputs/outputs with propagate and some without.
                //
                // For the `join` operator for this discussion,
                // the left argument of `join` is just a carry bit, and the
                // right operand is a carry and propagate bit. The result
                // is just a carry bit. So `join` changes 3 bits to 1 bit.

                // inputs[0] ultimately determines an index in w.bits, which we
                // here compute as `decision_ix`.
                let cookie = *cookie;

                // I panicked a little after I went to the bother of making
                // this work left-to-right, because the multipliers were no
                // faster! Turns out, they are *sporadically* faster, starting
                // at 213u x 213u:
                //
                // right-to-left: (14, 8819, 11658888, 77)
                // left-to-right: (13, 8819, 11659944, 77) -- faster!
                //
                // I feel better about my work now! 22Aug2022
                let decision_ix = if true {     // MWA Leave this 'true'.
                    // LEFT-TO-RIGHT (I *thought* this would work better.)
                    let rmost = inputs.len() == 1;
                    let lsi = inputs.last().unwrap();
                    let mut decision_ix = prop_carry_ix(w, cookie, rmost, *lsi);
                    let mut decision_layer = lsi.0 + cookie - 1;
                    for i in (0..(inputs.len().saturating_sub(1))).rev() {
                        let lsi = inputs[i];
                        let right_ix = prop_carry_ix(w, cookie, i == 0, lsi);
                        decision_layer = 1 + decision_layer.max(lsi.0 + cookie);

                        let (ram, new_ix) = prop_carry_compose
                                (w, decision_ix, right_ix, joiner_map);
                        decision_ix = new_ix;

                        if let Some(ram) = ram {
                            while w.ram_layers.len() <= decision_layer {
                                w.ram_layers.push(Vec::new());
                            }
                            w.ram_layers[decision_layer].push(ram);
                        }
                    }
                    decision_ix
                } else {
                    // RIGHT-TO-LEFT (Original version was easier to code.)
                    let mut decision_ix =
                            prop_carry_ix(w, cookie, true, inputs[0]);
                    let mut decision_layer = inputs[0].0 + cookie - 1;
                    for lsi in &inputs[1..] {
                        let left_ix = prop_carry_ix(w, cookie, false, *lsi);
                        decision_layer = 1 + decision_layer.max(lsi.0 + cookie);

                        let (ram, new_ix) = prop_carry_compose
                                (w, left_ix, decision_ix, joiner_map);
                        decision_ix = new_ix;

                        if let Some(ram) = ram {
                            while w.ram_layers.len() <= decision_layer {
                                w.ram_layers.push(Vec::new());
                            }
                            w.ram_layers[decision_layer].push(ram);
                        }
                    }
                    decision_ix
                };

/*              // ORIGINAL COMMENTS -- AN UPDATE ABOVE WOULD BE NICE!
                // `decision_ix` will be the carry decision to the
                // CarrySkipFinish RAM for this stage. But the decision
                // isn't final until we exhaust `inputs` and build out
                // joiner RAMs. FIXME Doc that this is now a prop-carry pair.
                let mut decision_ix = prop_carry_ix(w, cookie, true, inputs[0]);

                // `decision_layer` is the *RAM* layer where `decision_ix`
                // comes from. We'll increase this monotonically so that
                // joiner RAMs are correctly seqeuential in layers so that
                // there are no timing violations.
                let mut decision_layer = inputs[0].0 + cookie - 1;

                // inputs[0] always exists, so we were safe thus far.
                // If there are more inputs, we have to do joins to obtain
                // the ultimate carry decision. We don't check that
                // inputs[0] had no propagate, and all other inputs do,
                // but we rely on that here.
                for lsi in &inputs[1..] {
//p!(B "  next: ({}, {}, {})", layer_ix, stage_ix, inst_ix);

                    // `decision_ix`, `prop_ix`, and `carry_ix` are inputs
                    // for a new joiner RAM. We need it to be one layer later
                    // than the latest input came from. That's also our new
                    // decision_layer. FIXME Redoc here.
                    let left_ix = prop_carry_ix(w, cookie, false, *lsi);
                    decision_layer = 1 + decision_layer.max(lsi.0 + cookie);

                    let (ram, new_ix) =
                            prop_carry_compose(w, left_ix, decision_ix);
                    decision_ix = new_ix;

                    // Insert this RAM at the appropriate layer.
                    // This process is very different from the non-Joiner
                    // RAM additions, which are always to a new layer
                    // being attended. But this layer may be to an appending
                    // or a previous layer.
                    while w.ram_layers.len() <= decision_layer {
                        w.ram_layers.push(Vec::new());
                    }
                    w.ram_layers[decision_layer].push(ram);
                }
*/

                // Now that the Propagator and Joiner trees are finished,
                // we append the CarrySkipFinish for this stage.

                // CarrySkipFinish RAMs will be in a single layer, beneath
                // the layer with the Joiner tree root. That was the layer
                // with the PackageKind::Joiner entries. Its output wiring
                // layer is numerically the same as the CarrySkipFinish RAM
                // layer.
                let finish_layer = out_layer + 25;   // FIXME just out_layer 15July

                // `finish_layer` is a RAM layer. Its corresponding output
                // wiring layer is one higher. Shadow.
                let out_layer = 1 + finish_layer;

                // The main CarrySkipFinish inputs are the corresponding
                // CarrySkipSetup addend outputs. That DOESN'T include prop
                // and carry outputs, and since the setup isn't rightmost
                // and only supports one carry bit for our scheme, there is
                // is precisely one prop and one carry. We determine:
                let setup_outs = &w.ram_layers[cookie - 1][*stage].outputs;
                let stage_width = setup_outs.len() - 2;
                let mut inputs = setup_outs[0..stage_width].to_owned();

                let base_col = w.ram_layers[cookie - 1][*stage].marker;
                assert!(base_col != 0);

                // Encode where the carry decision appears in the
                // CarrySkipFinish inputs, and that it's one bit wide.
                let finish_srcs = vec![(stage_width as u64, 1)];

                // Create outputs for the CarrySkipFinish and connect these
                // to the wiring.
                let mut outputs = Vec::new();
                for i in 0..stage_width {
                    let new_index = w.add_node(out_layer, base_col + i);
                    outputs.push(new_index);
                }
                let stage_widths = vec![stage_width as u64];

                // Include the carry decision as the final input.
                inputs.push(decision_ix.1);

                // Here's our finish RAM.
                let ram = RAM {
                    dims: w.specs.l0_ram_dims,
                    inputs,
                    outputs,
                    func: Funky(carry_skip_finish_ram_fn),
                    func_parms: RAMParms::CarrySkipFinish {
                        finish_srcs,
                        n_finishes_so_far: 0,
                        stage_widths,
                    },
                    marker: base_col,
                };

                // Place the RAM where it belongs.
                while w.ram_layers.len() <= finish_layer {
                    w.ram_layers.push(Vec::new());
                }
                w.ram_layers[finish_layer].push(ram);

                // The "regular" CarrySkipFinish code is not to do anything,
                // because we've already generated this RAM.
                *nofinish = true;
            },

            _ => {
                panic!("not yet supported: {:?}", p.kind);
            },
        }
    }

    // How we conclude depends whether this was a Joiner layer or
    // some other layer.
    if *nofinish {                      // Joiner

        // The rightmost place value covered by a CarrySkipFinish is not
        // the LSB for the product. Additional bits need carried down from
        // the first CarrySkipSetup, the outputs of which are final, and
        // already-final bits prior to that. Fuse together the necessary nodes.
        let min_pv = w.ram_layers.last().unwrap()[0].marker;
        let from_layer = *cookie;
        let to_layer = w.ram_layers.len();
        for pv in 0..min_pv {
            let bit_ix = w.from_lpi[from_layer][pv][0];
            w.fuse_node(to_layer, pv, bit_ix);
        }

    } else {                            // non-Joiner
        // Fuse nodes together instead of synthesizing wireflops.
        // This modifies w and `from_bits`.
        dont_make_wireflops(w, out_layer, &mut from_bits);

        // We didn't actually attach `rams` to `w` yet -- usually.
        w.ram_layers.push(rams);
    }
}

// Convert PropNodeID to (prop_ix, carry_ix).
fn prop_carry_ix(
    w: &Wires,
    cookie: usize,
    rightmost: bool,
    lsi: PropNodeID,                    // (layer, stage, instance)
) -> (usize, usize) {

    // Determine the actual *wiring* (not RAM) layer.
    let (layer_ix, stage_ix, inst_ix) = lsi;
    let layer_ix = layer_ix + cookie;

    // Find the pseudo-pvs for locating the prop & carry bits.
    // Beware, there are no bits at prop_pv if rightmost.
    let prop_pv = w.product_width + 2 * stage_ix;
    let carry_pv = prop_pv + 1;

    // Find the offset into w.bits for prop and carry. Note offset 0
    // was long-ago taken for subproduct expansion, so I use 0 here
    // to indicate "no prop bit" if this PropNodeID is rightmost.
    let prop_ix = if rightmost { 0 } else {
        w.from_lpi[layer_ix][prop_pv][inst_ix]
    };
    let carry_ix = w.from_lpi[layer_ix][carry_pv][inst_ix];

    (prop_ix, carry_ix)
}

// Combine two (prop_ix, carry_ix) to create a RAM and its 
// output (prop_ix, carry_ix).
fn prop_carry_compose(
    w: &mut Wires,
    hi: (usize, usize),                 // more significant
    lo: (usize, usize),                 // less significant
    joiner_map: &mut Memo,              // reuse existing RAMs if available
) -> (Option<RAM>, (usize, usize)) {

    assert!(hi.0 != 0);                 // The high guy must support propagate.

    // If we've already created this exact same RAM, don't duplicate it.
    if let Some(prop_carry_ix) = joiner_map.get(&(hi, lo)) {
p!(B "Spared a joiner!");
        return (None, *prop_carry_ix);
    }

    let rightmost = lo.0 == 0;
    let inputs = if rightmost {         // RAM inputs
        vec![lo.1, hi.1, hi.0]
    } else {
        vec![lo.1, lo.0, hi.1, hi.0]
    };

    let outputs = if rightmost {        // RAM outputs
        vec![w.alloc_bit_ix()]
    } else {
        vec![w.alloc_bit_ix(), w.alloc_bit_ix()]
    };

    let prop_carry_ix = if rightmost {  // indices of composed outputs
        (0, outputs[0])                 // carry output only
    } else {
        (outputs[1], outputs[0])        // prop and carry output
    };

    // A "Joiner" RAM is a special case of a "Propagator."
    let ram = RAM {
        dims: w.specs.l0_ram_dims,
        inputs,
        outputs,
        func: Funky(propagator_ram_fn),
        func_parms: RAMParms::Propagator {
            n_stages: 2,
            rightmost,
        },
        marker: 0,
    };

    // If given this (hi, lo) again, we already have wires for it
    // and don't need yet another RAM. Remember that.
    //
    // TODO This doesn't actually help when using left-to-right associativity,
    // which produces one fewer layer, and doesn't help much with right-to-left.
    // If this optimization is to ever work, a mixed associativity scheme that
    // appends the rightmost term last but does everything else right-to-left
    // would be the next thing to consider. 25Aug2022
    joiner_map.insert((hi, lo), prop_carry_ix);

    (Some(ram), prop_carry_ix)
}
