///! tally.rs March 2022
///! A `tally` is an unsigned bit vector of unlimited width that
///! we can increment at any place value. A principal use is managing
///! the multiplier's total additive offset.

use crate::*;

#[derive(Clone)]
pub struct Tally {
    pub bits: Vec<bool>,
}

impl Tally {

    /// Create a tally with value 0.
    pub fn new() -> Tally {
        Tally {
            bits: Vec::new(),
        }
    }

    /// Add 2**pv to this tally.
    pub fn add_pv(&mut self, mut pv: usize) {

        loop {
            // Ensure enough bits in `bits` to support this place value.
            while self.bits.len() <= pv {
                self.bits.push(false);
            }

            // If this place value is already set, we need a carry.
            if self.bits[pv] {
                self.bits[pv] = false;
                pv += 1;
                continue;
            }

            // Otherwise, set the place value.
            self.bits[pv] = true;
            break;
        }
    }

    /// Compute the additive complement of a tally modulo 2**`width`.
    pub fn complement(&mut self, width: usize) -> Tally {

        // There shouldn't be any tally bits beyond the indicated width.
        if self.bits.len() > width {
            fail("Tally complement does not make sense.");
        }

        // Copy `bits` to future complement with specified width.
        let mut c = self.bits.clone();
        c.resize(width, false);

        // Complement `c` and wrap in a Tally.
        let mut once_true = false;
        for bit in c.iter_mut() {
            *bit = *bit != once_true;
            once_true = once_true || *bit;
        }
        Tally { bits: c }
    }

    /// Display a tally for diagnostic purposes.
    fn show(&self, zero: &str, one: &str, prefix: &str, postfix: &str) {
        pp!("{}", prefix);
        for b in self.bits.iter().rev() {
            pp!("{}", if *b { one } else { zero });
        }
        pp!("{}", postfix);
    }
}

/// function to spot-check Tally struct
pub fn test_tally() {
    p!(Y "test_tally output:\n");

    let mut sum = Tally::new();

    // `sum` will be 66666 = 0b10000010001101010.
    // This means we need at least 17 bits complemented.
    for v in &[12345, 54321] {
        for i in 0..16 {
            if v & 1 << i != 0 {
                sum.add_pv(i);
            }
        }
    }

    sum.show("0", "1", "sum         0b", "\n");
    let comp = sum.complement(24);
    comp.show("0", "1", "comp 0b", "\n");
    p!();
}
