///! tests.rs April 2022
///! This module contains "established tests" that I would rather not
///! have cluttering up main.rs.

use crate::*;

/// Given an already-built multiplier, try some number of test multiplications
/// with random inputs. For Either signage, the signage is randomized.
/// Returns (something_failed, n_tests_right, n_tests_wrong).
pub fn test_mul(w: &mut Wires, iters: usize) -> (bool, u64, u64) {

    if false {                          // diagnostic
        for i in 5..w.ram_layers.len() {
            p!(R "****** LAYER {} ******", i);
            for ram in &w.ram_layers[i] {
                p!("{:?}", ram.func_parms);
            }
        }
    }

    let mut rng = Urand64::new();
    let mut trouble = false;
    let mut n_right = 0;
    let mut n_wrong = 0;

    for _ in 0..iters {

        // Factors with `Either` signage will need their control decoder
        // signages assigned at random. Here are some bits for that.
        let rand_bits = rng.next();

        let l_is_signed = match w.specs.sue_l {
            Unsigned => false,
            Signed => true,
            Either => rand_bits & 1 != 0,
        };

        let r_is_signed = match w.specs.sue_r {
            Unsigned => false,
            Signed => true,
            Either => rand_bits & 2 != 0,
        };

        // Come up with some numbers to multiply.
        let l_factor = BigRand::new(l_is_signed, w.specs.width_l, &mut rng);
        let r_factor = BigRand::new(r_is_signed, w.specs.width_r, &mut rng);
        let product_wanted = l_factor.multiply(&r_factor);

        // Copy the right and left BigRand factors into multiplier nodes.
        // This happens one bit at a time.
        for i in 0..w.specs.width_r {
            let bit_index = w.from_lpi[0][w.r_pvs.start + i][0];
            w.bits[bit_index] = r_factor.read_bit(i);
        }

        for i in 0..w.specs.width_l {
            let bit_index = w.from_lpi[0][w.l_pvs.start + i][0];
            w.bits[bit_index] = l_factor.read_bit(i);
        }

        // Connect right and left signage, if any.
        if let Some(pv) = w.r_signage_pv {
            let bit_index = w.from_lpi[0][pv][0];
            w.bits[bit_index] = r_is_signed;
        }

        if let Some(pv) = w.l_signage_pv {
            let bit_index = w.from_lpi[0][pv][0];
            w.bits[bit_index] = l_is_signed;
        }

        // Multiply by actuating RAMs in order.
        w.simulate();

        // What's the sum from the last RAM layer?
        let sum = w.final_sign_extended_layer_sum(product_wanted.is_signed);

        // Verify product length. It's not likely to miss.
        assert_eq!(sum.bits.len(), product_wanted.n_bits);

        // Verify product.
        let mut ok = true;
        for i in 0..product_wanted.n_bits {
            if sum.bits[i] != product_wanted.read_bit(i) {
                ok = false;
            }
        }

        if ok {
            n_right += 1;
        } else {
            n_wrong += 1;
            trouble = true;
            if true {                  // Diagnostic hook.
                let got = BigRand::from(&sum);
                product_wanted.print_bin("want: ", "\n");
                got.print_bin(" got: ", "\n\n");
            }
        }
    }

    (trouble, n_right, n_wrong)
}

/// This function constructs a variety of multipliers, and then
/// runs `n_test_muls` trial multiplications to validate that the
/// correct products are being generated.
///
/// `lo_ram_dims` provides the number of address and I/O bits in the
/// subproduct RAMs.
///
/// `n_test_muls` provides the number of random-number multiplications
/// to test on each config. There is no output on success, but the
/// program will panic if the multiplier's answers do not match what
/// the BigRand module thinks those answers should be. You may use 0
/// if you only want to build the configs, but not try any test muls.
///
/// `sue_ls` and `sue_rs` are strings containing 's', 'u', and 'e' to
/// indicate signages to construct for the left and right factors.
/// If `sue_rs` is empty, `sue_ls` will be applied to both factors.
/// If `sue_ls` is empty, you won't get any configs. If characters
/// are repeated in a string, you will get repeated configs.
///
/// `width_ls` and `width_rs` are (informally speaking) ranges of
/// factor widths to config. Both must use Rust's .. notation, but
/// their interpretation is unique to this module:
///
/// a) Upper bounds are *inclusive*, even though you don't use =.
///
/// b) If `width_ls` is `..`, you won't get any configs.
///
/// c) If `width_rs` is `..`, `width_ls` will also be used as `width_rs`
///    for all configurations. Note this produces many configs fewer than
///    if you explicitly supply both width ranges, because the `..` case
///    excludes unequal factor widths from the generated configs.
///
/// d) If an upper bound is supplied without a lower bound, 1 will be
///    used as the lower bound.
///
/// e) If a lower bound is supplied without an upper bound, the lower
///    bound will also be used as the upper bound, meaning that factor
///    will be configured with the same width for all tests.
///
/// f) If both bounds are supplied in a range, that range is inclusive.
///    It doesn't make sense to support an empty range, as no configs
///    would result.
///
/// g) If both bounds are supplied and the lower bound is greater,
///    no configs will be generated.
///
pub fn try_configs(
    width_ls: impl std::ops::RangeBounds<usize>,
    width_rs: impl std::ops::RangeBounds<usize>,
    sue_ls: &str,
    sue_rs: &str,
    l0_ram_dims: (usize, usize),
    scope: &Scope,
) {
    use std::ops::{RangeBounds, Bound::*};

    fn range(r: impl RangeBounds<usize>) -> (usize, usize) {
        match (r.start_bound(), r.end_bound()) {
            (Included(s), Excluded(e)) if *s > 0 => (*s, *e + 1),
            (Included(s), Unbounded) if *s > 0 => (*s, *s + 1),
            (Unbounded, Excluded(e)) => (1, *e + 1),
            (Unbounded, Unbounded) => (0, 0),
            _ => panic!("unsupported bounds combination"),
        }
    }

    fn sues(sue: &str) -> Vec<Signage> {
        let mut v = Vec::new();
        for c in sue.chars() {
            v.push(match c {
                's' => Signed,
                'u' => Unsigned,
                'e' => Either,
                _ => panic!("specify signage as 's', 'u', or 'e'"),
            });
        }
        v
    }

    let mut tasker = Tasker::with_threads(N_THREADS);

    let width_ls = range(width_ls);
    if width_ls.0 == 0 {
        return;
    }
    let width_rs = range(width_rs);
    let sue_rs = sues(if sue_rs == "" { sue_ls } else { sue_rs });
    let sue_ls = sues(sue_ls);

    // `mul_width` here is defined to be the sum of the input widths,
    // which is typically *but not always* the multiplier output width.
    let max_mul_width = width_ls.1 + if width_rs.0 == 0
            { width_ls.1 } else { width_rs.1 };
    for mul_width in 2..=max_mul_width {
        for width_l in width_ls.0..width_ls.1 {
            let width_rs = if width_rs.0 == 0 {
                (width_l, width_l + 1)
            } else { width_rs };
            for width_r in width_rs.0..width_rs.1 {
                if width_l + width_r != mul_width {
                    continue;
                }
                for sue_l in &sue_ls {
                    for sue_r in &sue_rs {
                        let specs = Specs {
                            width_l, width_r, l0_ram_dims,
                            sue_l: *sue_l, sue_r: *sue_r,
                        };
                        tasker.go(&specs, scope);
                    }
                }
            }
        }
    }
}

/// Documentation purpose. The design range this program covers may be
/// seriously reduced if usize is less than 64.
pub fn test_usize() {
    assert_eq!(usize::BITS, 64);
}

/// FIXME This should not be distributed, but only because it won't be
/// useful to most, and I haven't documented anything about it.
/// Also that my batch files aren't included here.
/// I used this for some cluster test runs with Slurm.
pub fn bender_job() {
    let sue = environ_string("LMUL_SUE", "");
    let dims = environ_usize("SLURM_ARRAY_TASK_ID", 0);
    if sue.is_empty() || dims == 0 {
        fail("Please humor me and set LMUL_SUE and SLURM_ARRAY_TASK_ID.");
    }
    if N_THREADS != 16 {
        fail("Pretty please run this with 16 threads on bender.");
    }

    let scope = Scope {
        n_test_muls: 1000,
        draw: Best,
        synth: Best,
        test: Best,
        abandon_overcost: true,
        d_flip_flops_avail: true,
        strategy: OptimizeAll,
        join_reservation: 0,
    };

    p!("Big run: dims ({}, {}), sue {}, scope {:?}", dims, dims, sue, scope);

    try_configs(120..65535, .., &sue, "", (dims, dims), &scope);
}

/// Test aggregator_ram_fn()
pub fn test_arf() {

    fn arf_print(bits: u64, n_stages: usize) {
        for stage in (0..n_stages).rev() {
            let c = match bits >> 2 * stage & 3 {
                0 => '-',
                1 => 'g',
                2 => 'p',
                _ => '?',               // errror
            };
            pp!("{}", c);
        }
    }

    for rightmost in [false, true] {
        p!(G "testing with rightmost {}", rightmost);
        for n_stages in 1..8 {
            let parms = RAMParms::Propagator { n_stages, rightmost };
            let n_tests = 3_u64.pow(n_stages as u32);
            for test in 0..n_tests {
                let mut i = 0;              // input bits to test
                let mut convert = test;
                for pos in 0..n_stages {
                    i |= convert % 3 << 2 * pos;
                    convert /= 3;
                }

                // If rightmost, input cannot be propagate.
                if rightmost && i & 2 != 0 {
                    continue;
                }

                arf_print(i, n_stages);
                pp!(" -> ");


                // If rightmost, there is no p_0 input bit. Delete bit pos 1.
                if rightmost {
                    i = i >> 1 & !1 | i & 1;
                }

                let o = propagator_ram_fn(i, &parms);
                arf_print(o, n_stages - 1);
                p!();
            }
        }
    }
}

// Test special_case_add() from BigRand.
pub fn test_special_case_add() {
    let mut rng = Urand64::new();
    let width = 160;

    // random number check
    p!(B "Verify with Python please:\n");
    for _ in 0..8 {
        let a = BigRand::new(false, width, &mut rng);
        let b = BigRand::new(false, width, &mut rng);
        let c = a.special_case_add(&b);
        a.print_hex("  0x", "\n");
        b.print_hex("+ 0x", "\n");
        c.print_hex("= 0x", "\n\n");
    }

    // extremely basic carry check
    let mut a = BigRand::zero(false, width);
    let mut b = BigRand::zero(false, width);
    for i in 0..width - 1 {
        a.set_bit(i, true);
        b.set_bit(i, true);
    }
    let c = a.special_case_add(&b);
    p!(B "Verify visually please:\n");
    a.print_hex("  0x", "\n");
    b.print_hex("+ 0x", "\n");
    c.print_hex("= 0x", "\n\n");
}

/// Create a large addition spec with 2 bits per place value,
/// synthesize a solution, and test it. This will be used to
/// check multi-layer carry-skip synthesis, which is difficult to
/// test thoroughly in real multipliers because of how rarely
/// many-layer carry skips occur.
pub fn test_many_stage(width: usize, n_test_muls: usize) {

    // Here is a means to check this test jig until code to
    // synthesize many-stage adders is available.
    const MOCKUP: bool = false;

    // struct Wires is intended for synthesizing multipliers.
    // But here we only want an adder. So we kind of end-run
    // around our usual manner of instantiating one.
    let heading = ReportHeading {
        n_l_parts: 0,
        n_r_parts: 0,
        est_n_subprod_bits: 0,
    };

    let specs = Specs {
        width_l: 0,
        width_r: 0,
        l0_ram_dims: (0, 0),
        sue_l: Unsigned,
        sue_r: Unsigned,
    };

    let mut w = Wires {
        specs,
        product_width: width,
        heading,
        bits: Vec::new(),
        from_lpi: Vec::new(),
        ram_layers: Vec::new(),
        l_pvs: 0..0,
        r_pvs: 0..0,
        l_signage_pv: None,
        r_signage_pv: None,
        overshoot: Tally::new(),
        l0_trunc_to: 0,
        l0_cost: (0, 0, 0, 0),
        l0_backup: Vec::new(),
        l0_dirty: false,
        n_compos: 0,
    };

    // Provide two addends.
    for pv in 0..width {
        w.add_node(0, pv);
        w.add_node(0, pv);
    }

    if MOCKUP {
        // Provide a sum.
        for pv in 0..width {
            w.add_node(1, pv);
        }
    }

    // ---------------------------------------------------
    // FIXME TODO Implement the adder you're testing here.
    // ---------------------------------------------------

    // Simulate some test additions.
    let mut n_correct = 0;
    let mut n_wrong = 0;
    let mut rng = Urand64::new();
    let last = w.from_lpi.len() - 1;

    for _ in 0..n_test_muls {

        // Put the addition problem into the wiring.
        let a = BigRand::new(false, width, &mut rng);
        let b = BigRand::new(false, width, &mut rng);
        for i in 0..width {
            w.bits[w.from_lpi[0][i][0]] = a.read_bit(i);
            w.bits[w.from_lpi[0][i][1]] = b.read_bit(i);
        }

        // Compute the solution by simulating the circuit.
        w.simulate();

        if MOCKUP {
            let mut carry = 0;
            for i in 0..width {
                let a = w.bits[w.from_lpi[0][i][0]] as u32;
                let b = w.bits[w.from_lpi[0][i][1]] as u32;
                let mut c = a + b + carry;
                carry = c >> 1;
                c = c & 1;
                w.bits[w.from_lpi[last][i][0]] = c != 0;
            }
        }

        // Compute the desired solution, and see if we reached it.
        let c = a.special_case_add(&b);
        let mut ok = true;
        for i in 0..width {
            if w.bits[w.from_lpi[last][i][0]] != c.read_bit(i) {
                ok = false;
                break;
            }
        }

        if ok {
            n_correct += 1;
        } else {
            n_wrong += 1;
        }
    }

    p!("{} good and " R "{} bad" " test additions", n_correct, n_wrong);
}
