///! util.rs April 2022
///! This succinct module contains (non-macro) utility functions.

/// End program with an error message. It's distinct from panic!
/// in that it's a usage error, not a programmer or logic error.
pub fn fail(s: &str) -> ! {
    p!("{}", s);
    std::process::exit(1);
}

/// Let user press Enter before continuing.
pub fn pause() {
    use std::io::Write as _;
    let mut ignore = String::new();
    let _ = std::io::stdout().flush();
    let _ = std::io::stdin().read_line(&mut ignore);
}

/// How many bits are needed to represent some value `n`?
pub trait HasBitLength {
    fn bit_length(self) -> usize;       // min. # bits to represent this int
    fn smo(self) -> Self;
}

impl HasBitLength for usize {
    fn bit_length(self) -> usize {
        (usize::BITS - self.leading_zeros()) as usize
    }

    /// "So many ones" converts 5 into 0b11111, etc. Saturates.
    fn smo(self) -> Self {
        if self >= usize::BITS as usize {
            usize::MAX
        } else {
            (1 << self) - 1
        }
    }
}

/// Fetch a numeric variable from the environment, or supply a default.
pub fn environ_usize(key: &str, def: usize) -> usize {
    match std::env::var(key) {
        Ok(s) => {
            match s.parse() {
                Ok(n) => n,
                Err(_) => {
                    p!(e "failed parsing {}={}, so using {}", key, s, def);
                    def
                },
            }
        },
        Err(_) => {
            p!(e "no env var {}, using default {}", key, def);
            def
        }
    }
}

/// Fetch a string variable from the environment, or supply a default.
pub fn environ_string(key: &str, def: &str) -> String {
    match std::env::var(key) {
        Ok(s) => s,
        Err(_) => def.to_string(),
    }
}




