///! wires.rs February 2022
///! This module models the connections between RAMs.

// This diagnostic const can increase the number of overshoot complement
// bits the adder stage needs to incorporate. We do this in order to
// purposely strain the adder when testing for correct processing of
// thse bits. But for normal use of this program, have this const false.
const DIAGNOSTIC_AVOID_WIDENS: bool = false;

use crate::*;

/// `Wires` represents the set of all circuit nodes within a multiplier.
/// Every node has a layer `l`, a place value `p`, and (because there may
/// be multiple nodes at a given layer and place value), an instance `i`.
/// The term `lpi` refers to such a triple to identify a particular node.
///
/// This struct also holds the RAMs that connect its wires. At present
/// they are stratified by layer.
#[derive(Clone)]
pub struct Wires {

    // ------------------------------------------------------------
    // The following fields DO NOT CHANGE during recursive searches
    // for the best multiplier circuit.
    // ------------------------------------------------------------

    /// `specs` holds design parameters used to build the multiplier.
    pub specs: Specs,

    /// `product_width` is the output width of the multiplier.
    ///
    /// From wiring layer 1 and up, there will always be at least
    /// one bit for each place value 0..product_width.
    ///
    /// We also need nodes for propagates and carries. The custom is
    /// to stack these as instances; e.g., if the rightmost setup
    /// RAM produces 3 carry bits, they're at pv == product_width
    /// and ordered from LSB to MSB. The next setup RAM's propagate
    /// and carry bits stack at product_width + 1 and product_width + 2
    /// respectively, and so on for further setup RAMs. So it's very
    /// easy to know physically where these bits appear in the wiring.
    pub product_width: usize,

    /// Here are the number of left and right subwords,
    /// and a starting point for the number of subproduct bits.
    pub heading: ReportHeading,

    /// `l_pvs` and `r_pvs` give the place values at level 0 where
    /// multiplicands are introduced. To get a `bits` index, we must
    /// still look these up via `from_lpi`. Likewise, `l_signage_pv`
    /// and `r_signage_pv` give place values at level 0 where the
    /// control decoder specifies signage, but only if the operand
    /// supports `Either` signage, and still must look up via `from_lpi`.
    pub l_pvs: std::ops::Range<usize>,
    pub r_pvs: std::ops::Range<usize>,
    pub l_signage_pv: Option<usize>,
    pub r_signage_pv: Option<usize>,

    /// `overshoot` is the fixed additive error introduced by the
    /// multiplier in order to support signed arithmetic.
    pub overshoot: Tally,

    /// `l0_trunc_to` is where to cut `bits` to when we shed
    /// subproduct adder layers in order to test other adder configurations.
    pub l0_trunc_to: usize,

    /// `l0_cost` is the cost of the subproduct expansion RAMs.
    pub l0_cost: SolutionCost,

    /// `l0_backup` and `l0_dirty` let us restore the subproduct RAM
    /// layer for backtracking if it's been altered by composition.
    pub l0_backup: Vec<RAM>,
    pub l0_dirty: bool,

    /// `n_compos` is managed by compos.rs and indicates how many RAMs
    /// were eliminated by composition.
    pub n_compos: usize,

    // ------------------------------------------------------------
    // FIXME The below comment is no longer factual, but I would
    //       like something said about these fields.
    // The following fields GROW AND CONTRACT during recursive
    // searches for the best multiplier circuit.
    // ------------------------------------------------------------

    /// `bits` is a flat vector of all nodes within the multiplier.
    /// All RAM I/O is via this vector, so that simulations run reasonably
    /// fast. (Unsafe Rust may be faster, but we won't.) The index within
    /// `bits` can be encoded by `from_lpi`.
    pub bits: Vec<bool>,

    /// Convert lpi to an index within `bits` via this uneven vector nest.
    /// Access looks like: from_lpi[layer_num][pv][instance].
    pub from_lpi: Vec<Vec<Vec<usize>>>,

    /// RAMs by layer.
    pub ram_layers: Vec<Vec<RAM>>,
}

impl Wires {

    /// Create the wiring nodes for a new multiplier.
    /// XXX This only creates the input nodes where the factors
    ///     and control decoder signages are applied. There may
    ///     be naming and refactoring work ahead.
    pub fn new(s: &Specs) -> Wires {

        // Determine the width of the final product.
        let product_width = predict_product_width(
                s.width_l, s.width_r, s.sue_l, s.sue_r);

        // Fill in where the multiplicand inputs will be.
        let mut top = s.width_l + s.width_r;
        let r_pvs = 0..s.width_r;
        let l_pvs = s.width_r..top;

        // Fill in the left and right signage input location
        // from the control decoder, when applicable.
        let mut r_signage_pv = None;
        if s.sue_r == Either {
            r_signage_pv = Some(top);
            top += 1;
        }

        let mut l_signage_pv = None;
        if s.sue_l == Either {
            l_signage_pv = Some(top);
            top += 1;
        }

        let heading = ReportHeading {
            n_l_parts: 0,
            n_r_parts: 0,
            est_n_subprod_bits: 0,
        };

        // Instantiate the struct with its specs.
        let mut w = Wires {
            specs: *s,
            product_width,
            heading,
            bits: Vec::new(),
            from_lpi: Vec::new(),
            ram_layers: Vec::new(),
            l_pvs,
            r_pvs,
            l_signage_pv,
            r_signage_pv,
            overshoot: Tally::new(),
            l0_trunc_to: 0,
            l0_cost: (0, 0, 0, 0),
            l0_backup: Vec::new(),
            l0_dirty: false,
            n_compos: 0,
        };

        // Add one node for each incoming factor bit and signage.
        for i in 0..top {
            w.add_node(0, i);
        }

        w
    }

    /// Simulate the multiplier by actuating all RAMs.
    pub fn simulate(&mut self) {
        for layer in &self.ram_layers {
            for ram in layer {
                ram.simulate(&mut self.bits);
            }
        }
    }

    /// Part 1 of add_node():
    /// Given a layer and place value, make sure from_lpi
    /// has at least an empty list of indices there.
    pub fn reserve_for_node(&mut self, layer: usize, pv: usize) {

        // Force the layer to exist and fetch it.
        while self.from_lpi.len() <= layer {
            self.from_lpi.push(Vec::new());
        }
        let lay = &mut self.from_lpi[layer];

        // Force the place value to exist and fetch it.
        while lay.len() <= pv {
            lay.push(Vec::new());
        }
    }

    /// Part 2 of add_node():
    /// Add a node instance to an existing layer and place value,
    /// and return its location within `bits`.
    pub fn add_node_fast(&mut self, layer: usize, pv: usize) -> usize {

        // Add a node for this place value, noting its index in `bits`.
        let index = self.bits.len();
        self.bits.push(false);

        // Map from lpi form to `index`.
        self.from_lpi[layer][pv].push(index);

        // Return `index` within `bits`.
        index
    }

    /// Given a layer and place value, add a node that can hold
    /// a bit and return its location within `bits`. This function
    /// does not exist the layer and place value exist yet. If you
    /// know they exist, you can call add_node_fast() instead.
    pub fn add_node(&mut self, layer: usize, pv: usize) -> usize {
        self.reserve_for_node(layer, pv);
        self.add_node_fast(layer, pv)
    }

    // Given a layer, place value, and bit index of an existing node,
    // include that node in a new place value for the layer.
    // This is like add_node, except that this node is fused across
    // layers. This can be used in lieu of a wireflop for simulation.
    pub fn fuse_node(&mut self, layer: usize, pv: usize, bit_index: usize) {
        self.reserve_for_node(layer, pv);
        self.from_lpi[layer][pv].push(bit_index);
    }

    /// After new() is called, the Wires have a layer of nodes for the
    /// factors and (if any) control decoder signages. This function sets
    /// down the subproduct RAMs, connects those to the first layer of
    /// nodes, then creates and connects a second layer of nodes with
    /// the subproduct outputs. Returns a list of overshoot complement
    /// pvs that haven't been included in the subproduct.
    ///
    ///
    /// XXX APPLICABILITY: This has been tested for left and right
    /// factors of 1 through 36 bits in all 36**2 combinations, with all
    /// 9 SUE combinations, for all "RAM sizes" of 4 through 63 bits,
    /// where the L0 RAM  has an equal number of input and output bits.
    /// This test was done 27 April 2022.
    ///
    pub fn lay_subproduct_rams(&mut self) -> Vec<usize>{

        // XXX There is some overlap in logic with split.rs here
        //     that can possibly be factored. But I wish to keep
        //     the high speed of the split.rs tasks.

        // This creates RAM layer 0.
        self.ram_layers.push(Vec::new());

        // number of left and right subwords
        let (n_l_parts, n_r_parts, n_bits) =
                select_subproduct_split(&self.specs);
        self.heading.n_l_parts = n_l_parts;
        self.heading.n_r_parts = n_r_parts;
        self.heading.est_n_subprod_bits = n_bits;

        // [subword_lengths]
        let parts_l = partition_word_size(self.specs.width_l, n_l_parts);
        let parts_r = partition_word_size(self.specs.width_r, n_r_parts);

        // `pv_l` and `pv_r` will accumulate subword offsets within factors.
        // Their sum is the offset within a product.

        // Loop over left subwords, obtaining their lengths and signage.
        // `pv_l` is the current subword's place value within the factor.
        let mut pv_l = 0;
        for (index_l, sw_width_l) in parts_l.iter().enumerate() {
            let signage_l = if index_l + 1 == parts_l.len()
                    { self.specs.sue_l } else { Unsigned };

            // Loop over right subwords, obtaining their lengths and signage.
            // `pv_r` is the current subword's place value within the factor.
            let mut pv_r = 0;
            for (index_r, sw_width_r) in parts_r.iter().enumerate() {
                let signage_r = if index_r + 1 == parts_r.len()
                        { self.specs.sue_r } else { Unsigned };

                // Create a subproduct RAM right here, right now.
                // Once created, its `inputs` and `outputs` are empty
                // vectors, so we have to assemble them in correct order.
                let parms = RAMParms::subproduct_from(
                    *sw_width_l, *sw_width_r,   // widths
                    signage_l, signage_r,       // signages
                    pv_l + pv_r,                // LSB's global place value
                    self.product_width,         // width of final product
                );
                let mut ram = RAM::new
                    (self.specs.l0_ram_dims, Funky(subproduct_ram_fn), parms);

                // *** Connect right factor inputs.

                // Loop over the right factor bits on the RAM.
                for i in 0..*sw_width_r {

                    // Get the corresponding place value in the input nodes.
                    let pv = self.r_pvs.start + pv_r + i;

                    // Get the corresponding index in `bits`.
                    // Note we're on layer 0, and there's just one instance.
                    let bit_index = self.from_lpi[0][pv][0];

                    // That's the next RAM input connection.
                    ram.inputs.push(bit_index);
                }

                // *** Connect left factor inputs.

                // Loop over the left factor bits on the RAM.
                for i in 0..*sw_width_l {

                    // Get the corresponding place value in the input nodes.
                    let pv = self.l_pvs.start + pv_l + i;

                    // Get the corresponding index in `bits`.
                    // Note we're on layer 0, and there's just one instance.
                    let bit_index = self.from_lpi[0][pv][0];

                    // That's the next RAM input connection.
                    ram.inputs.push(bit_index);
                }

                // *** Connect signage(s) from control decoder, if any.

                if signage_r == Either {
                    let pv = self.r_signage_pv.unwrap();
                    let bit_index = self.from_lpi[0][pv][0];
                    ram.inputs.push(bit_index);
                }

                if signage_l == Either {
                    let pv = self.l_signage_pv.unwrap();
                    let bit_index = self.from_lpi[0][pv][0];
                    ram.inputs.push(bit_index);
                }

                // *** Connect outputs.

                if let RAMParms::Subproduct {
                    width, offset_used, ..
                } = ram.func_parms {

                    // Loop over the output bits on the RAM.
                    for i in 0..width {

                        // Get corresponding place value of output nodes.
                        let pv = pv_r + pv_l + i;

                        // Although some RAMs can produce place values
                        // beyond the width of the full multiplier product,
                        // they are superfluous and undesirable.
                        if pv >= self.product_width {
                            // This can cause the number of subproduct bits
                            // to be fewer than estimated within
                            // select_subproduct_split().
                            continue;
                        }

                        // Get the corresponding index in `bits`.
                        // This requires creating a new node in node layer 1.
                        let bit_index = self.add_node(1, pv);

                        // That's the next RAM output connection.
                        ram.outputs.push(bit_index);

                        // If this bit has offset associated, we need to
                        // accumulate its contribution to multiplier overshoot.
                        if offset_used & 1 << i != 0 {
                            self.overshoot.add_pv(pv);
                        }
                    }
                }
                else {
                    fail("Bug: apparently a non-Subproduct");
                }

                // Store this RAM in the multiplier.
                self.ram_layers[0].push(ram);

                // Update right subword offset for next iteration.
                pv_r += *sw_width_r;
            }

            // Update left subword offset for next iteration.
            pv_l += *sw_width_l;
        }

        // TODO comment
        let monoram = n_l_parts * n_r_parts == 1;
        let overshoot_pvs = self.fix_overshoot(monoram);

        // Store enough information to revert to containing only
        // subproduct wiring.
        self.l0_trunc_to = self.bits.len();

        // Record the cost of this layer, to speed up cost accounting.
        self.l0_cost = self.actual_cost();

        // Make a backup copy of this layer for backtracking after composition.
        self.l0_backup = self.ram_layers[0].clone();

        overshoot_pvs
    }

    /// A helper for `lay_subproduct_rams()`, this function adds fixed
    /// offsets to certain RAMs in order to null out `self.overshoot`.
    /// Once this function has been called, adding all subproduct bits
    /// in their correct columns will produce the correct final product
    /// for the multiplier. These further fixed offsets sometimes
    /// require increasing the number of used output bits of a RAM by
    /// one, which correspondingly increases the total number of
    /// subproduct bits. The largest penalty of this kind is a total
    /// of three additional subproduct bits.
    ///
    /// Returns a list of overshoot complement
    /// pvs that haven't been included in the subproduct.
    fn fix_overshoot(&mut self, monoram: bool) -> Vec<usize> {

        // If there is just one RAM, widening is mandatory where needed
        // to include overshoot complement bits. Otherwise, the subproduct
        // adder layers can include these.
        let widen_allowed = (monoram || !DIAGNOSTIC_AVOID_WIDENS) as usize;

        // The offset for this multiplier is fully computed.
        // We need its complement.
        let mut complement = self.overshoot.complement(self.product_width);

        // Here is the unhandled portion of the complement in pv-list form.
        let mut overshoot_pvs = Vec::new();

        // Let's try to undo `complement` early via layer 1.
        let mut count_added_bits = 0;
        let layer = &mut self.ram_layers[0];
        let mut deferred_new_pvs = Vec::new();
        'next: for pv in 0..complement.bits.len() {
            if !complement.bits[pv] {
                continue;
            }
            for widen_first in &[false, true][..=widen_allowed] {
                for ram in layer.iter_mut() {
                    let (offset_removed, just_widened)
                            = ram.try_to_offset(pv, *widen_first);
                    if offset_removed {
                        complement.bits[pv] = false;

                        // If we had to widen the RAM to remove the
                        // offset, that RAM has a new output bit to
                        // feed to the next layer now.
                        if just_widened {
                            // This can cause the number of subproduct bits
                            // to be more than estimated within
                            // select_subproduct_split().
                            count_added_bits += 1;

                            // Need the place value of the new output bit.
                            if let RAMParms::Subproduct {
                                width, global_pv, ..
                            } = ram.func_parms {
                                let new_pv = global_pv + width - 1;

                                // We need to add a wiring node at (1, new_pv),
                                // but right now we're boxed into a corner
                                // by the Rust borrow checker and can't
                                // `self.add_node(1, new_pv);`. But we know
                                // what the return value from this function
                                // is going to be, since it's just extending
                                // `self.bits`. This lets us let the RAM
                                // know where the output will go, and in a
                                // moment we'll have released the `self`
                                // borrow and be able to finish up.
                                let bit_index = self.bits.len()
                                        + deferred_new_pvs.len();
                                ram.outputs.push(bit_index);
                                deferred_new_pvs.push(new_pv);
                            }
                            else {
                                panic!("Not a RAMParms::Subproduct!");
                            }
                        }
                        continue 'next;
                    }
                }
            }

            // Non-square ram_dims like (5, 4) or (18, 16) aren't able to
            // remove all overshoot at layer 0. Overshoot may also remain
            // if this module is configured to not widen (where possible).
            // Keep a list of the complements we still need.
            overshoot_pvs.push(pv);
        }

        // Procedurally we could have added these wiring nodes to the
        // output earlier, but a Rust mutable borrow precluded that.
        // Do that task now.
        for new_pv in deferred_new_pvs {
            self.add_node(1, new_pv);
        }

        // Certain expectations as to how many bits must be added to
        // de-offset can be verified here.
        if count_added_bits != 0 {
            assert!(count_added_bits <= 3);
            if !(
                self.specs.width_l == 1 && self.specs.sue_l == Signed
             || self.specs.width_r == 1 && self.specs.sue_r == Signed
             || self.specs.sue_l == Signed && self.specs.sue_l == Signed
             || self.specs.l0_ram_dims.0 < 8
             || self.specs.l0_ram_dims.1 < 8
            ) {
                panic!("novel case needs documented");
            }
        }

        overshoot_pvs
    }

    /// Tally the final wiring layer's bits and the subproduct complement.
    /// This will be the product of the two factors. This fn is for testing.
    pub fn final_layer_sum(&self) -> Tally {
        let mut o = Tally::new();

        // Access the place values of the last layer.
        // This is a Vec<Vec<usize>> and identifies locations in `self.bits`.
        if let Some(pvs) = self.from_lpi.last() {

            // Loop through place values of this layer. This is `usize`.
            for pv in 0..pvs.len() {

                // Go through the instances at that place value.
                for i in &pvs[pv] {

                    // Pick up the true or false value of that instance,
                    // and add it to the sum we're running.
                    if self.bits[*i] {
                        o.add_pv(pv);
                    }
                }
            }
        }

        // The final tally needs to be the multiplier output width.
        // Depending on the data we have, it may be too narrow or
        // too wide, but the result will be correct once the width
        // is set.
        o.bits.resize(self.product_width, false);
        o
    }

    /// Test code probably doesn't realize that certain m*n-bit multipliers
    /// have output widths less than m*n. This would cause regression tests
    /// to fail even with both product correct. This function is able to
    /// produce full m*n outputs. The caller has to specify whether the
    /// output is signed, because that depends on the test case when a factor
    /// signage is Either. 
    pub fn final_sign_extended_layer_sum(&self, is_signed: bool) -> Tally {
        let mut o = self.final_layer_sum();
        let naive_width = self.specs.width_l + self.specs.width_r;
        let sign_bit = is_signed && *o.bits.last().unwrap();

        assert!(o.bits.len() <= naive_width);
        while o.bits.len() < naive_width {
            o.bits.push(sign_bit);
        }
        o
    }

    /// Return counts of all subproduct bits by place value.
    pub fn solver_counts_hook(&self) -> Vec<usize> {
        let mut v = Vec::new();
        for pv_bit_indices in &self.from_lpi[1] {
            v.push(pv_bit_indices.len());
        }

        v
    }

    /// Return ascending place values of all subproduct bits.
    pub fn solver_pvs_hook(&self) -> Vec<usize> {
        let mut v = Vec::new();
        for (pv, bit_indices) in self.from_lpi[1].iter().enumerate() {
            for _ in bit_indices {
                v.push(pv);
            }
        }
        v
    }

    /// Truncate circuit to subproduct expansion only.
    pub fn truncate_to_one_layer(&mut self) {
        self.bits.truncate(self.l0_trunc_to);
        self.from_lpi.truncate(2);
        self.ram_layers.truncate(1);
        if self.l0_dirty {
            // TODO I have yet to see runs where bypassing this
            //      restoration causes incorrect test muls. What's up?
            self.ram_layers[0] = self.l0_backup.clone();
            self.l0_dirty = false;
        }
    }

    /// Compute actual cost of this circuit as it now exists.
    /// The metric is (n_layers, n_rams, n_ram_cells_used).
    pub fn actual_cost(&self) -> SolutionCost {
        let mut c = (self.ram_layers.len(), 0, 0, self.n_compos);

        for layer in &self.ram_layers {
            c.1 += layer.len();
            for ram in layer {
                c.2 += ram.outputs.len() << ram.inputs.len();
            }                
        }

//p!(R "actual cost counted is {:?}", c);
        c
    }

    /// This is like reserve_for_node() in that it provides for a
    /// a new node in self.bits, but it's sloppy in that it doesn't
    /// use the from_lpi mechanism (which is very clunky and rather
    /// inappropriate for non-addend uses). To know what the bit does,
    /// one would have to look at the RAMs it's attached to. It's
    /// kind of ad-hoc. It's actually a sensible idea, the main problem
    /// is that much of my code that would benefit from this idea follows
    /// a different idea.
    pub fn alloc_bit_ix(&mut self) -> usize {
        let index = self.bits.len();
        self.bits.push(false);
        index
    }

    /// Check the multiplier to ensure that any bit that a RAM reads
    /// was written by a prior layer, not the same layer, a later layer,
    /// or not even written at all.
    pub fn time_sanity_check(&self) {
        let mut last_write = vec![None; self.bits.len()];
        let mut first_read = vec![None; self.bits.len()];

        for (when, layer) in self.ram_layers.iter().enumerate() {
            for ram in layer {
                for bit in ram.outputs.iter().copied() {
                    match last_write[bit] {
                        None => last_write[bit] = Some(when),
                        Some(time) => last_write[bit] = Some(time.max(when)),
                    }
                }
                if when != 0 {      // Layer 0 inputs are written by test code.
                    for bit in ram.inputs.iter().copied() {
                        match first_read[bit] {
                            None => first_read[bit] = Some(when),
                            Some(time) => first_read[bit] = Some(time.min(when)),
                        }
                    }
                }
            }
        }

        for bit in 0..self.bits.len() {
            match first_read[bit] {
                None => (),
                Some(r_time) => match last_write[bit] {
                    None => panic!("input bit {} was never written", bit),
                    Some(w_time) => if r_time <= w_time {
                        panic!("input bit {} not written early enough", bit);
                    },
                },
            }
        }
    }
}
